//
// GameX - ImageX Class Header 
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source Liscense. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
										// **** Image includes
#include "gamex-buffer.hpp"
#include "gamex-file.hpp"
#include "gamex-win-dx.hpp"

#ifndef IMAGEX_DEF
	#define	IMAGEX_DEF
										// **** Image Switches
	#define IMAGEX_DEBUG				// Image debugging (bounds, null data, etc.)
	//#define IMAGEX_SUPPORT_RAW			// Image support for RAW file load/save
	#define IMAGEX_SUPPORT_TIFF			// Image support for TIFF file load/save
	#define IMAGEX_SUPPORT_BMP			// Image support for BMP file load/save
	
	#define DRAW_UNDEF		0x0000		// can't be drawn or operated on

#ifndef IMAGE_LOAD_MODES_DEFINED
	#define IMAGE_LOAD_MODES_DEFINED

	// Load Flags: (combinable)
	// Use these flags when loading an image with Load() or when preparing an image with Create()
	#define LOAD_ALPHA		0x01 // load using a transparency layer that's part of the image -- currently only affects DRAW_PLAIN and DRAW_ADDED drawing of this image
	#define LOAD_MASKED		0x02 // load using a mask to exclude 0-value pixels, or converts alpha layer to mask if LOAD_ALPHA is specified too
	#define LOAD_TARGETABLE	0x04 // required when loading or any image that you are going to draw into
	#define LOAD_NON3D		0x08 // tells GameX to load this image without support for 3D-accelerated drawing (from)
							     // use this only for images that will be drawn with no alpha or blending or rotation
								 // and usually no scaling (can be masked or non-masked) with DRAW_PLAIN only
							     // in which case it will draw about 2x faster, otherwise this flag will cause disastrously slow drawing
							     // (this flag may not be fully supported and is for advanced users)
//	#define LOAD_DITHER		0x10 // NOT YET IMPLEMENTED -- dither the image when loading it to preserve the appearance of true-color quality even in high-color mode
//	#define LOAD_OPTIMIZED	0x20 // NOT YET IMPLEMENTED -- force loaded image to a size such as 128x128 or 256x256 that D3D can best handle and can automatically draw with wrapping
	#define LOAD_GRAY		0x40 // NOT YET TESTED -- convert color to grayscale upon loading (doesn't change the image file of course)
//	#define LOAD_ALLOWLOWQ	0x80 // NOT YET IMPLEMENTED -- if this image is loaded with LOAD_ALPHA, allow GameX to reload this image with LOAD_MASKED instead if in low-quality mode.
	#define LOADMODESMASK	0x4F // DO NOT SPECIFY THIS FLAG -- this is a mask for all of the load modes and is only to be used internally

	#define LOAD_SCREENTARGET (LOAD_TARGETABLE|LOAD_NON3D) // if copying from the screen, copy into an image loaded with this, otherwise the copy will be pretty slow

#endif



	#define IMG_OK			0
	#define IMG_NOTREADY	1
	#define IMG_IGNORE		2
										// **** Image Pixel Channels
	#define CHAN_INDEXED		0	
	#define CHAN_RED			1
	#define CHAN_GREEN			2
	#define CHAN_BLUE			3
	#define CHAN_ALPHA			4
	#define CHAN_GRAY			5
										// **** Image Pixel-Format Information
	#define IMG_UNDEF			0x00	// Image information is undefined.
	#define IMG_ALPHA			0x01	// Image has an alpha channel (follows color data)
	#define IMG_MERGEALPHA		0x02	// Alpha channel is merged with data at pixel level (RGBA)	
	#define IMG_INDEXED			0x04	// Image is indexed
	#define IMG_LOWCOLOR		0x08	// Image is low color.	(BPP=4-4-4-[A])
	#define IMG_HIGHCOLOR		0x10	// Image is high color. (BPP=5-6-5-[A])
	#define IMG_TRUECOLOR		0x20	// Image is true color. (BPP=8-8-8-[A])
	#define IMG_TRUECOLORX		0x40	// Image is true color. (BPP=8-8-8-8X)
	#define IMG_MASKED			0x80	// Image uses a 0-mask (instead of alpha)

										// **** File Formats
	#define IMG_RAW_FORMAT		0		// RAW format
	#define IMG_TIFF_FORMAT		1		// TIFF format	


	typedef unsigned char		ImageOps;	
	typedef unsigned int		ImageUsage;	

	#define DEGtoRAD	((float)(3.141592/180))

	class ImageX {
	public:		
		ImageX ();										// Constructors
		ImageX (int xr, int yr, ImageOps ops);
		~ImageX ();										// Destructor

		// Functions for game to call:

//		void ScaleAndRotate (ImageX *dest, float ang, float scale, ImageUsage usagePlus = 0);
//		void Rotate (ImageX *dest, float ang, ImageUsage usagePlus = 0) {ScaleAndRotate(dest,ang,1.0f,usagePlus);}
//		void Scale (ImageX *dest, float scale, ImageUsage usagePlus = 0) {ScaleAndRotate(dest,0,scale,usagePlus);}
		int GetWidth (void)	{return m_xres;}
		int GetHeight (void) {return m_yres;}
		void Create (int xr, int yr, ImageUsage u);
		void Fill (int r, int g, int b);
		void SetPixel (int x, int y, int r, int g, int b);
		void GetPixel (int x, int y, int &r, int &g, int &b);

		// Functions that only ImageX, ImageExt, and WindowsDX should call:
		
		void Size (int xr, int yr, ImageOps ops = IMG_HIGHCOLOR);	// Construct Image
		inline void TurnOn (ImageOps opt);					// Turn on a PixelFormat option
		inline void TurnOff (ImageOps opt);					// Turn off a PixelFormat option
		inline int GetBytesPerPixel (ImageOps ops);			// Get Bytes-Per-Pixel
		inline int GetFormat (ImageOps ops);				// Get Format 
		inline int GetAlpha (ImageOps ops);					// Get Alpha
		inline ImageUsage GetUsage (void)	{return m_usage;}
		void UseWith (ImageUsage what);
		void DeleteData (void) { if(m_data!=NULL) {delete [] m_data; m_data = NULL;}
								 if(m_alpha!=NULL) {delete [] m_alpha; m_alpha = NULL;} }
		void Reset (void);
		void Restore (void);

		// Variables that only ImageX, ImageExt, and WindowsDX should access:

		LPDIRECTDRAWSURFACE7 m_surface;						// Color Data
		int m_hassurface; // whether or not this image has a surface / texture at the moment
		int m_lost_counter; // how many times the surface has been restored
		int m_surf_width, m_surf_height, m_surf_bpp; // texture dimension for accurate mapping
		LPDIRECT3DDEVICE7 m_device; // D3D accelerator
		XBYTE *m_data;										// Color Data
		XBYTE *m_alpha;										// Alpha Channel
		int m_xres, m_yres, m_size;							// Resolution/Size info
		ImageUsage m_usage;									// Usage info
		ImageOps m_options;									// Pixel Format
		char status;										// Image Ready?
		char * m_filename;									// Filename of ImageX
	private:

		// Functions that only ImageX should call:

		void DataToSurface (void);
//		void RotateImage (ImageX *dest, float ang, float scale, bool toVideo = false);
//		void RotateAlpha (ImageX *dest, float ang, float scale);
		inline void AddUsage (ImageUsage u) {m_usage |= u;}
		inline void RemoveUsage (ImageUsage u) {m_usage ^= u;}
		void ConvertFormat (ImageOps pf);
		void ConvertAlphaToMask (void);
		void ConvertMaskToAlpha (void);
		void ConvertMDataToGrayscale (void);
//		void ConvertAlphaToBlend (void);
		void AllocateSurface (void);
	};

	class ImageExt {
	public:
		ImageExt ();
		ImageExt (ImageX* pImg);
		~ImageExt ();
		void Load (char *filename, ImageX *img, ImageUsage what=0);
	private:
		#ifdef IMAGEX_SUPPORT_TIFF
			void LoadTiff (char *filename, ImageX *img); // implemented in gamex-image-tiff.cpp
			void SaveTiff (char *filename, ImageX *img);
			void LoadTiffDirectory (File &tiff, Buffer &code);
			void LoadTiffEntry (File &tiff, Buffer &code);
			void LoadTiffStrips (File &tiff, Buffer &code);
			void LoadTiffData (File &tiff, Buffer &code, unsigned long count, unsigned long offset, int row);
			void SaveTiffDirectory (File &tiff, Buffer &code);
			void SaveTiffEntry (File &tiff, Buffer &code, unsigned int tag);
			void SaveTiffExtras (File &tiff, Buffer &code, unsigned int tag);
			void SaveTiffData (File &tiff, Buffer &code);			
		#endif
		#ifdef IMAGEX_SUPPORT_BMP
			void LoadBMP (char *filename, ImageX *img); // implemented in gamex-image-bmp.cpp
		#endif

		ImageX*				m_pImg;					// Pointer to an ImageX		
		char				m_filename[128];		// Filename of ImageX
		int					m_status;				// Status of ImageX
		
		#ifdef IMAGEX_SUPPORT_RAW					// RAW Format Info
			int				m_bpp;
			int				m_channel;
			int				m_header;
		#endif

		#ifdef IMAGEX_SUPPORT_TIFF					//**** TIFF Format Info
			int				m_xres, m_yres;			// TIFF Resolution
			int				m_ops;					// TIFF Pixel-Format Options for ImageX
			int				m_mode;					// TIFF Mode
			int				m_alpha;				// TIFF Alpha Channel Present?
			int				m_num_chan;				// TIFF Number of Channels
			unsigned long	m_bpcoff;				// TIFF 
			int				m_bpc[5];				// TIFF Bits per Channel
			int				m_compress;				// TIFF Compressed?
			int				m_photo;
			unsigned long	m_num_strips;			// TIFF Number of Strips
			unsigned long	m_strip_offsets;		// TIFF Strip Offsets
			unsigned long	m_strip_counts;			// TIFF Strip Counts
			int				m_rps;					// TIFF Rows per Strip
			int				m_bpp;					// TIFF Bits per Pixel
			int				m_bpr;					// TIFF Bits per Row
		#endif
	};
#endif
