//
// GameX - WindowsDX Class Code (including WinMain and WinProc)
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source Liscense. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

// this is a version of GameX that has been modified (by Justin Pease)



// performance note: drawing into an image is currently not hardware-accelerated,
// except for non-scaled non-rotated non-blended DRAW_PLAIN or MASKED drawing.
// if you need to draw into an image using another blending mode or with rotation etc.,
// consider drawing into the VIEWPORT and doing the blending there, then copying back out of the viewport,
// and then doing the actual drawing on the viewport that will be seen each frame.
// This is only if you need to do it every frame of course, the slowness is acceptable if you can
// manage to do it only once and then use the modified image as a source to draw into the viewport.


// known bugs:
//
// *** high priority:
//
// (The many bugs that were once in this category now appear to be dead.)
//
// *** medium-high priority:
//
// Certain video cards misinterpret Direct3D 7.0 drawing mode instructions.
// I've only actually seen this with old RAGE cards.
// The result is that, on these older computers,
// DRAW_SUBTRACT and DRAW_INTENSIFY (and possibly others) are interpreted as DRAW_PLAIN,
// which makes games that use these draw modes look horrible.
// One could argue that they should look horrible anyway on such old video cards
// but GameX should really have as much compatibility as possible.
// An acceptable fix would be to find a way to detect for draw mode failures
// and revert to direct non-accelerated pixel drawing with these modes
// or use a backup 3D device that doesn't go through the video card (MMX hopefully)
// to draw only images that use these modes, if the video card misinterprets them.
// The problem is that detecting for such failures is difficult because the video card itself
// does not admit that it can't handle these cases, so maybe the easiest and most reliable
// way to detect for failures is to draw a few really tiny sample images during initialization
// like a 4x4 white box on a grey background, and if a resulting pixel is way off (> 25% difference from expected)
// then remember to revert to 2D drawing or the backup D3D device when drawing in failed modes
// of course, this would make the initialization process longer and add even more overhead to drawing...
//
// certain video cards (again, old RAGE cards)
// don't draw images using their transparency values
// (i.e drawing a normal image at 50% transparency results in it being drawn at 100%)
// I have no idea if this problem is more widespread, and if it is,
// there is probably something wrong with the 3D drawing code,
// some simple flag I forgot to set that some video cards assume is set anyway, for instance.
//
// (should fix soon since doing so makes GameX input more convenient to check for)
// when GetKey() is called to find if a key was just pressed,
// it returns true if the key was just pressed, BUT, if GetKey() is called
// for the same key in the same call of main(), it will then return false
// even though it should return true again because the game is still
// at the same instant in time as far as the programmer is concerned.
// CURRENT WORKAROUND for game programmer: if you need to GetKey() check for
// a keypress in multiple places in main(), use GetKey() on that key
// once per call to main() and store the result in a boolean variable,
// then refer to that variable later in main() to check whether the key is down.
// the same applies to AnyKeyPress(), GetMouseClick(), and AnyMouseClick()
// this also applies to the Peek functions if they are called after these functions
// (which generally they shouldn't be)

// *** medium priority:
//
// (would fix this if I knew how to, maybe it requires using a newer version of Direct3D):
// drawing an image with D3D into another image (not into the back buffer)
// does not take advantage of 3D hardware acceleration, only MMX software acceleration at most
// this is bad because it makes a lot of complex offscreen compositing techniques too slow to be practical
// unless you do some trick like using the back buffer as a temporary offscreen surface between frames
// update: this technique is supported by SOME video cards to allow for reflection and soft shadow effects
//
// when resuming music that has been paused, oftentimes a single note will play for quite a few seconds
// over the rest of the music which is continuing as normal.
// this is due to the tricky use of tempo change to pause the music,
// a better solution would be to remember where it is and stop it, then start again at that point on resume,
// but this requires making use of DirectMusic 8.0 interface for PlayEx()
// (which is possible to do while remaining compatible with computers that only have DX7)
//
// (maybe more a lack of a feature than a bug though):
// there's currently no way to load an image with an alpha channel from two BMP files
// this one's easy to fix, just haven't gotten around to it yet...
//
// (could be a pain to fix but should be done soon):
// Blur() causes a mosaic effect (extremely low quality blurring) on some images,
// such as an alpha image, due to lack of hardware resampling support for the pixel format.
// need to make a check for 2D capabilities and switch to 3D (DrawImage3D) if not otherwise supported
//
// some video cards misinterpret certain D3D draw modes
// specifically, they draw DRAW_SUBTRACT and DRAW_INTENSIFY as if it were DRAW_PLAIN
// which looks horrible, should find a way to check for this support
// (may have to use kludgy way of drawing and comparing pixels)
// and if it isn't supported, treat DRAW_INTENSIFY like DRAW_ADD and DRAW_SUBTRACT like ...?
// or switch to pixel-by-pixel drawing (hardware emulation) drawing of these two modes then
//
// *** medium-low priority:
//
// images loaded with the LOAD_NON3D flag cannot be rotated, mirrored, or vertex-shaded
// workaround: draw the image into an image that wasn't loaded with the LOAD_NON3D flag and rotate it from there,
//			   or just don't load it with the LOAD_NON3D flag in the first place
//
// support for non-D3D non-2D accelerated image drawing
// is currently only implemented for 16-bit display mode.
// maybe there's an elegant way to extend the current support to 24 and 32 bit modes
// without making 2 duplicates of 6 drawing functions?
//
// *** low priority:
//
// (should fix eventually):
// X1R5G5B5 display mode may not actually be supported,
// because I've never actually seen a monitor that uses 5,5,5 mode instead of 5,6,5 for 16-bit
// i.e., R5G5B5 mode *should* work but has not been tested and so probably doesn't work.
//
// dialog boxes don't work right when page flipping is on.
// temporary fix: all dialog boxes are skipped when page flipping is on
// better fix that's not yet implemented: switch out of exclusive mode temporarily
// for the dialog box, then switch back in
//
// drawing WITHOUT D3D with DRAW_MULTIPLY fails completely.
//
// *** very low priority:
//
// (maybe won't fix):
// images drawn with D3D with DRAW_SUBTRACT aren't subtracted with complete accuracy,
// because D3D 7 doesn't support Blend Ops besides Add,
// so currently using a workaround that's close to subtracted but a little bit off
// (since (dest - source) can't be achieved exactly without the Subtract BlendOp,
//  DRAW_SUBTRACT currently uses (dest - 2*source*dest) instead.)
//
// drawing WITHOUT D3D with DRAW_GHOST (or DRAW_BURN) is disastrously slow and isn't completely accurate,
// because the operation that DRAW_GHOST performs is somewhat complex
// and I haven't figured it out yet, so in the meantime without D3D it just subtracts then adds
// to get an effect that's kind of similar to DRAW_GHOST.
// This is low priority because you should never draw an image in DRAW_GHOST without D3D anyway.






// other todo:

// clean up some code, for instance, there are public variables that really ought to be private

// move the GetClock() and GetClocksPerSec() code somewhere, maybe into a new source file
// as part of a ClockX object or something that games can use for time limits
// (although games should really use the CPS and count of calls to main() for time limits,
//  these functions are really designed to get ultra-accurate time measurements needed
//  (or at least very helpful) for frame rate and game speed management)

// allow for the game to get a rough idea of how fast a computer it's running on
// so game programmers can, for example, leave out slow drawing operations
// if the computer/graphics card is too slow to draw them each frame

// make the GetKey()-type functions work even if called twice in a row
// which means only updating what GetKey() was called on
// inside UpdateInput()
// update: this is now considered a bug too

// improve drawing interface, integrate things better

// better support for alpha, for instance, ability to copy only alpha or only non-alpha between images,
// ability to turn the alpha layer off temporarily for images (maybe)

// allow specification of 4 x,y coordinate pairs when drawing to explicitly define the quadrilateral
// polygon for warping/3D effects of image textures

// because of the many features DrawImage provides,
// it seems to have considerable overhead,
// which perhaps advanced users should have some means
// of skipping past to use only the features they want
// at the expense of more confusing drawing code in their game

// enable/fix on-draw rotation for non-D3D images that may be slow but must be stable enough to replace ImageX::Rotate() (which is currently broken, but trivially)
// non-3D support for gauraud shading of all 4 colors given to Draw(), including alpha blending for DrawBlended()
// combine all the 2D drawing functions (DrawImage(), DrawAlpha(), DrawAdded(), etc.) into one function
// enable horizontal and vertical mirroring for 2D drawing functions (although it really slows down the 2D-accelerated ones when used)

// implement a method of easily saving and loading the game
// update: maybe the functions in gamex-file.hpp are sufficiently simple to use for this purpose
//         however, if an interface for netplay is implemented (see below),
//         part of it may be so similar that it might as well be adapted
//         to provide an easy interface for saving/loading games as well

// cutscene fmv movie player, using the cutscene directshow sample code
// (what version of DirectX would this require? If higher than 7.0,
//  it would still be possible to play the game with DirectX 7.0
//  (unless the code for it is added sloppily) but the fmv cutscenes wouldn't show)
// of course, cutscenes of this variety are non-interactive and gigantic memory-wise
// but they're common enough that it should be possible to do them with GameX

// enable easy networkable multiplayer game support via DirectPlay
// there are two types of networking situations with DirectPlay:
// Peer-to-peer -- simple, no central server hub, can't support large amounts of players very well, but that's OK for most games
// Client-server -- a server manages game and synchronizes things, can support many players, requires a decent server, may be easier to implement though
// Lobby -- I think this is a combination of peer-to-peer and client-server, a remote server manages lobby and sessions, may require separate lobby application (which wouldn't fit in well)...looks very complicated to implement
//
// I think it would be best to base this feature off of the SimplePeer example in the DX9SDK Samples.
// the actual structuring of the networked game is the tricky part though, since it has to be
// something that makes sense for most games and is extremely easy to implement from the game programmer's point of view.
// ideally it would be automatic beyond main() and draw() having to react differently depending on a variable
// that says who the current player is (i.e. which computer this is)
// how I think it could work:
// the host computer calls main() like normal, and draw() to update the screen like normal,
// the other computers do not call main(), but they still call draw()
// thus, the main() called by the host computer responds to EVERYONE's input and manages the whole game
// there are two types of messages that are automatically sent to maintain the network game:
// INPUT MESSAGES -- sent from all computers to the host computer
//					 these messages contain a list of relevant keys and/or mouse info,
//					 which the host uses in main() to determine what each player's input causes
// UPDATE MESSAGES -- sent from the host computer to all computers
//					  these messages contain a list of relevant update variables
//					  which the computers use to update their display (and audio)
//
// the host computer can be a player in the game too, in which case it can "send messages to itself" (optimized to just use own data)
// and update the game with draw() like a client, otherwise it won't need to call draw() or generate input messages / respond to update messages
// 
// all computers including the host still call initialize(),
// which, for network support, must in addition to normal initialization:
// * specify which variables are "net variables" (i.e. must be synchronized for proper game updates)
//	 (to be on the safe side and allow for non-hosts to become hosts should the host disconnect,
//	  you should make everything that is used in the main() or draw() function a "net variable"
//	  unless it is completely readied in initialize() or not used during gameplay
//	  now, this may seem like a lot of variables, but if you have a bunch of variables
//	  in a class or structure that is used by main(), just specify the class/structure as a variable
//	  and everything inside that class/structure will automatically be taken care of)
//	 (Some things should NOT be made net variables: ImageX, SoundX, or MusicX objects
//	  these should be loaded in initialize(), and if not, they should be loaded by each computer
//	  according to signal variables to avoid having to transmit entire images over the network.
//	  if you need to update an object that has an ImageX in it...umm, still working on it, but for now just don't
//	  possible solution: use RemoveNetVar() on the specific parts of it that you don't want updated,
//	  GameX will update everything and then un-update the variables in the removed list
//	  if the removed variables aren't exactly an entire var added with AddNetVar()
//	  also: maybe the effects of AddNetVar() will be needed to enable game-saving too, so might rename these)
//	 Net variables are specified by calling AddNetVar, like so:  GameX.AddNetVar(&var, sizeof(var));
//	 this enables GameX to handle messages/packets such that these variables are synchronized
//	 (which it does by associating each variable with an index, and operating under the assumption
//	  that each computer will add the same variables in the same order during initialize(), 
//	  and stores all of the relevant variables on the host computer into a message which is received
//	  and used to refresh each net variable on each computer)
// * specify network settings such as the maximum number of players (no minimum; the host can enforces that if desired)
//	 and the desired means of connection
//	 with a call to GameX.NetInitialize() (which would be in addition to GameX.Initialize())
//	 (the actual connection is made
//	  after a dialog box that asks whether you're hosting or joining and where, then it starts...
//	  other players can continue to join until the max, the host decides what to do with each player
//	  so for instance it could keep new additions just watching the action until their turn to play
//	  by sending them the proper update info and ignoring the spectator's input data)
// * call GameX.NetStart() if the game is network-only,
//	 otherwise call GameX.NetStart() in main() when a choice to enter/begin a multiplayer game is made
//	 (before GameX.NetStart() is called, both main() and draw() are called as normal for single-player)
//	 GameX.NetEnd() can be called to return everyone to a single-player non-networked state.
//	 also, GameX.NetHost() and GameX.NetJoin() can be called if the game has a mechanism for
//	 choosing whether to host or join, instead of GameX.NetStart() which pops up a dialog asking which to do.
//
// there should probably be a simple "chat" system implemented, that can optionally be activated by the game,
// or even better, the network code would be good enough that implementing a custom chat system
// is very easy to do in the game code
//
// how would main() respond to input?
// the input-getting functions will each take an extra parameter that defaults to 0,
// where 0 means all players or the only player if not networked,
// and 1,2,3,4,etc. mean the input of player 1, 2, 3, or 4 (etc.), respectively 
// and input for each player is kept separately so main() can
// loop through the number of current players and respond to the input of each one

// full polygonal 3D support where each polygon can be texture mapped and/or gauraud shaded
// need to decide on a format for 3D object files then, maybe...

// allow for 24 and 32-bit images to be loaded with their high color level preserved,
// so they can be drawn dithered for 16 bit
// right now, you can kind of do this by loading the image with alpha,
// and then draw it with dithering enabled
// but the color preservation is just a side-effect of the 32-bit format with alpha imbedded,
// and would result in bad performance compared to dithering it to 16-btt upon loading

// better 24 and 32 bit support, and 8 bit depth support

// move from DirectX 7 to DirectX 8 for some things, such as DirectMusic and DirectSound,
// which have some cool features in DX8 that aren't in DX7 (such as wav file loader for DirectMusic
// and ability to apply sound FX to sound buffers in DirectSound and through it DirectMusic too)

// joystick support using DirectInput's 2.0 interfaces
// the next possible step would be simple force-feedback capability

// implement some mechanism to allow custom drawing functions to be possible
// at the least, a GetImagePixels function that must be followed by ReleaseImagePixels after drawing to it

// find some way to accelerate text drawing or at least allow for multiple sizes of text

// allow specification of colors even when flags aren't specified

#include "gamex-win-dx.hpp"

// WindowsDX Constructor
WindowsDX::WindowsDX (void)
{	
	win_request_xsize = -1;
	win_request_ysize = -1;
	win_status = GAMEX_NO_INIT;
	table_alpha = NULL;
	num_sounds = 0;
	max_sounds = 0;
	sound_list = NULL;
	ds_sound_list = NULL;
	sound_support = false;
	drawLocked = false;

	SetGameName ("");
	SetGameXVer ("GameX R5.0beta");
}

void WindowsDX::UpdateCaption ()
{	
	if (strlen(game_name)==0) sprintf (game_name, "UNTITLED");
	sprintf (game_caption, "%s - %s", game_name, gamex_ver);
	
	if (win_hwnd!=NULL) SetWindowText (win_hwnd, game_caption);
}

WindowsDX::~WindowsDX (void)
{
	if (table_alpha!=NULL) delete [] table_alpha;
}

// WinMain and WindowFunc functions supplied by WindowsDX
LRESULT CALLBACK WindowFunc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message) {
	case WM_MOVE:
		GameX.SetPosition (LOWORD(lParam), HIWORD(lParam)) ;
		GameX.SetMouseStop(1); // so window movement doesn't register as mouse movement to the game
		break;
	case WM_CHAR: // for GetKey() buffered keyboard input
		if (wParam==27) {
			if(GameX.CheckQuit()) {
				GameX.SetStatus(GAMEX_SHUTTINGDOWN);
				DestroyWindow (hWnd) ;
			}
		}
		GameX.AddBufferedKeyPress ((TCHAR) wParam) ;
		return 0;
		break;
	case WM_SYSKEYDOWN:
		if(wParam == VK_MENU)
			return 0; // prevent pressing the alt key on its own from pausing/freezing the game
		break;
	case WM_SYSCOMMAND:
		if(wParam == SC_KEYMENU || wParam == SC_PREVWINDOW/* || wParam == SC_NEXTWINDOW || wParam == SC_SCREENSAVE*/)
			return 0; // prevents ALT-anykey from blinking the window and beeping
					  // and prevents ALT-F6 from pausing the game
					  // but allows ALT-F4 to continue functioning
		// no break
	case WM_ACTIVATEAPP:
		// no break
	case WM_ENTERMENULOOP:
		if(GameX.CanDraw() && GameX.CanAutoPause()) {
			GameDraw();
			GameX.win_alternator = 0;
			GameX.SpecialViewFade(160);
		}
		GameX.ViewUpdate() ;
		break;
	case WM_KILLFOCUS:
		if(GameX.win_active) {
			GameX.SetActivation(false) ;
			if(GameX.CanHideCursor())
				GameX.UnhideCursor();
			if(GameX.CanAutoPause())
				GameX.PauseMusic();
			GameX.SetMouseStop(5); // skip 5 input loops, the mouse should be ready for input by then
		}
		break;
	case WM_SETFOCUS:
		if(!GameX.win_active && GameX.win_realy>1) {
			GameX.SetActivation (true) ;
			if(GameX.IsFullScreen() && GameX.CanHideCursor())
				GameX.HideCursor();
			if(GameX.CanAutoPause())
				GameX.ResumeMusic();
		}
		break;
	case WM_PAINT:
		GameX.ViewUpdate();
		break;
	case WM_CLOSE:
		if(GameX.CheckQuit()) {
			GameX.SetStatus(GAMEX_SHUTTINGDOWN);
			DestroyWindow (hWnd) ;
		}
		else
			return 0;
		break;
	case WM_DESTROY:
		PostQuitMessage(0) ;
		break;


	// these are used to check to make sure DirectInput's mouse presses are within the window
	case WM_LBUTTONDOWN:	GameX.SetMousePress(MOUSE_LEFT);			return 0;
	case WM_MBUTTONDOWN:	GameX.SetMousePress(MOUSE_MIDDLE);			return 0;
	case WM_RBUTTONDOWN:	GameX.SetMousePress(MOUSE_RIGHT);			return 0;

	// these are used to allow double-click detection (only trigger if CS_DBLCLKS is set on our WNDCLASS.style)
	case WM_LBUTTONDBLCLK:	GameX.SetMouseDoubleClick(MOUSE_LEFT);		return 0;
	case WM_MBUTTONDBLCLK:	GameX.SetMouseDoubleClick(MOUSE_MIDDLE);	return 0;
	case WM_RBUTTONDBLCLK:	GameX.SetMouseDoubleClick(MOUSE_RIGHT); 	return 0;

	case WM_SIZE:
		if(GameX.GetStatus() == GAMEX_READY) {

			if(HIWORD(lParam) < 1) {
				PostMessage(GameX.GetWindow(),WM_KILLFOCUS,0,0);
			} else {
				PostMessage(GameX.GetWindow(),WM_SETFOCUS,0,0);
			}

			GameX.win_realx = LOWORD(lParam);
			GameX.win_realy = HIWORD(lParam);

			if(GameX.win_realy < 1) GameX.win_realy = 1; // avoids crash
			if(GameX.win_realx < 1) GameX.win_realx = 1;
			GameX.SetMouseStop(1); // so window resize doesn't register as mouse movement to the game
		}
		break;

// obselete:
//	case WM_ACTIVATEAPP:	GameX.SetActivation((BOOL)wParam) ; break;
//	case WM_KEYDOWN:		GameX.SetKeyDown ((TCHAR) wParam) ; break;
//	case WM_KEYUP:			GameX.SetKeyUp ((TCHAR) wParam) ;	break;
//	case WM_SYSKEYDOWN: 	GameX.ClearPressed() ;				break; 
//	case WM_ENTERIDLE:		break;
//	case WM_CREATE: 		break;
//	case WM_LBUTTONUP:		GameX.SetMouseRelease(MOUSE_LEFT);			return 0;
//	case WM_MBUTTONUP:		GameX.SetMouseRelease(MOUSE_MIDDLE);		return 0;
//	case WM_RBUTTONUP:		GameX.SetMouseRelease(MOUSE_RIGHT); 		return 0;
//	// used to track the absolute position of the mouse,
//	// while DirectInput handles the relative position (and does a much better job of it)
//	case WM_MOUSEMOVE:		GameX.SetMouseMove(LOWORD(lParam), HIWORD(lParam)) ; return 0;

	}
	
	return DefWindowProc (hWnd, message, wParam, lParam) ;		
}

bool WindowsDX::CheckQuit(void)
{
	// if quit confirmation is off
	// or if we're page flipping in which case a dialog often won't show up
	// or if the user is holding the left shift key,
	if(win_request_no_confirm_quit || win_page_flipping || IsKeyDown(DIK_LSHIFT))
		return true; // signal to exit immediately
	else {
		int choice = 0;

		bool hide_cursor = (win_fullscreen && !win_request_no_hide_cursor);
		if(hide_cursor) {
			UnhideCursor();
			win_request_no_hide_cursor = true; // avoids re-hiding
			win_dialog_up = true;
		}

		choice = MessageBox (win_hwnd, "Are you sure you want to exit?", GameX.GetName(), MB_YESNO|MB_DEFBUTTON1|MB_ICONQUESTION);

		if(hide_cursor) {
			win_dialog_up = false;
			win_request_no_hide_cursor = false; // reset to hide
			HideCursor();
		}

		return (choice == IDYES); // if true, the game quits, if false, the game keeps going
	}
}

void WindowsDX::QualityCheck (void)
{
	// unless we're already in low quality mode,
	// and unless we're page flipping in which case a dialog often won't show up
	if(!win_page_flipping && !win_lowquality) {
		int choice = 0;

		bool hide_cursor = (win_fullscreen && !win_request_no_hide_cursor);
		if(hide_cursor) {
			UnhideCursor();
			win_request_no_hide_cursor = true; // avoids re-hiding
			win_dialog_up = true;
		}

#ifdef _DEBUG
		choice = MessageBox (win_hwnd, "This game appears to be running at an unacceptably low frame rate.\nThis may be because your video card, video driver, processor and/or BIOS are outdated.\nIf not, you may be using too many drawing functions/features,\n changing the game speed or sound/music effects too often,\nor using a very slow algorithm somewhere.\nPlease click No and ignore this message if it does not apply.\nOtherwise, would you like GameX to temporarily lower graphical output quality to attempt to compensate?\n(Your choice will remain in effect until you quit the game.)", "GameX Notice", MB_YESNO|MB_DEFBUTTON1|MB_ICONQUESTION) ;
#else
		choice = MessageBox (win_hwnd, "This game appears to be running at an unacceptably low frame rate.\nLook in \"readme.txt\" for possible causes and solutions to this problem.\nFor now, would you like GameX to lower the graphical quality to attempt to compensate?\n(Your choice will remain in effect until you quit the game.)", "GameX Notice", MB_YESNO|MB_DEFBUTTON1|MB_ICONQUESTION) ;
#endif

		if(hide_cursor) {
			win_dialog_up = false;
			win_request_no_hide_cursor = false; // reset to hide
			HideCursor();
		}
		
		if(choice == IDYES) {
			win_lowquality = true;
		}
	}
}


// returns the current time since startup with as much precision as this hardware supports,
// this will probably be in some ultra-precise unit, could be in fractions of a nanosecond
// (use GetClocksPerSec() to find out exactly how many of these units are in a second)
static unsigned long GetClock(void)
{
	LARGE_INTEGER time;
	if(QueryPerformanceCounter(&time)) // get the High Performance Timer count
		return time.QuadPart;
	else // if QueryPerformanceCounter() failed, that means the hardware doesn't support high-resolution timers
		return clock(); // so use the built-in low-resolution timer, clock(), instead. (only accurate to a 100th of a second)
}

// returns the number of time units (that GetClock() counts in) that are in one second.
// for instance, if this function returns 3579545, that means there are 3,579,545 clock ticks per second.
// note: you can assume the return value is a constant, although it may be different on different computers.
static unsigned long GetClocksPerSec(void)
{
	LARGE_INTEGER frequency;
	if(QueryPerformanceFrequency(&frequency)) // get the High Performance Timer frequency in Hz
		return frequency.QuadPart;
	else // if QueryPerformanceFrequency() failed, that means the hardware doesn't support high-resolution timers
		return CLOCKS_PER_SEC; // so use the low-resolution timer's "per second" value
}

// sleeps for at least the given number of time units,
// at most a bit over that amount,
// and increments GameX's idle counter accordingly.
// will only Sleep() in units of sleep_step milliseconds,
// so higher sleep_step means more accuracy but less actual sleeping  
// sleep_step should be an integer from 1 to 10 inclusive,
// clocks is the amount of time to sleep/delay in terms of GetClocksPerSec(), NOT milliseconds
inline static void AccurateSleep(long clocks, int sleep_step = 1)
{
	if(clocks <= 0) return; // wouldn't want to freeze upon sleeping a negative amount, also return quickly on 0 sleep
	long sleeptimer = GetClock();
	int sleepAmount = (clocks+1)*CLOCKS_PER_SEC/GetClocksPerSec();
	Sleep((sleepAmount/sleep_step)*sleep_step); // sleep for the amount of time that fits into sleep_step millisecond units
	GameX.sleep_record += GetClock() - sleeptimer;
	while(GetClock() - sleeptimer < clocks) ; // loop tightly for any time leftover
}


int WINAPI WinMain (HINSTANCE hThisInst, HINSTANCE hPrevInst, LPSTR lpszArgs, int nWinMode)
{
	MSG msg;
	
	GameX.WindowInitInstance (hThisInst, nWinMode) ;	

	GameInit();

	if (GameX.GetStatus()!=GAMEX_READY) GameX.ReportProblem() ; 
	GameX.SetStatus(GAMEX_READY) ;

	int forceDrawAll = 0;
	bool forceUpdateAll = false;

	int frame=0, cycle=0; // so we can keep track of the frame rate
	GameX.sleep_record = 0; // to keep track of how much time we spend sleeping

	const int clocksPerSec = GetClocksPerSec(); // get the number of time units in a second (varies from computer to computer)

	int cps = GameX.win_cps; // cycles per second (a cycle meaning a call to main(), not a CPU cycle)
	int fps = GameX.win_fps; // frames per second (a frame meaning a call to draw() and ViewUpdate())
	int tpc = clocksPerSec / (cps?cps:1); // time per cycle (number of clock ticks between each game update)
	int tpf = clocksPerSec / (fps?fps:1); // time per frame (number of clock ticks between each view update)
	bool needUpdate = true;
	int fps_too_low = 0; // number of times in a row that the fps is way too low
	int last_cpf_counter = fps?cps/fps+1:0;
	unsigned long cycleTime, frameTime, frameClock; // timers for game update, view update, and framerate calculation update
				  cycleTime =frameTime =frameClock =GetClock(); // (respectively)
	for(;;) {
		if(PeekMessage (&msg, NULL, 0, 0, PM_REMOVE)) {
			if (msg.message == WM_QUIT) break;
			TranslateMessage(&msg) ;
			DispatchMessage(&msg) ;
		}
		if(!GameX.win_active && GameX.CanAutoPause()) { // if the window isn't active, and we can pause,
			Sleep(10) ; // sleep so we don't max out the cpu in the background
			cycleTime = GetClock(); // set the timers to the current time
			frameTime = GetClock(); // so we don't skip a ton of frames when resuming

// THIS SHOULD GO INTO A FUNCTION CALLED GameX.DrawStatus ();
/*#ifdef _DEBUG
	if(!GameX.IsFullScreen() && (clock()/10)%100 == 0) { // every second if not in fullscreen mode
		char str [256];
		sprintf(str,"%s (Paused) - %d/%d fps - %d/%d cps - %d/%d idle",GameX.GetName(),0,GameX.GetTargetFrameRate(),0,GameX.GetTargetCycleRate(),100-rand()*rand()*2/RAND_MAX/RAND_MAX,100);
		SetWindowText (GameX.GetWindow(), str);
	}
#endif*/
		} else {
			// AUTOMATIC GAME SPEED AND FRAME RATE MANAGEMENT:
			// This ensures that the game always runs at a given speed (chosen by the game programmer)
			// regardless of how quickly the computer is able to execute the game's drawing/instructions.
			// What this means is that the programmer need not worry about speed fluctuations which normally
			// occur, and moreover that the game will remain playable on computers of the past and future.
			// No matter what drawing commands the game uses, and no matter how fast the computer is,
			// the game will never run too fast or too slow (within reasonable limits).

			// For these reasons, this speed management is desirable for almost all games imaginable. However,
			// if for some reason it is undesirable, RUN_UNMANAGED can be set which causes the speed to be unregulated.
			// Note that a game can change win_cps while running to achieve effects like fast-forward or slow-motion.

			// As far as the game programmer is concerned, here's basically how it works:
			// 1.) You put all drawing code into a function called draw(), and all other code into main() (these can still call other functions of course)
			// 2.) You choose a game speed that you want the game to run at (60 to 100 cps is the recommended range). Set this with GameX.Initialize().
			// 3.) The game is automatically managed to run at that rate and the maximum framerate (by timing the calls to draw() and main()).
			// 4.) Even if the game can't hit the target frame rate, it will still go as fast as if it did.
			// 5.) You can call GameX.GetCurrentFrameRate() at any time to get the frame rate, no need to calculate it.

			// CPS means Cycles Per Second. Higher CPS means faster gameplay.	main() is called win_cps times per second.
			// FPS means Frames Per Second. Higher FPS means smoother gameplay. draw() is called win_fps times per second. 
			// Attaining the target CPS (game speed) is nearly guaranteed under normal circumstances, (< 10000 cps)
			// whereas attaining the target FPS (game smoothness) isn't always possible, depending on the game and cpu.
			// To find out how many FPS a game is actually getting, have it call GameX.GetCurrentFrameRate().
			// (To find out how many FPS it's trying to get, call GameX.GetTargetFrameRate())

			// A reasonable CPS setting is between 30 and 150 fps. 60 to 100 is recommended.
			// (If you think you need an unreasonable CPS speed, try altering the speed of object movement in main() instead.)
			// The target FPS is automatically set to the minimum of the monitor refresh rate and the CPS.
			// (You can set it manually if you wish, but there is rarely any reason to do this.)

			// Warning: the following code is VERY "fragile" due to its time-sensitive nature, so even
			// the most innocent modification is likely to cause subtle problems if not drastic failure.

			if (GameX.CanManageSpeed()) { // if speed management hasn't been disabled:

				// first check to see if the game changed the cps or fps and adjust accordingly
				if(cps != GameX.win_cps) {
					last_cpf_counter = last_cpf_counter*GameX.win_cps/cps;
					cps = GameX.win_cps;
					tpc = clocksPerSec / cps;
					if(fps > cps) tpf = tpc;


// THIS SHOULD GO INTO A FUNCTION CALLED GameX.DrawStatus()
/*					#ifdef _DEBUG
						if(!GameX.IsFullScreen()) {
							char str [256];
							if(GameX.CanUseMouse())
								sprintf(str,"%s - %d/%d fps - %d/%d cps - %d/%d idle - (%d,%d) mouse",GameX.GetName(),GameX.GetCurrentFrameRate(),GameX.GetTargetFrameRate(),GameX.GetCurrentCycleRate(),GameX.GetTargetCycleRate(),GameX.GetCurrentIdlePercent(),100,GameX.GetMouseX(),GameX.GetMouseY());
							else
								sprintf(str,"%s - %d/%d fps - %d/%d cps - %d/%d idle",GameX.GetName(),GameX.GetCurrentFrameRate(),GameX.GetTargetFrameRate(),GameX.GetCurrentCycleRate(),GameX.GetTargetCycleRate(),GameX.GetCurrentIdlePercent(),100);
							SetWindowText(GameX.GetWindow(),str);
						}
					#endif*/

				}
				if(fps != GameX.win_fps) {
					fps = GameX.win_fps;
					tpf = clocksPerSec / fps;
					tpf = max(tpf,tpc);
					
// THIS SHOULD GO INTO A FUNCTION CALLED GameX.DrawStatus()
/*					#ifdef _DEBUG
						if(!GameX.IsFullScreen()) {
							char str [256];
							if(GameX.CanUseMouse())
								sprintf(str,"%s - %d/%d fps - %d/%d cps - %d/%d idle - (%d,%d) mouse",GameX.GetName(),GameX.GetCurrentFrameRate(),GameX.GetTargetFrameRate(),GameX.GetCurrentCycleRate(),GameX.GetTargetCycleRate(),GameX.GetCurrentIdlePercent(),100,GameX.GetMouseX(),GameX.GetMouseY());
							else
								sprintf(str,"%s - %d/%d fps - %d/%d cps - %d/%d idle",GameX.GetName(),GameX.GetCurrentFrameRate(),GameX.GetTargetFrameRate(),GameX.GetCurrentCycleRate(),GameX.GetTargetCycleRate(),GameX.GetCurrentIdlePercent(),100);
							SetWindowText(GameX.GetWindow(),str);
						}
					#endif*/
				}

				// sleep while draw update would exceed game update
				if(!needUpdate)
					AccurateSleep(tpc+cycleTime-GetClock());
				
				unsigned long thisTime; // variable to hold pre-action reference time

				thisTime = GetClock();
				int cpf_counter = 0; // number of times to update the game at this instant,
									 // if > 1 then we skip (cpf_counter-1) frames by updating again before drawing them

				// calculate how many times we need to update the game
				while((int)(thisTime-cycleTime) > tpc) {
					cycleTime += tpc;
					cpf_counter++;
				}
				// limit number of times to the maximum allowable updates per frame
				if(cpf_counter > GameX.win_max_cpf) {
					cpf_counter = GameX.win_max_cpf;
					frameTime = GetClock();
				}

				// convert the current update number to a weighted average of itself and
				// the number of updates last time, to smooth out delay discrepencies between frames
				int ave = (cpf_counter+cpf_counter+cpf_counter + last_cpf_counter) / 4;
				int total = cpf_counter + last_cpf_counter;
				cpf_counter = total - ave; // takes into account the integer-divided nature of ave
				last_cpf_counter = ave;

				if(cpf_counter) { // if we need to update the game state at all since last time:
					while(cpf_counter--) { // now update (without drawing) as many times as calculated
						GameX.UpdateInput(); // first update the input so main() has fresh input data
						GameMain(); // call main() function, whose contents decide the behaviour of the game
						if(forceDrawAll==2 && GameX.CanDraw()) GameDraw();
						cycle++; // increment counter so we can know how many cps we're really getting
						if(/*cycle >= GameX.win_cps ||*/ cycleTime > thisTime) { // correct overflow
							cycleTime = thisTime;
							last_cpf_counter += cpf_counter;
							cpf_counter = 0;
						}
					}
					needUpdate = true; // remember that the game screen now needs to be drawn again
				}

				// sleep until the next draw update:
				AccurateSleep(tpf+frameTime-GetClock());

				// draw the game as necessary:
				thisTime = GetClock();
				if((int)(thisTime-frameTime) > tpf) { // executes at most once per loop
					if(needUpdate || forceUpdateAll) {
						if(!forceDrawAll && (needUpdate || forceUpdateAll == 1) && GameX.CanDraw())
							GameDraw(); // call draw() function, whose contents draw the current game state
						GameX.ViewUpdate();
						frame++; // increment counter so we can know how many fps we're really getting
						GameX.win_alternator = !GameX.win_alternator;
						frameTime = min(frameTime+tpf, thisTime);
						needUpdate = false;
					}
				} 
			} else { // if RUN_UNMANAGED is on
				GameX.UpdateInput();
				GameMain();
				cycle++;

				if(GameX.CanDraw()) GameDraw();
				GameX.ViewUpdate(); // (note: will lock game speed to refresh rate unless VIDEO_NOREFRESHSYNC is on too)
				frame++;
			}

			DWORD thetime = GetClock() - frameClock;// calculate the frame and cycle rate every second
			if(thetime >= clocksPerSec) {			// so the game doesn't need to calculate it itself
				GameX.f_rate = frame;				// and can instead call GameX.GetCurrentFrameRate().
				GameX.c_rate = cycle;
				GameX.p_usage = GameX.sleep_record*(CLOCKS_PER_SEC/10)/clocksPerSec; // calculate idle time percent (percent of power conserved) -- have to be careful of overflow resulting in negative or integer divide resulting in 0, so order of operations is important here
				if(GameX.p_usage < 0 || GameX.p_usage > 99) GameX.p_usage = 99;
				frame = cycle = GameX.sleep_record = 0;
				frameClock = GetClock();
				frameTime = frameClock; // prevent long-term frame loss compensation

// THIS SHOULD GO INTO A FUNCTION CALLED GameX.DrawStatus()
/*	
#ifdef _DEBUG
	if(!GameX.IsFullScreen()) {
		char str [256];
		if(GameX.CanUseMouse())
			sprintf(str,"%s - %d/%d fps - %d/%d cps - %d/%d idle - (%d,%d) mouse",GameX.GetName(),GameX.GetCurrentFrameRate(),GameX.GetTargetFrameRate(),GameX.GetCurrentCycleRate(),GameX.GetTargetCycleRate(),GameX.GetCurrentIdlePercent(),100,GameX.GetMouseX(),GameX.GetMouseY());
		else
			sprintf(str,"%s - %d/%d fps - %d/%d cps - %d/%d idle",GameX.GetName(),GameX.GetCurrentFrameRate(),GameX.GetTargetFrameRate(),GameX.GetCurrentCycleRate(),GameX.GetTargetCycleRate(),GameX.GetCurrentIdlePercent(),100);
		SetWindowText(GameX.GetWindow(),str);
	}
#endif
	*/
				if (fps_too_low <= 12) {
					if (GameX.GetCurrentFrameRate() < GameX.GetTargetFrameRate()/3) {
						fps_too_low++;
						if(fps_too_low > 12)
							GameX.QualityCheck();
					} else
						fps_too_low = 0;
				}
			}
		}
	}
	GameX.CleanUp() ;
	return msg.wParam;
}


// WindowsDX Member Functions

// The game can call this to exit
void WindowsDX::Quit (void)
{
	if(!(win_request_no_confirm_quit || win_page_flipping || IsKeyDown(DIK_LSHIFT))) {
		GameDraw();
		GameX.SpecialViewFade(160);
		GameX.ViewUpdate();
	}

	SendMessage(win_hwnd,WM_CLOSE,0,0); // SendMessage instead of PostMessage, for immediate action
}

// This function must be called by the game's initialize() function
// Parameters: Name, Flags, ViewWidth, ViewHeight, [optional:]	Speed, Smoothness, SkipLimit 
bool WindowsDX::Initialize (char * name, int options, int xres, int yres, int cps, int fps, int max_cpf)
{
	strcpy (game_name, name);

#ifdef _DEBUG
	if(options & VIDEO_WINDOWED && options & VIDEO_FULLSCREEN) {
		MessageBox (win_hwnd, "WARNING: VIDEO_WINDOWED and VIDEO_FULLSCREEN are mutually exclusive flags.\nLeave out both flags if you want the user to choose.", "GameX Warning",MB_OK|MB_ICONWARNING) ;
		options ^= VIDEO_WINDOWED;
		options ^= VIDEO_FULLSCREEN;
	}
#endif

	if(options & VIDEO_WINDOWED)
		win_request_full = 0;
	else if(options & VIDEO_FULLSCREEN)
		win_request_full = 1;
	else
		win_request_full = 2;

	if(options & VIDEO_ALLOWPAGEFLIP) {
		win_request_no_flip = false;
#ifdef _DEBUG
		if(win_request_full == 0) {
			MessageBox (win_hwnd, "WARNING: VIDEO_ALLOWPAGEFLIP has no effect in windowed mode.", "GameX Warning",MB_OK|MB_ICONWARNING) ;
		}
#endif
	} else
		win_request_no_flip = true;

	if(options & VIDEO_LETTERBOX) {
		win_request_no_stretch = true;
#ifdef _DEBUG
		if(win_request_full == 0) {
			MessageBox (win_hwnd, "WARNING: VIDEO_LETTERBOX has no effect in windowed mode.", "GameX Warning",MB_OK|MB_ICONWARNING) ;
		}
#endif
	} else
		win_request_no_stretch = false;

	if(win_request_full == 2) {
		int style = MB_YESNOCANCEL|MB_ICONQUESTION;

		#ifdef _DEBUG
			style |= MB_DEFBUTTON2;
		#endif

		int choice = MessageBox (NULL, "Do you want to play this game in fullscreen mode?", game_caption, style) ;
		if(choice == IDYES)
			win_request_full = 1;
		else if(choice == IDNO)
			win_request_full = 0;
		else if(choice == IDCANCEL)
			ExitProcess(0);
	}

	if(options & VIDEO_NOREFRESHSYNC)
		win_request_vsync = 0;
	else
		win_request_vsync = 1;

	if(options & VIDEO_LOWQUALITY) {
#ifdef _DEBUG
		MessageBox (win_hwnd, "WARNING: Entering low-quality mode because VIDEO_LOWQUALITY was turned on.\nThis is just a reminder: don't forget to remove that flag.", "GameX Notice",MB_OK|MB_ICONINFORMATION) ;
#endif
		win_lowquality = true;
	}
	else
		win_lowquality = false;

	if(options & VIDEO_SOFTWAREONLY) {
#ifdef _DEBUG
		MessageBox (win_hwnd, "WARNING: Entering software-only mode because VIDEO_SOFTWAREONLY was turned on.\nThis is just a reminder: don't forget to remove that flag.", "GameX Notice",MB_OK|MB_ICONINFORMATION) ;
#endif
		win_request_software_only = true;
	}
	else
		win_request_software_only = false;

	if(options & VIDEO_NORESIZE) {
		win_request_can_resize = false;
#ifdef _DEBUG
		if(win_request_full == 1) {
			MessageBox (win_hwnd, "WARNING: VIDEO_NORESIZE has no effect if VIDEO_FULLSCREEN is set.", "GameX Warning",MB_OK|MB_ICONWARNING) ;
		}
#endif
	} else {
		win_request_can_resize = true;
	}


	win_request_bpp = 0;
	if(options & VIDEO_8BIT) {
		win_request_bpp = 8;
		MessageBox (win_hwnd, "WARNING: 8-bit indexed color mode is not yet supported by GameX.\nPlease initialize GameX with the VIDEO_16BIT flag instead.", "GameX Warning",MB_OK|MB_ICONWARNING) ; win_request_bpp = 16;
	} else if(options & VIDEO_16BIT) {
		win_request_bpp = 16; // 16 bit color is fully supported by GameX
	} else if(options & VIDEO_24BIT) {
		win_request_bpp = 24;
		MessageBox (win_hwnd, "WARNING: 24-bit RGB color mode has not yet been tested.", "GameX Warning",MB_OK|MB_ICONWARNING) ;
	} else if(options & VIDEO_32BIT) {
		win_request_bpp = 32; // 32 bit color is fully supported by GameX -- may be a little slow though
	}

#ifdef _DEBUG
	if(1 < !!(options & VIDEO_8BIT)+!!(options & VIDEO_16BIT)+!!(options & VIDEO_24BIT)+!!(options & VIDEO_32BIT)) {
		MessageBox (win_hwnd, "ERROR: You cannot select more than one video mode at once.\nPlease initialize GameX with only one VIDEO_??BIT flag.", "GameX Error",MB_OK|MB_ICONSTOP) ;
	} else if(win_request_bpp == -1) {
		MessageBox (win_hwnd, "WARNING: You should select a video mode by initializing GameX with a video color depth flag.\nFor example,\nGameX.Initialize(VIDEO_16BIT,640,480);\nor\nGameX.Initialize(VIDEO_32BIT|VIDEO_FULLSCREEN,800,600);", "GameX Warning",MB_OK|MB_ICONWARNING) ;
	}
#endif

	win_request_xsize = xres;
	win_request_ysize = yres;

	win_cps = cps; // see WinMain() for usage

	if(win_cps < 1 || win_cps > 1000) {
#ifdef _DEBUG
		char disp [500] ;
		sprintf (disp, "WARNING: You have specified %d CPS, which is invalid. The CPS must be at least 1, at most 1000, and normally should be between 30 and 150.", win_cps) ;
		MessageBox (win_hwnd, disp, "GameX Warning",MB_OK|MB_ICONWARNING) ;
#endif
		win_cps = 5;
	}

	if(fps < 1) win_fps = -1;
	else win_fps = fps;
	if(max_cpf < 2) win_max_cpf = win_cps; // default to skip no more than 1 second of frames per instant
	else win_max_cpf = max_cpf; // otherwise, use user-set limit -- can limit game speed if set too low

#ifdef _DEBUG
	if(win_fps > 150 && win_fps != win_cps) {
		char disp [500] ;
		sprintf (disp, "WARNING: You have specified %d FPS, which is unreasonably high.\nIt puts unnecessary strain on the system to draw extra frames that you won't notice anyway.\n60 FPS is a good, average frame rate, and 120 FPS is the absolute maximum recommended.", win_fps) ;
		MessageBox (win_hwnd, disp, "GameX Warning",MB_OK|MB_ICONWARNING) ;
	}
#endif


	if(options & AUDIO_DISABLE)
		win_request_no_sound = true;
	else
		win_request_no_sound = false;

	if(options & AUDIO_NOSTEREO) {
		win_request_no_stereo = true;
#ifdef _DEBUG
		if(win_request_no_sound)
			MessageBox (win_hwnd, "WARNING: AUDIO_NOSTEREO has no effect if AUDIO_DISABLE is set.", "GameX Warning",MB_OK|MB_ICONWARNING) ;
#endif
	} else
		win_request_no_stereo = false;

	if(options & AUDIO_REVERSESPEAKERS) {
		win_request_reverse_speakers = true;
#ifdef _DEBUG
		if(win_request_no_sound)
			MessageBox (win_hwnd, "WARNING: AUDIO_REVERSESPEAKERS has no effect if AUDIO_DISABLE is set.", "GameX Warning",MB_OK|MB_ICONWARNING) ;
		else if(win_request_no_stereo)
			MessageBox (win_hwnd, "WARNING: AUDIO_REVERSESPEAKERS has no effect if AUDIO_NOSTEREO is set.", "GameX Warning",MB_OK|MB_ICONWARNING) ;
#endif
	} else
		win_request_reverse_speakers = false;



	if(options & RUN_BACKGROUND)
		win_request_no_auto_pause = true;
	else
		win_request_no_auto_pause = false;

	if(options & RUN_KEEPCURSOR)
		win_request_no_hide_cursor = true;
	else
		win_request_no_hide_cursor = false;

	if(options & RUN_NOMOUSEINPUT)
		win_request_no_mouse_input = true;
	else
		win_request_no_mouse_input = false;

	if(options & RUN_ASSUMEVRAMSAFE)
		win_request_delete_mdata = true;
	else
		win_request_delete_mdata = false;


	if(options & RUN_NOCONFIRMQUIT)
		win_request_no_confirm_quit = true;
	else
		win_request_no_confirm_quit = false;

	if(options & RUN_UNMANAGED) {
		win_request_no_manage_speed = true;
#ifdef _DEBUG
		if(win_cps != 60 || win_fps != min(win_cps,90)) {
			char disp [500] ;
			sprintf (disp, "WARNING: You have specified RUN_UNMANAGED, but you have also specified a game speed, %d CPS, which will be ignored.\nNote that RUN_UNMANAGED is highly unrecommended unless your game has it's own custom speed management system\n(which isn't a good idea for the vast majority of games).", win_cps) ;
			MessageBox (win_hwnd, disp, "GameX Warning",MB_OK|MB_ICONWARNING) ;
		}
#endif
		win_cps = win_fps = 0;
	}
	else
		win_request_no_manage_speed = false;
 
	bool success = true;

	debug.Open ("debug.txt") ;
#ifdef WINDX_DEBUG
	debug.Output ("WindowsDX: Initialization started.\n", "") ; 
#endif

	if (GameX.InitWindows()==false) success = false;
	if(success)
	if (GameX.InitDirectX()==false) success = false;
	
	if(!success) {
		ReportProblem(GameX.GetStatus());
#ifdef WINDX_DEBUG
	debug.Output ("WindowsDX: Initialization failed!\n", "") ;
#endif
		return false;
	}
#ifdef WINDX_DEBUG
	debug.Output ("WindowsDX: Initialization finished OK.\n", "") ;
#endif
//			GameX.SetMouseStop(1);
	
//	GameX.DrawClear() ; GameX.ViewUpdate() ;

	return true;
}

bool WindowsDX::InitWindows (void)
{	
	if (WindowInitClass()==false)		return false;
	if (WindowInit()==false)			return false;
	if (WindowDraw()==false)			return false;
	return true;
}

bool WindowsDX::InitDirectX (void)
{	
	if(FAILED(CoInitialize(NULL))) {
		win_status = GAMEX_DDSTART_FAILED;
		return false;
	}

	bool success = true;

	if(DirectDrawInit()==false) {
		success = false;
		ReportProblem();
	}
	
	if(Direct3DInit()==false) {
		success = false;
		ReportProblem();
	}
	
	if(DirectInputInit()==false) {
		success = false;
		ReportProblem();
	}
	
	if(DirectSoundInit()==false) {
		success = false;
		ReportProblem();
	}
	
	if(DirectMusicInit()==false) {
		success = false;
		ReportProblem();
	}
	
	VSyncInit(); // should be after other initializations
	win_keys = 0;
	return success;
}

void WindowsDX::WindowInitInstance (HINSTANCE hThisInst, int nWinMode)
{
	win_hinst = hThisInst;
	win_hmode = nWinMode;
}
		
bool WindowsDX::WindowInitClass (void)
{
	WNDCLASS wcl;

	wcl.hCursor 		= LoadCursor(NULL, IDC_ARROW) ;
	wcl.hIcon			= LoadIcon(NULL, IDI_APPLICATION);
	wcl.lpszMenuName	= "Menu" ;
	wcl.lpszClassName	= "GameX" ;
	wcl.hInstance		= win_hinst;
	wcl.style			= CS_VREDRAW | CS_HREDRAW;
	if(!win_request_no_mouse_input) wcl.style |= CS_DBLCLKS; // enable double-click recognition in the game window
	wcl.lpfnWndProc 	= (WNDPROC) WindowFunc;
	wcl.cbClsExtra		= 0;
	wcl.cbWndExtra		= 0;
	if(win_fullscreen)
		wcl.hbrBackground	= (HBRUSH) GetStockObject(BLACK_BRUSH) ;
	else
		wcl.hbrBackground	= (HBRUSH) GetStockObject(HOLLOW_BRUSH) ; // avoid flickering since we'll handle background clearing ourself

	if (!RegisterClass(&wcl)) { 
		win_status = GAMEX_INITCLASS_FAILED;
		return false;
	}
	win_status = GAMEX_INITCLASS_OK;
	return true;
}

bool WindowsDX::WindowInit (void)
{
	DWORD style;

	// Initialize View Format (Bits-Per-Pixel)
	if(win_request_bpp != 0)
		win_screenbpp = win_request_bpp; // set bpp to requested setting
	else {
		HDC hdc = GetDC (win_hwnd) ; // current setting requested
		win_screenbpp = GetDeviceCaps (hdc, BITSPIXEL) ;
		ReleaseDC (win_hwnd, hdc) ;
	}

	// Initialize View Size & FullScreen Mode
	win_screenx = GetSystemMetrics(SM_CXSCREEN) ;
	win_screeny = GetSystemMetrics(SM_CYSCREEN) ;
	win_fullscreen = win_request_full ? true : false;
	win_dialog_up = win_no_acc_err = false;

	if(win_request_xsize == 0) win_request_xsize = win_screenx;
	if(win_request_ysize == 0) win_request_ysize = win_screeny;
	
	if (win_request_xsize > 2048 || win_request_ysize >= 2048 // (the > and >= is not a typo)
	||	win_request_xsize < 64	 || win_request_ysize < 64) {
		win_status = GAMEX_REQSIZE_FAILED;
		return false;
	}

	// Initialize View Width & Height
	if (win_fullscreen==TRUE)	{
		style = WS_POPUP;	
		win_width = win_request_xsize;
		win_height = win_request_ysize;
	} else {
		style = WS_OVERLAPPED | WS_SYSMENU | WS_MINIMIZEBOX;
		if(win_request_can_resize) style |= WS_SIZEBOX | WS_MAXIMIZEBOX;
		win_width = win_request_xsize;		
		win_height = win_request_ysize;
	}

	win_realx = win_screenx;
	win_realy = win_screeny;
	int window_x = min(win_width, win_fullscreen ? win_screenx : win_realx-4); // temporary variable for actual width of window
	int window_y = min(win_height, win_fullscreen ? win_screeny : win_realy-32);

	// Initialize Drawing Viewport
	win_viewx1 = 0;
	win_viewy1 = 0;
	win_viewx2 = win_width;
	win_viewy2 = win_height;
		
	// Create window adjusting for desired client area size
	// NOTE: For some reason (windows crap??) the AdjustWindowRectEx function
	// is not working.
	RECT	rect;
	int 	sx, sy;
	HWND	temp_hwnd;

	// Create temporary window to determine extended window size
	temp_hwnd = CreateWindowEx(WS_EX_APPWINDOW, "GameX", game_caption, style, 0, 0, window_x, window_y, 0, 0, win_hinst, 0) ;
	if (temp_hwnd==NULL) {
		win_status = GAMEX_INITWINTEMP_FAILED;
		return false;		
	}
	win_status = GAMEX_INITWINTEMP_OK;
	GetClientRect (temp_hwnd, &rect) ;

	// set the rect to be the proper size and center it onscreen while we're at it
	sx = (window_x*2) - rect.right;
	sy = (window_y*2) - rect.bottom;
	if (win_fullscreen==TRUE) {
		rect.left = 0; rect.top = 0;
	} else {
		rect.left = (win_screenx/2) - (sx/2) ;
		rect.top = (win_screeny/2) - (sy/2) ;
	}	
	rect.right = rect.left + sx;		
	rect.bottom = rect.top + sy;

	// Create actual window (with correct size) which will be used

	win_hwnd = CreateWindowEx(
					WS_EX_APPWINDOW, 
					"GameX", 
					game_caption, 
					style, 
					rect.left, rect.top, 
					rect.right-rect.left, rect.bottom-rect.top,
					0, 
					0, 
					win_hinst, 
					0) ;
	if (win_hwnd==NULL) {
		win_status = GAMEX_INITWINREAL_FAILED;		
		return false;
	}

	// Double check to make sure we have correct client rect size	
	GetClientRect (win_hwnd, &rect) ;	

	win_cursor_hidden = false;

	win_status = GAMEX_INITWINREAL_OK;
	return true;
}

bool WindowsDX::WindowDraw (void)
{
	ShowWindow (win_hwnd, win_hmode) ;
	if (UpdateWindow (win_hwnd)==false) {
		win_status = GAMEX_SHOWWIN_FAILED;		
		return false;
	}
	FlashWindow(win_hwnd,0);
	BringWindowToTop(win_hwnd);
	win_status = GAMEX_SHOWWIN_OK;
	return true;
}

bool WindowsDX::DirectDrawInit (void)
{
	dd_last_error = DD_OK;
	dd_was_lost = false;

	HRESULT status = DirectDrawCreateEx(NULL, (void **) &dd_main, IID_IDirectDraw7, NULL);

	if(status != DD_OK) {
		DirectDrawError(status);
		win_status = GAMEX_DDSTART_FAILED;
		return false;
	}
	win_status = GAMEX_DDSTART_OK;


DDCAPS caps;
ZeroMemory(&caps,sizeof(caps));
caps.dwSize = sizeof(caps);
if(SUCCEEDED(dd_main->GetCaps(&caps, NULL))) {

#ifdef WINDX_DEBUG
	char stringy [64];
	sprintf(stringy,"WinDX: Video memory available: %d megabytes.\n",caps.dwVidMemFree/1024/1024);
	debug.Output(stringy,"");
#endif
	int memfree = caps.dwVidMemFree;
	if(memfree < 1500000 && !win_lowquality && !win_request_software_only) { // if there are less than about 1.5 megabytes of VRAM available, assume there won't be much if any video acceleration
		if(win_hmode)
			ShowWindow (win_hwnd, 0) ;
		int choice = MessageBox (win_hwnd, "Lack of support for video acceleration detected!\nThis can make the game unplayably choppy.\nWould you like to temporarily lower the graphical quality to compensate?", "GameX Notice", MB_YESNOCANCEL|MB_DEFBUTTON1|MB_ICONQUESTION) ;
#ifndef WINDX_DEBUG
		debug.Output("Warning: Not enough video memory (VRAM) available for video acceleration.","");
#endif
		if(choice == IDYES) {
			win_lowquality = true;
			win_fps = min(win_fps,30);
		} else if(choice == IDCANCEL) {
			ExitProcess(0); // bail if they hit cancel
		}

		if(win_hmode)
			ShowWindow (win_hwnd, 1) ;
	} else {
		int memtotal = caps.dwVidMemTotal;
		if(memtotal < 22000000 && !win_lowquality && !win_request_software_only) { // if there are less than about 22 megabytes of VRAM total, assume the video card is probably too old (yes, the memory number only gives a very, very rough estimate of how good the card is)
		int choice = MessageBox (win_hwnd, "Warning: Your video card appears to be outdated.\nConsider updating it if this game plays slowly or looks odd.", "GameX Notice", MB_ICONWARNING) ;
#ifndef WINDX_DEBUG
		debug.Output("Warning: Video card may be outdated...","");
#endif
		}
	}
}
 
	int style;
	if(win_fullscreen==TRUE) {
		style = DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE | DDSCL_ALLOWMODEX | DDSCL_ALLOWREBOOT;
	} else {
		style = DDSCL_NORMAL;
	}
	status = dd_main->SetCooperativeLevel (win_hwnd, style);
	if(status != DD_OK) {
		win_status = GAMEX_DDSETLEVEL_FAILED;
		return false;
	}
	win_status = GAMEX_DDSETLEVEL_OK;

	// now set the display mode:
	if (win_fullscreen==TRUE) {

		if(win_request_no_flip)
			win_page_flipping = 0; // don't allow page flipping if VIDEO_ALLOWPAGEFLIP is off
		else 
			win_page_flipping = 1; // otherwise set it on, may still become false later though

		win_realx = win_realy = 0;

		// enumerate available display modes and decide on one of them that best matches win_width, win_height, and win_screenbpp
		status = dd_main->EnumDisplayModes(0, NULL, NULL, (LPDDENUMMODESCALLBACK2) &DisplayEnumCallback);

		// this was a really annoying bug to track down:
		// fullscreen mode would always fail, but only in Release configuration, not in Debug configuration
		// the reason was that the "Assume no aliasing" optimization was set,
		// causing the changes that DisplayEnumCallback() makes to win_realx, win_realy to not take effect.
		// ideally the "assume no aliasing" optimization should not be turned on with GameX,
		// but another thing that fixes the problem is calling a function before checking win_realx, win_realy
		// so the following line is mainly to prevent this code from failing in the case that "assume no aliasing" is set:
		DirectDrawError(status);

		if(!win_realx || !win_realy) { // if DisplayEnumCallback didn't set a screen size
			
			// maybe the requested bit depth isn't supported, try for something higher
			while(win_screenbpp <= 32 && (!win_realx || !win_realy || !SupportsDepth(win_screenbpp))) {
				win_screenbpp += 8;
				status = dd_main->EnumDisplayModes(0, NULL, NULL, (LPDDENUMMODESCALLBACK2) &DisplayEnumCallback);
				DirectDrawError(status);
			}
			// if that didn't work, try for something lower
			while(win_screenbpp > 0 && (!win_realx || !win_realy || !SupportsDepth(win_screenbpp))) {
				win_screenbpp -= 4;
				status = dd_main->EnumDisplayModes(0, NULL, NULL, (LPDDENUMMODESCALLBACK2) &DisplayEnumCallback);
				DirectDrawError(status);
			}
			if(!win_realx || !win_realy) {
				win_status = GAMEX_SETMODE_FAILED;
				return false;
			}
		}
		
		if(win_page_flipping && win_realx != win_width || win_realy != win_height)
			win_page_flipping = false; // can't (or wouldn't be worth it to) page flip if we're stretching or only drawing to part of the screen

		status = dd_main->SetDisplayMode (win_realx, win_realy, win_screenbpp,0,0);

		win_screenx = GetSystemMetrics(SM_CXSCREEN) ;
		win_screeny = GetSystemMetrics(SM_CYSCREEN) ;

		if(status != DD_OK) {
			GameX.DirectDrawError(status);
			win_status = GAMEX_SETMODE_FAILED;
			return false;
		}
		win_status = GAMEX_SETMODE_OK;

	} else {
		win_page_flipping = false; // can't page flip in windowed mode

		if(win_screenbpp != GetDeviceCaps (GetDC (win_hwnd), BITSPIXEL)) { // if we aren't in desired bit mode
			if(SupportsDepth(win_screenbpp))
				status = dd_main->SetDisplayMode (win_screenx, win_screeny, win_screenbpp,0,0) ; // change to that many bpp
			else
				status = DDERR_GENERIC;
			if(status != DD_OK) {
				// maybe the requested bit depth isn't supported, try for something higher
				while(win_screenbpp <= 32 && (status != DD_OK || !SupportsDepth(win_screenbpp))) {
					win_screenbpp += 8;
					status = dd_main->SetDisplayMode (win_screenx, win_screeny, win_screenbpp,0,0) ;
				}
				// if that didn't work, try for something lower
				while(win_screenbpp > 0 && (status != DD_OK || !SupportsDepth(win_screenbpp))) {
					win_screenbpp -= 4;
					status = dd_main->SetDisplayMode (win_screenx, win_screeny, win_screenbpp,0,0) ;
				}
				if(status != DD_OK) {
					win_status = GAMEX_SETMODE_FAILED;
					return false;
				}
			}
			UpdateWindow (win_hwnd); // redraw the window borders after switching or they'll all be black for a while
		}
		win_realx = min(win_width, win_realx-4); 
		win_realy = min(win_height, win_realy-32);

		win_status = GAMEX_SETMODE_OK;
	}

	DWORD monitor_frequency = 0;
	dd_main->GetMonitorFrequency(&monitor_frequency);
	if(monitor_frequency == 0) monitor_frequency = 60; // just in case
	win_screenrefresh = monitor_frequency;
	win_vsync_enabled = false; // will enable later if we can

#ifdef WINDX_DEBUG
		char str [128];
		sprintf(str,"WinDX: Chosen display mode: %dx%d %d-bit %dHz %sexclusive.\n", win_screenx, win_screeny, win_screenbpp, win_screenrefresh,win_fullscreen?"":"non");
		debug.Output(str,"");
#endif

	win_xpos = 0;
	win_ypos = 0;

	if(DirectDrawSurfInit() == false)
		return false;

	return true;
}

static HRESULT DisplayEnumCallback(LPDDSURFACEDESC2 lpDDModeDesc, LPVOID lpContext)
{
/// char str [128];
/// sprintf(str,"WinDX: Inspecting available display mode: depth: %d   width: %d   height: %d   refresh: %d\n",lpDDModeDesc->ddpfPixelFormat.dwRGBBitCount, lpDDModeDesc->dwWidth, lpDDModeDesc->dwHeight, lpDDModeDesc->dwRefreshRate);
/// debug.Output(str,"");

	if(GameX.SupportsDepth(lpDDModeDesc->ddpfPixelFormat.dwRGBBitCount)
	&& lpDDModeDesc->ddpfPixelFormat.dwRGBBitCount == GameX.win_screenbpp) { // if the bit depth matches
		if(lpDDModeDesc->dwWidth == GameX.win_width && lpDDModeDesc->dwHeight == GameX.win_height) { // if the resolution matches
			GameX.win_realx = GameX.win_width;
			GameX.win_realy = GameX.win_height;
			return DDENUMRET_CANCEL; // exact match, end enumeration with requested screen width and height
		}

		if(lpDDModeDesc->dwWidth >= GameX.win_width && lpDDModeDesc->dwHeight >= GameX.win_height // if (this bigger than requested
		&& lpDDModeDesc->dwWidth <= GameX.win_realx && lpDDModeDesc->dwHeight <= GameX.win_realy // and this smaller than so far)
		|| (GameX.win_realx 	 <= GameX.win_width || GameX.win_realy		  <= GameX.win_height) // or (so far smaller than requested
		&& lpDDModeDesc->dwWidth >= GameX.win_realx && lpDDModeDesc->dwHeight >= GameX.win_realy) { // and this smaller than so far)
			GameX.win_realx = lpDDModeDesc->dwWidth;
			GameX.win_realy = lpDDModeDesc->dwHeight;
			return DDENUMRET_OK; // best match yet but not exact, remember this resolution and continue enumeration
		}
	}

	// there doesn't seem to be a signal to see if the callback is done, and it just keeps calling
	// with crazy surface descriptions and then crashes... as far as I can tell, these crazy surface descriptions
	// all have a weird bit depth like 13 but that can't be the intended way of knowing the enumeration is over...
	// (shouldn't the enumeration stop on its own when there's nothing more to enumerate?)
	// well, for now we'll end enumeration if the bit depth is 13 or the resolution is insane, but hopefully this is temporary:

	if(lpDDModeDesc->ddpfPixelFormat.dwRGBBitCount == 13 || lpDDModeDesc->dwWidth > 5000 || lpDDModeDesc->dwHeight > 5000
	|| lpDDModeDesc->ddpfPixelFormat.dwRGBBitCount <= 0  || lpDDModeDesc->dwWidth <= 0	 || lpDDModeDesc->dwHeight <= 0)
		return DDENUMRET_CANCEL; // end enumeration
	else
		return DDENUMRET_OK; // continue enumeration
}

bool WindowsDX::DirectDrawSurfInit (void)
{
	// DirectX 6.0		
	HRESULT status;
	DDSURFACEDESC2 ddsd;	
	POINT pt;

	if (dd_back_buf)	dd_back_buf->Release(), 	dd_back_buf = NULL;
	if (dd_front_buf)	dd_front_buf->Release(),	dd_front_buf = NULL;	

	ZeroMemory (&ddsd, sizeof (ddsd)) ;
	ddsd.dwSize = sizeof (ddsd) ;
 
	LPDIRECTDRAWPALETTE lpddpal = NULL; // palette interface
	if(win_screenbpp == 8) { // 8-bit mode is indexed, so do palette stuff here
		PALETTEENTRY palette[256]; // palette storage
		int shades [16][3] = 
		{{128,128,255},{  0,128,255},{	0,	0,255},{128,  0,255},
		 {255,	0,255},{255,  0,128},{255,	0,	0},{255,128,  0},
		 {255,255,	0},{128,255,  0},{	0,255,	0},{  0,255,128},
		 {	0,255,255},{128,255,128},{255,255,255},{128,255,128},};
		for (int index=0; index < 256; index++) {
			int sector = index/16;
			int pad = (index%16) - 8;
			if(sector == 15) pad = -pad+1;
			palette[index].peRed = min(255,max(0,shades[sector][0] + pad * 32));
			palette[index].peGreen = min(255,max(0,shades[sector][1] + pad * 32));
			palette[index].peBlue = min(255,max(0,shades[sector][2] + pad * 32));
			palette[index].peFlags = PC_NOCOLLAPSE;
		}
		palette[0].peRed = palette[0].peGreen = palette[0].peBlue = 0;
		palette[255].peRed = palette[255].peGreen = palette[255].peBlue = 255;
		if(!win_fullscreen) {
			for (index = 0; index < 10 ; index++)
			{
				palette[index].peFlags = PC_EXPLICIT;
				palette[index+246].peFlags = PC_EXPLICIT;
			}
		}

		if(FAILED(dd_main->CreatePalette(DDPCAPS_8BIT | DDPCAPS_INITIALIZE, palette, &lpddpal, NULL)))
			debug.Output("Palette creation error.\n","");
	}

	if (win_page_flipping) {
		// Set front surface description
		ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		ddsd.dwBackBufferCount = 1;
/// 	ddsd.dwBackBufferCount = 2; // triple buffering? -- would need to enumerate surfaces on init
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP | DDSCAPS_3DDEVICE; // (the 3d device is for the dd_back_buf)
//		ddsd.ddsCaps.dwCaps2 = DDCAPS2_FLIPNOVSYNC; // apparently not necessary

		// Try to create primary surface 
		status = dd_main->CreateSurface (&ddsd, &dd_front_buf, NULL) ;

		// No good, return error
		if (status!=DD_OK) {
			GameX.DirectDrawError(status);
			win_status = GAMEX_FRONTSURFACE_FAILED; 		
			return false;
		}			
		if(lpddpal != NULL) {
			if (FAILED(dd_front_buf->SetPalette(lpddpal))) {
				return false;
			}
		}
		
		win_status = GAMEX_FRONTSURFACE_OK;
		// Set back surface description

		// DirectX 6.0		
		DDSCAPS2 caps;	
		ZeroMemory(&caps,sizeof(caps));
		caps.dwCaps = DDSCAPS_BACKBUFFER;
		if(win_page_flipping)
			status = dd_front_buf->GetAttachedSurface (&caps, &dd_back_buf) ;

		if (status!=DD_OK) {
			GameX.DirectDrawError(status);
			win_status = GAMEX_BACKSURFACE_FAILED;
			return false;
		}
		if(lpddpal != NULL) {
			if (FAILED(dd_back_buf->SetPalette(lpddpal))) {
				return false;
			}
		}
		win_status = GAMEX_BACKSURFACE_OK;
	} else {

		// Set front surface description
		ddsd.dwFlags = DDSD_CAPS;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

		// Try to create primary surface 
		status = dd_main->CreateSurface (&ddsd, &dd_front_buf, NULL);
		if (status!=DD_OK) {			
			GameX.DirectDrawError(status);
			win_status = GAMEX_FRONTSURFACE_FAILED; // No good, return error
			return false;
		}
//			LPDIRECTDRAWPALETTE currpal;
//			dd_front_buf->GetPalette(&currpal);
		if(lpddpal != NULL) {
			if (FAILED(dd_front_buf->SetPalette(lpddpal))) {
				return false;
			}
		}
		
		win_status = GAMEX_FRONTSURFACE_OK;
		// Set back surface description
		ddsd.dwSize = sizeof (ddsd) ;
		ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_3DDEVICE;
		ddsd.dwWidth = win_width;
		ddsd.dwHeight = win_height;

/*		DDPIXELFORMAT ddpf;
		ddpf.dwSize = sizeof(ddpf);
		dd_front_buf->GetPixelFormat(&ddpf);
		ddsd.dwFlags |= DDSD_PIXELFORMAT;
		ddsd.ddpfPixelFormat = ddpf;
*/
/*		if(win_screenbpp == 8) {
			DDPIXELFORMAT ddpf;
			ddpf.dwSize = sizeof(ddpf);
			dd_front_buf->GetPixelFormat(&ddpf);
			ddpf.dwFlags = DDPF_RGB;
			ddpf.dwRGBBitCount	   = 16;
			ddpf.dwRGBAlphaBitMask = 0x0000;
			ddpf.dwRBitMask 	   = 0xF800;
			ddpf.dwGBitMask 	   = 0x07E0;
			ddpf.dwBBitMask 	   = 0x001F;
			ddsd.dwFlags |= DDSD_PIXELFORMAT;
			ddsd.ddpfPixelFormat = ddpf;
		}*/

		// Try to create secondary surface 
		status = dd_main->CreateSurface (&ddsd, &dd_back_buf, NULL);
		if (status!=DD_OK) {
			GameX.DirectDrawError(status);
			win_status = GAMEX_BACKSURFACE_FAILED; // No good, return error
			return false;
		}

		if(lpddpal != NULL) {
			if (FAILED(dd_back_buf->SetPalette(lpddpal))) {
				return false;
			}
		}
		
		win_status = GAMEX_BACKSURFACE_OK;
	}

	// Get Surface Description (pixel format, pitch, etc.)	
	ddsd.dwSize = sizeof(ddsd) ;
	ddsd.dwFlags = DDSD_HEIGHT | DDSD_WIDTH | DDSD_PITCH | DDSD_PIXELFORMAT;
	dd_front_buf->GetSurfaceDesc(&ddsd) ;
	if (ddsd.ddsCaps.dwCaps & DDSCAPS_VIDEOMEMORY) {
#ifdef WINDX_DEBUG 
		debug.Output ("WinDX: Front buffer is in video memory.\n", "") ;
#endif
	} else {
#ifdef WINDX_DEBUG 
		debug.Output ("WinDX: Front buffer is NOT IN VIDEO MEMORY.\n", "") ;
#else
		debug.Output ("Warning: Front buffer is not in video memory.\n", "") ;
#endif
	}
	dd_back_buf->GetSurfaceDesc(&ddsd) ;
	if (ddsd.ddsCaps.dwCaps & DDSCAPS_VIDEOMEMORY) {		
#ifdef WINDX_DEBUG 
		debug.Output ("WinDX: Back buffer is in video memory.\n", "") ;
#endif
	} else {
#ifdef WINDX_DEBUG 
		debug.Output ("WinDX: Back buffer is NOT IN VIDEO MEMORY.\n", "") ;
#else
		debug.Output ("Warning: Back buffer is not in video memory.\n", "") ;
#endif
	}
#ifdef WINDX_DEBUG
	if (ddsd.ddsCaps.dwCaps & DDSCAPS_BACKBUFFER) {
		debug.Output ("WinDX: Page flipping is on.\n", "") ;
	} else {
		debug.Output ("WinDX: Page flipping is OFF.\n", "") ;
	}	
#endif

	// Initialize Drawing Functions, Pixel Masks, Pitch and Width 
	// based on Surface Description 

	win_pitch = ddsd.lPitch;
	win_adjustedpitch = win_pitch * (win_screenbpp/8);

	win_maskr = ddsd.ddpfPixelFormat.dwRBitMask;
	win_maskg = ddsd.ddpfPixelFormat.dwGBitMask;		
	win_maskb = ddsd.ddpfPixelFormat.dwBBitMask;
	
	win_shiftr = GetShift (win_maskr) ; // 11 for 16 bit, 10 for 15 bit, 16 for 24 or 32 bit
	win_shiftg = GetShift (win_maskg) ; // 5 for 16 or 15 bit, 8 for 24 or 32 bit
	win_shiftb = GetShift (win_maskb) ; // 0

	win_maxr = GetMax (win_maskr) ;
	win_maxg = GetMax (win_maskg) ;
	win_maxb = GetMax (win_maskb) ;

	switch (win_screenbpp) {
	case 8: // currently unsupported
//		FuncDrawPixel = &DrawPixel8;
//		FuncDrawPixelAdded = &DrawPixelAdded8;
		FuncDrawPixel = &DrawPixel16; ///
		FuncDrawPixelAdded = &DrawPixelAdded16; ///
		break;
	case 16:
		FuncDrawPixel = &DrawPixel16;
		FuncDrawPixelAdded = &DrawPixelAdded16;
		break;
	case 24: // currently unsupported
//		FuncDrawPixel = &DrawPixel24;
//		FuncDrawPixelAdded = &DrawPixelAdded24;
		FuncDrawPixel = &DrawPixel16; ///
		FuncDrawPixelAdded = &DrawPixelAdded16; ///
		break;
	case 32:
//		FuncDrawPixel = &DrawPixel32;
//		FuncDrawPixelAdded = &DrawPixelAdded32;
		FuncDrawPixel = &DrawPixel16; ///
		FuncDrawPixelAdded = &DrawPixelAdded16; ///
		break;
	default:
//		win_width = ddsd.dwWidth;
		FuncDrawPixel = NULL;
		FuncDrawPixelAdded = NULL;
		break;
	}

	temp_surface = CreateImageSurface(6,6,0,0,true);

	// now set a clipper to the back surface so we don't have to clip bitmap blits ourselves
	// (we still have to clip things like DrawAdded(), just not things like DrawImage() that use Blt())
	{
		RECT boundary[1] = {0,0,win_width,win_height};
		LPDIRECTDRAWCLIPPER clipper;
		LPRGNDATA rgn_data;
		if (FAILED(dd_main->CreateClipper(0,&clipper,NULL))) {
			win_status = GAMEX_CLIPPER_FAILED;
			return false;
		}
		int rectSize = sizeof(RECT);
		int rgnHeaderSize = sizeof(RGNDATAHEADER);
		rgn_data = (LPRGNDATA)malloc(rgnHeaderSize+rectSize);
		memcpy(rgn_data->Buffer, boundary, rectSize);
		rgn_data->rdh.dwSize = rgnHeaderSize;
		rgn_data->rdh.nRgnSize = rectSize;
		rgn_data->rdh.nCount = 1;
		rgn_data->rdh.iType = RDH_RECTANGLES;
		rgn_data->rdh.rcBound = boundary[0];
		if (FAILED(clipper->SetClipList(rgn_data,0)) || FAILED(dd_back_buf->SetClipper(clipper))) {
			free(rgn_data);
			win_status = GAMEX_CLIPPER_FAILED;
			return false;
		}
		free(rgn_data);
	}

//	if (!win_fullscreen) { // if windowed, clip drawing to the window bounds
		
		LPDIRECTDRAWCLIPPER clipper = NULL; // hold the clipper
		dd_main->CreateClipper(0,&clipper,NULL);
		clipper->SetHWnd(0, win_hwnd);
		dd_front_buf->SetClipper(clipper);
//	}

	// Precompute: y-table look up 
	for (int y=0; y<=win_height; y++)
		win_ytable[y] = y*win_adjustedpitch; // the >> is to shift it to current bit pitch
		
	// Precompute: screen size (total number of bytes)
	win_size = win_pitch*win_height;
	
	// Precompute: client rectangle
	
	RECT destrect;

	GetClientRect(win_hwnd, &destrect) ;

	win_realx = destrect.right - destrect.left; // necessary on occasions where the OS gives us a different window size
	win_realy = destrect.bottom - destrect.top; // than we requested, such as when the requested width is very small

	if(win_fullscreen && win_request_no_stretch)
		SetProportional();
	else
		win_xoffset = win_yoffset = 0;

	pt.x = win_xpos;
	pt.y = win_ypos;
	ClientToScreen(win_hwnd, &pt) ;
	OffsetRect(&destrect, pt.x, pt.y) ;
	win_xpos = pt.x;
	win_ypos = pt.y;
	
	win_active = true;	
	win_status = GAMEX_READY;

	DrawClear() ;
	DrawColor(255,255,255); 		// Set text paint color to white
	DrawText(win_width-90,win_height-30,"Loading...") ;
	ViewUpdate() ;

	return true;
}

void WindowsDX::VSyncInit (void)
{
#ifdef WINDX_DEBUG
	char str [128];
#endif

	if(win_request_vsync && win_fps > win_screenrefresh || win_fps == -1)
		win_fps = win_screenrefresh; // set the frame rate to not exceed the refresh rate

	if(win_request_vsync) {
		
		// test scanline speed:
		// sample 100 times at 10 milliseconds of delay each,
		// this adds slightly over one second to the load time of GameX
		// but any less and we risk losing accuracy in the speed averaging
		// which would cause either very choppy gameplay
		// or very unsynchronized refreshes

		DWORD scanline [100]; // array of consecutive scanline locations
		int diff [100]; // array of valid scanline location differences
		int n; // number of valid (positive) differences

		int clocksPerMillisec = GetClocksPerSec()/1000;
		for(int i = 0 ; i < 100 ; i++) {
			long sleeptimer = GetClock();
			while(GetClock() - sleeptimer < clocksPerMillisec) ;

			dd_main->GetScanLine(&scanline[i]);
		}

		n = 0;

		for(i = 0 ; i < 100 ; i++) {
			if(i && scanline[i]) {
				if(scanline[i] > scanline[i-1]) {
					diff[n] = scanline[i] - scanline[i-1];
					n++;
				}
			}
		}

		if(n < 3)
			win_request_vsync = 0;
		else {
			int total = 0;

			for(int i = 0 ; i < n ; i++) {
				total += diff[i];
			}

			win_scanline_speed = total/n;
			if(win_scanline_speed > 0) {
				if(win_screeny/win_scanline_speed*win_screenrefresh > CLOCKS_PER_SEC) {
#ifdef WINDX_DEBUG
					sprintf(str,"WinDX: Refresh synchronization OFF; %d scanlines per millisecond is too high.\n", win_scanline_speed);
#endif
				} else {
					win_vsync_enabled = true;
	//				win_blit_latency = 0;
#ifdef WINDX_DEBUG
					sprintf(str,"WinDX: Refresh synchronization on, %d scanlines per millisecond.\n", win_scanline_speed);
#endif
				}
			}
		}
//		char str [64];
//		sprintf(str,"n: %d   win_scanline_speed: %d\n",n,win_scanline_speed);
//		debug.Output(str,"");
	}

#ifdef WINDX_DEBUG
	{
		long resolution = GetClocksPerSec();
		char str [64];
		sprintf(str,"WinDX: Timer resolution: %d ticks per second.\n");
		debug.Output(str,"");
	}
#endif

#ifdef WINDX_DEBUG
	if(!win_vsync_enabled)
		sprintf(str,"WinDX: Refresh synchronization OFF.\n");
	debug.Output(str,"");
#endif
}

bool WindowsDX::Direct3DInit (void)
{
	if(dd_main==NULL || FAILED(dd_main->QueryInterface(IID_IDirect3D7, (LPVOID *)& d3d_main))) {
		debug.Output ("Direct3D: Device query failed.\n", "") ;
		win_status = GAMEX_D3DSTART_FAILED;
		return false; 
	}

/*	LPDIRECTDRAWPALETTE lpddpal = NULL; // palette interface
	if(FAILED(dd_main->CreatePalette(DDPCAPS_8BIT, NULL, &lpddpal, NULL)))
		debug.Output("Palette creation error in Direct3DInit.\n","");
	if(lpddpal != NULL) {
		if (FAILED(dd_back_buf->SetPalette(lpddpal))) {
			int p = 5;
		}
	}

	LPDIRECTDRAWPALETTE currpal;
	dd_back_buf->GetPalette(&currpal);*/
	// make sure we assign a 3D device to the back buffer
	// or it probably won't get 3D hardware acceleration
	if(CreateDeviceForSurface(d3d_device, dd_back_buf) == false) {
		win_status = GAMEX_D3DSTART_FAILED;
		return false;
	}

	// set some render and texture state values, these may be the defaults, but just in case:

	d3d_device->SetTextureStageState(0,D3DTSS_ADDRESSU,D3DTADDRESS_CLAMP); // we don't want texture wrapping
	d3d_device->SetTextureStageState(0,D3DTSS_ADDRESSV,D3DTADDRESS_CLAMP); // that doesn't work right on most image sizes anyway

	d3d_device->SetRenderState(D3DRENDERSTATE_LIGHTING, FALSE); // we have no D3D light sources, for now
	d3d_device->SetRenderState(D3DRENDERSTATE_COLORVERTEX, TRUE); // we want to allow per-vertex coloring
	d3d_device->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE); // we want to allow alpha blending

	D3DVIEWPORT7 view;
	view.dwX=0;
	view.dwY=0;
	view.dwWidth=win_width;
	view.dwHeight=win_height; 
	view.dvMinZ=0.0f;
	view.dvMaxZ=1.0f;
	d3d_device->SetViewport(&view);

	return true;
}

bool WindowsDX::CreateDeviceForSurface (LPDIRECT3DDEVICE7 & device, LPDIRECTDRAWSURFACE7 & surface)
{
	if(win_request_software_only || d3d_main->CreateDevice(IID_IDirect3DTnLHalDevice, surface, &device) != D3D_OK) { // the best 3D hardware device
		if(win_request_software_only || d3d_main->CreateDevice(IID_IDirect3DHALDevice, surface, &device) != D3D_OK) { // second-best 3D hardware
			if(d3d_main->CreateDevice(IID_IDirect3DMMXDevice, surface, &device) != D3D_OK) { // best 3D software device
				// failing up until now is normal, although undesirable
				HRESULT status = d3d_main->CreateDevice(IID_IDirect3DRGBDevice, surface, &device); // backup 3D software
				if(status != D3D_OK) {
					// that last try should never fail, something's very wrong if we get here
					DirectDrawError(status);
					Direct3DError(status);
					win_status = GAMEX_D3DSTART_FAILED;
					debug.Output ("Direct3D: Device create failed.\n", "") ;
					return false; // should never get here; even if there's no 3D device, the D3DRGB device is software emulated
				}
			}
		}
	}
	return true;
}
/*
inline int WindowsDX::AssignDeviceToSurface (LPDIRECT3DDEVICE7 & device, LPDIRECTDRAWSURFACE7 & surface)
{
	return ((FAILED(device->SetRenderTarget(surface, 0))) ? 0 : 1);
}
*/
bool WindowsDX::DirectSoundInit (void)
{
	if(win_request_no_sound) return true;

	if(FAILED(DirectSoundCreate (NULL, &ds_main, NULL))) {
		win_status = GAMEX_SOUNDINIT_FAILED;
		ds_main = NULL;
		sound_support = false;
		return false;
	} else
	if(FAILED(IDirectSound_SetCooperativeLevel(ds_main, win_hwnd, DSSCL_NORMAL))) {
		IDirectSound_Release(ds_main) ;
		win_status = GAMEX_SOUNDCOOP_FAILED;
		ds_main = NULL;
		sound_support = false;
		return false;
	}

	master_music_tempo = master_music_volume = 1.0f;

	sound_support = true;
	return true;
}

bool WindowsDX::DirectMusicInit (void)
{
	if(win_request_no_sound) return true;

	// set up DirectMusic and performance/ports:

	if(FAILED(CoCreateInstance(CLSID_DirectMusicPerformance, NULL, CLSCTX_INPROC,
 								IID_IDirectMusicPerformance, (void**)&dm_performance)))
		return(false);

	if(FAILED(dm_performance->Init(NULL, (sound_support ? ds_main : NULL), win_hwnd)))
		return(false);

	if(FAILED(dm_performance->AddPort(NULL)))
		return(false);

	// prepare DirectMusic for loading MIDI files:

	if(FAILED(CoCreateInstance(CLSID_DirectMusicLoader, NULL, CLSCTX_INPROC,
								IID_IDirectMusicLoader, (void**)&dm_loader)))
		return(false);

	dm_musicx_index = 0;

	return true;
}

bool WindowsDX::DirectInputInit (void)
{
	// Create the DirectInput object:

	if(FAILED(DirectInputCreate(win_hinst, DIRECTINPUT_VERSION, &di_main, NULL))) {
//		win_status = GAMEX_CREATEINPUT_FAILED;
		return false;
	}


	// Create and set up the keyboard device:

	if(FAILED(di_main->CreateDevice(GUID_SysKeyboard, &di_keyboard, NULL))) {
		return false;
	}

	if(FAILED(di_keyboard->SetCooperativeLevel(win_hwnd, DISCL_BACKGROUND | DISCL_NONEXCLUSIVE))) {
		return false;
	}

	if(FAILED(di_keyboard->SetDataFormat(&c_dfDIKeyboard))) {
		return false;
	}

	if(FAILED(di_keyboard->Acquire())) {
		return false;
	}

	for(int k = 0 ; k < 255 ; k++) {
		di_keystate[k] = di_keypressed[k] = 0;
	}

	if(!win_request_no_mouse_input) {

		// Create and set up the mouse device:

		if(FAILED(di_main->CreateDevice(GUID_SysMouse, &di_mouse, NULL))) {
			return false;
		}

		if(FAILED(di_mouse->SetCooperativeLevel(win_hwnd, DISCL_BACKGROUND | DISCL_NONEXCLUSIVE))) {
			return false;
		}

		if(FAILED(di_mouse->SetDataFormat(&c_dfDIMouse))) {
			return false;
		}

		if(FAILED(di_mouse->Acquire())) {
			return false;
		}
	}

	UpdateInput(); // so the game can check for initial input inside initialize() right after calling GameX.Initialize()

	return true;
}

bool WindowsDX::UpdateInput (void)
{
	if(FAILED(di_keyboard->GetDeviceState(256, (LPVOID)di_keystate))) {
		return false;
	}

	// fix windows oddity that makes the right shift button register as unpressed
	// if an arrow key is pressed while it's down
	if(!IsKeyDown(DIK_RSHIFT) && (GetAsyncKeyState(VK_RSHIFT) & 0x8000))
		di_keystate[DIK_RSHIFT] |= INPUT_PRESSED;

	// transfer/record unique (non-held) keypresses
	for(int k = 0 ; k < 255 ; k++) {
		if(di_keypressed[k] & INPUT_HELD) {
			if(!di_keystate[k])
				di_keypressed[k] = di_keystate[k];
		} else {
			di_keypressed[k] = di_keystate[k];
		}
	}

	if(!win_request_no_mouse_input) {

		if(FAILED(di_mouse->GetDeviceState(sizeof(DIMOUSESTATE), (LPVOID)&di_mousestate))) {
			return false;
		}

		// transfer/record unique (non-held) mousepresses
		for(int b = 0 ; b < 4 ; b++) {
			if(!di_mousestate.rgbButtons[b])
				mouse_down[b] = 0;
			else if(!mouse_down[b])
///				di_mousestate.rgbButtons[b] = 0; // seems to cause mouse click detection to lag after typing
				;
			else if(mouse_down[b] == 2 && di_mousestate.rgbButtons[b])
				di_mousestate.rgbButtons[b] |= 0x01;
			if(di_mousepressed[b] & INPUT_HELD) {
				if(!di_mousestate.rgbButtons[b])
					di_mousepressed[b] = di_keystate[b];
			} else {
				di_mousepressed[b] = di_mousestate.rgbButtons[b];
			}
		}
		if(di_mousestop) {
			di_mousestate.lX = di_mousestate.lY = di_mousestate.lZ = 0;
			if(win_active)
				di_mousestop--;

			// unregister all clicks at this time (so the game doesn't respond to window-activating clicks)
			for(int b = 0 ; b < 4 ; b++) {
 				mouse_down[b] = 0;
				di_mousepressed[b] = 0;
			}
		}

		POINT mousePoint;
		GetCursorPos(&mousePoint);
		mousePoint.x -= win_xpos+win_xoffset;
		mousePoint.y -= win_ypos+win_yoffset;
		SetMouseMove(mousePoint.x,mousePoint.y);
	}

	return true;
}

bool WindowsDX::AnyKeyPress (void)	
{
	for(int n=0; n<256; n++)
		if(GetKeyPress(n))
			// some keys don't count as "any key", such as volume control keys
			if(n != DIK_CAPITAL && n != DIK_MUTE && n != DIK_VOLUMEDOWN && n != DIK_VOLUMEUP && n != DIK_SYSRQ
			&& n != DIK_POWER && n != DIK_SLEEP && n != DIK_WAKE && n != DIK_NUMLOCK && n != DIK_SCROLL
			&& n != DIK_NEXTTRACK && n != DIK_WEBSEARCH && n != DIK_WEBFAVORITES && n != DIK_WEBREFRESH && n != DIK_WEBSTOP
			&& n != DIK_WEBHOME && n != DIK_WEBFORWARD && n != DIK_WEBBACK && n != DIK_MYCOMPUTER && n != DIK_MAIL
			&& n != DIK_MEDIASELECT && n != DIK_PLAYPAUSE && n != DIK_MEDIASTOP && n != DIK_CALCULATOR)
				return true;
	return false;
}

bool WindowsDX::PeekAnyKeyPress (void)	
{
	for(int n=0; n<256; n++)
		if(PeekKeyPress(n))
			// some keys don't count as "any key", such as volume control keys
			if(n != DIK_CAPITAL && n != DIK_MUTE && n != DIK_VOLUMEDOWN && n != DIK_VOLUMEUP && n != DIK_SYSRQ
			&& n != DIK_POWER && n != DIK_SLEEP && n != DIK_WAKE && n != DIK_NUMLOCK && n != DIK_SCROLL
			&& n != DIK_NEXTTRACK && n != DIK_WEBSEARCH && n != DIK_WEBFAVORITES && n != DIK_WEBREFRESH && n != DIK_WEBSTOP
			&& n != DIK_WEBHOME && n != DIK_WEBFORWARD && n != DIK_WEBBACK && n != DIK_MYCOMPUTER && n != DIK_MAIL
			&& n != DIK_MEDIASELECT && n != DIK_PLAYPAUSE && n != DIK_MEDIASTOP && n != DIK_CALCULATOR)
				return true;
	return false;
}

// gets amount to shift a color component of the given mask
int WindowsDX::GetShift (int mask)
{
	if(mask == 0) return 0;
	int count;
	for (count=0; (mask & 1) == 0; count++)
		mask >>= 1;
	return count;
}

int WindowsDX::GetMax (int mask)
{
	if(mask == 0) return 0;
	int count, max; 
	for (count=0; (mask & 1) == 0; count++)
		mask >>= 1;
	for (max=0; mask > 0; max++)
		mask >>= 1;
	return max;
}

LPDIRECTDRAWSURFACE7 WindowsDX::CreateImageSurface (int width, int height, bool texture, int create_alpha, bool targetable)
{
	LPDIRECTDRAWSURFACE7 surface; // the surface we'll create and return

	DDSURFACEDESC2 ddsd;
	ZeroMemory (&ddsd, sizeof(DDSURFACEDESC2)) ;
	ddsd.dwSize=sizeof(ddsd);

	DDPIXELFORMAT ddpf;
	ZeroMemory (&ddpf, sizeof(DDPIXELFORMAT)) ;
	ddpf.dwSize = sizeof(ddpf);

	ddsd.dwHeight=height;
	ddsd.dwWidth=width;
	ddsd.dwFlags=DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_CKSRCBLT;
	ddsd.ddckCKSrcBlt.dwColorSpaceLowValue = 0; // specify black as transparent if drawn masked
	ddsd.ddckCKSrcBlt.dwColorSpaceHighValue = 0;
	if(texture) {
		ddsd.ddsCaps.dwCaps = DDSCAPS_TEXTURE;
		D3DDEVICEDESC7 ddDesc;
		if(d3d_device != NULL)
			d3d_device->GetCaps(&ddDesc);
		if (ddDesc.deviceGUID == IID_IDirect3DHALDevice || ddDesc.deviceGUID == IID_IDirect3DTnLHalDevice) {
			ddsd.ddsCaps.dwCaps2 = DDSCAPS2_TEXTUREMANAGE;
		}
		else {
			ddsd.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;
		}

		if(targetable) { // && texture
			ddsd.ddsCaps.dwCaps |= DDSCAPS_3DDEVICE; // to allow the surface to request a 3D device
			ddsd.ddsCaps.dwCaps2 |= DDSCAPS2_HINTDYNAMIC; // could speed up images that are frequently drawn to
		} else {
			ddsd.ddsCaps.dwCaps2 |= DDSCAPS2_HINTSTATIC; // could speed up images that are never drawn to
//			ddsd.ddsCaps.dwCaps2 |= DDSCAPS2_OPAQUE; // maybe this is better, have to test it though
		}

		if(ddDesc.dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_POW2) {
			// if the texture dimensions must be a power of 2,
			// increase the surface size accordingly (the change is transparent to the user and programmer)
			for(ddsd.dwWidth  = 1; ddsd.dwWidth  < width ; ddsd.dwWidth  <<= 1);
			for(ddsd.dwHeight = 1; ddsd.dwHeight < height; ddsd.dwHeight <<= 1);
		}
		if(ddDesc.dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_SQUAREONLY || targetable) { // for some reason DDSCAPS_3DDEVICE causes non-square textures to fail drawing
			// if the texture has to be square
			if(ddsd.dwWidth > ddsd.dwHeight) ddsd.dwHeight = ddsd.dwWidth;
			else							 ddsd.dwWidth  = ddsd.dwHeight;
		}
	} else {
		// if we're not making a 3D accelerated surface, just try for a 2D video memory accelerated one
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
/// 	ddsd.ddsCaps.dwCaps |= DDSCAPS_VIDEOMEMORY; // no need to fail if video memory is unavailable
	}

			if(create_alpha) {
				dd_back_buf->GetPixelFormat(&ddpf);
				ddpf.dwFlags = DDPF_RGB | DDPF_ALPHAPIXELS;
				ddpf.dwRGBBitCount	   = 32;
				ddpf.dwRGBAlphaBitMask = 0xFF000000;
				ddpf.dwRBitMask 	   = 0x00FF0000;
				ddpf.dwGBitMask 	   = 0x0000FF00;
				ddpf.dwBBitMask 	   = 0x000000FF;
				ddsd.dwFlags |= DDSD_PIXELFORMAT;
				ddsd.ddpfPixelFormat = ddpf;
			} /*else {
				dd_back_buf->GetPixelFormat(&ddpf);
				ddpf.dwFlags = DDPF_RGB;
				if(win_screenbpp == 24) {
					ddpf.dwRGBBitCount = win_screenbpp; 
					ddpf.dwRBitMask 	   = 0x00FF0000;
					ddpf.dwGBitMask 	   = 0x0000FF00;
					ddpf.dwBBitMask 	   = 0x000000FF;
				} else {
					ddpf.dwRGBBitCount = win_screenbpp; 
					ddpf.dwRBitMask 	   = win_maskr;
					ddpf.dwGBitMask 	   = win_maskg;
					ddpf.dwBBitMask 	   = win_maskb;
				}
				ddsd.dwFlags |= DDSD_PIXELFORMAT;
				ddsd.ddpfPixelFormat = ddpf;
			}*/
			
			/*else {
				DDPIXELFORMAT ddpf;
				ddpf.dwSize = sizeof(ddpf);
				dd_back_buf->GetPixelFormat(&ddpf);
				ddpf.dwFlags = DDPF_RGB;
				ddpf.dwRGBBitCount	   = 16;
				ddpf.dwRBitMask 	   = win_maskr;
				ddpf.dwGBitMask 	   = win_maskg;
				ddpf.dwBBitMask 	   = win_maskb;
				ddsd.dwFlags |= DDSD_PIXELFORMAT;
				ddsd.ddpfPixelFormat = ddpf;
			}*/

	HRESULT status = dd_main->CreateSurface(&ddsd, &surface, NULL);
	if(status != DD_OK) {
		if(ddsd.dwFlags & DDSD_PIXELFORMAT) {
			ddsd.dwFlags ^= DDSD_PIXELFORMAT; // try again, this time accepting any format it wants to use
			status = dd_main->CreateSurface(&ddsd, &surface, NULL);
		}
		if(status != DD_OK) {
			DirectDrawError(status);
			return NULL;
		}
	}

#ifdef WINDX_DEBUG
	if(win_status == GAMEX_READY) {
		surface->GetPixelFormat(&ddpf); // get the pixel format DirectDraw made the surface in,
		char stringy [64];
		sprintf(stringy,"Created %dx%d %d-bit surface for a %dx%d image.\n",ddsd.dwWidth,ddsd.dwHeight,ddpf.dwRGBBitCount,width,height);
		debug.Output(stringy,"");
	}
#endif


//	surface->Restore();
	return surface;
}


bool WindowsDX::DirectXCleanUp (void)
{
	// DirectMusic songs:
	while(dm_musicx_index--)
		if(dm_musicx[dm_musicx_index] != NULL)
			dm_musicx[dm_musicx_index]->Release(),	dm_musicx[dm_musicx_index] = NULL;

	// DirectMusic
	if (dm_performance) dm_performance->CloseDown(), dm_performance->Release(), dm_performance = NULL;
	if (dm_loader)		dm_loader->Release(),		 dm_loader = NULL;
	
	// DirectSound
	if (ds_main)		ds_main->Release(), 		ds_main = NULL;

	// DirectInput
	if (di_mouse)		di_mouse->Unacquire(),		di_mouse->Release(),		di_mouse = NULL;
	if (di_keyboard)	di_keyboard->Unacquire(),	di_keyboard->Release(), 	di_keyboard = NULL;
	if (di_main)		di_main->Release(), 		di_main = NULL;

	// Direct3D
	if (d3d_device) 	d3d_device->Release(),		d3d_device = NULL;	
	if (d3d_main)		d3d_main->Release(),		d3d_main = NULL;	

	// DirectDraw
	if (dd_back_buf)	dd_back_buf->Release(), 	dd_back_buf = NULL;
	if (dd_front_buf)	dd_front_buf->Release(),	dd_front_buf = NULL;	
	if (dd_main)		dd_main->Release(), 		dd_main = NULL;

	CoUninitialize();
	return true;
}

bool WindowsDX::CleanUp (void)
{
	if (DirectXCleanUp()==FALSE) {
		#ifdef WINDX_DEBUG
			debug.Output ("WinDX::CleanUp: Unable to properly shut down DirectX.\n", "") ;
		#endif			
		return false;
	}
	return true;
}

int WindowsDX::ViewUpdate (void)
{
	if(win_status != GAMEX_READY)
		return 0;

	// DirectX 8.1
	//dd_device->Present (NULL, NULL, win_hwnd, NULL) ;

	// DirectDraw
	HRESULT status; 
	
//	if(!CanDraw()) {
//		if(win_active) {
			if(dd_front_buf->IsLost()) {
				dd_front_buf->Restore();
//				dd_main->RestoreAllSurfaces();

				dd_was_lost=1;

#ifdef WINDX_DEBUG
				if(!dd_front_buf->IsLost())
					debug.Output("WinDX: DirectX surfaces were lost, but have now been properly restored.\n","");
//				else
//					debug.Output("WinDX: The front buffer has been lost and could NOT be restored.\n","");
#endif
				if(dd_front_buf->IsLost())
					return 0;
				else {
					dd_main->RestoreAllSurfaces();

					// now we restore the 3D drawing device.
					// for some reason, if we don't do so RIGHT HERE
					// strange and inexplicable image corruption occurs

					DrawClose() ;
					if(d3d_device != NULL) { 
						unsigned long temp_long = -1;
						status = d3d_device->ValidateDevice(&temp_long);

						if(status == DDERR_SURFACELOST) { // in newer versions of DirectX this is really D3DERR_DEVICELOST, which makes more sense
							d3d_device->Release();
							bool success = CreateDeviceForSurface(d3d_device, dd_back_buf); // may fail if DrawClose() is not done before this

				#ifdef WINDX_DEBUG
							if(success)
								debug.Output("WinDX: The 3D rendering device has been lost and is now properly restored.\n","");
							else
								debug.Output("WinDX: The 3D rendering device has been lost and could NOT be restored.\n","");
				#endif
						}
					}
				}
			}
//		}
//		return 1;
//	}
	
	DrawClose() ;

	if(win_page_flipping) {
		int flags = DDFLIP_WAIT;
		if(!win_vsync_enabled)
			flags |= DDFLIP_NOVSYNC;
//		if(win_vsync_enabled)
//			SyncWithRefresh();
///		else
//			flags |= DDFLIP_NOVSYNC;
		status = dd_front_buf->Flip(NULL, flags) ;	// Flip back buffer to front
	} else {
		// page flipping off, so do a "regular" blit... 
		// use Blt since we might want to scale the viewport to the screen
		// and, if in windowed mode, because we have a clipper set to the window

		// set the source and destination rectangles:
		RECT src_rect = {0, 0, win_width, win_height} ;
		RECT dst_rect = {win_xpos+win_xoffset, win_ypos+win_yoffset, win_xpos+win_xoffset+win_realx, win_ypos+win_yoffset+win_realy} ;

		if(win_vsync_enabled) {  // if vsync hasn't been disabled,
			SyncWithRefresh(); // wait until we're synchronized with the screen's refresh

			// debugging, delete the following code once vsync has been perfected:
			/*
					#ifdef WINDX_DEBUG
					// this code draws a red line where the scanline is,
					// so the success of refresh synchronization can be seen
					// by how infrequently the red line is over the game
						DWORD scanline = -1;
						dd_main->GetScanLine(&scanline); // delete this line
						int drawline = (scanline-win_ypos+win_yoffset)*win_height/win_realy;
						if(drawline > 0 && drawline < win_height)
							DrawBox(0,drawline,win_width,drawline+2,255,0,0); // red line for after refreshsync
						DrawClose() ;
					#endif
			*/
			// end of deletable debugging code
		}

		// do the blit:
		status = GameX.dd_front_buf->Blt(&dst_rect, dd_back_buf,
										 &src_rect,
										 DDBLT_WAIT,
										 NULL);
	}

	// we might clear the back buffer at this point to avoid visual remnants of previous frames hanging around,
	// but it might be intentional in some circumstances for such remnants to be visible,
	// so, to allow greater freedom for the game developer, the back buffer is not automatically cleared.

	if(status == DD_OK)
		return 1;
	else
		return 0;
}

// note: not entirely reliable yet, still working on it:
void WindowsDX::SyncWithRefresh (void)
{
//	dd_main->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN,NULL); // delete this line
//	return; // delete this line

// this function could just call dd_main->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN,NULL)
// and return, but there are two problems with this:
// 1.) It does not guarantee monitor refresh synchronization and in fact
//	   can cause worse graphical glitches than those that result from not synchronizing at all,
//	   depending on the adapter and video driver (and its version) and cpu speed.
// 2.) It uses all of the computer's processing power when waiting,
//	   which can slow down background processes and cause the CPU to overheat.
//
// so, here is what we're doing instead:
// 1.) get the current scanline location
// 2.) if the scanline is NOT over or almost over the game drawing area,
//	   do nothing since we're already synchronized if so.
// 3.) otherwise, since updating the view when the scanline is over the game causes graphical glitches,
//	   sleep until we think the scanline will have moved down far enough that it won't be over the game.
//
// this method is not perfect, and may allow for graphical glitches anyway, depending on the machine,
// but it is generally better than not synchronizing and it wastes a lot less CPU power than the standard way.
// as an added bonus, this method only synchronizes the part of the screen that has the game being drawn on it,


	DWORD scanline = 0; dd_main->GetScanLine(&scanline); // holds last scanline value so we can tell if/when wraparound occurs
	long timecalc = (win_realy+win_ypos+win_yoffset - scanline) * GetClocksPerSec() / CLOCKS_PER_SEC / win_scanline_speed;

	if(scanline <= win_ypos+win_yoffset || scanline >= win_realy+win_ypos+win_yoffset)
		return; // we can draw at this instant without tearing glitches

	AccurateSleep(timecalc,5); // sleep until we think we can draw without tearing glitches

	return;
}


inline void WindowsDX::SetActivation (BOOL act)
{
	win_active = act;
}

inline void WindowsDX::SetPosition (int x, int y)
{
	if(x > win_screenx) x -= 65535; // reverse unsigning to avoid totally wrong window position
	if(y > win_screeny) y -= 65535;

	win_xpos = x;
	win_ypos = y;
}

void WindowsDX::SetProportional (void)
{
	int screenx = win_realx;
	int screeny = win_realy;

	if(win_width * screeny > win_height * screenx) {
		win_realx = win_width*screenx/win_width;
		win_realy = win_height*screenx/win_width;
	} else {
		win_realx = win_width*screeny/win_height;
		win_realy = win_height*screeny/win_height;
	}

	win_xoffset = (screenx - win_realx)/2; // temporarily acts as initial offset, which doesn't change
	win_yoffset = (screeny - win_realy)/2; // for fullscreen mode, allowing the viewport to be centered
}

void WindowsDX::SetView (int x1, int y1, int x2, int y2)
{
	win_viewx1 = x1;
	win_viewy1 = y1;
	win_viewx2 = x2;
	win_viewy2 = y2;
}

inline void WindowsDX::AddBufferedKeyPress (char ch)
{
	win_keybuf[win_keys++] = ch;
	if (win_keys > WINDX_BUFSIZE) {
		win_keys = WINDX_BUFSIZE;
/// 	debug.Output ("WinDX: Keyboard buffer overflow. Consider calling GameX.ClearKeyBuffer()\n", "") ;
	}
}

bool WindowsDX::DrawOpen (void)
{
	if(drawLocked) return true;

	// DirectX 6.0
	DDSURFACEDESC2		ddsd;

	ddsd.dwSize = sizeof(ddsd) ;

	HRESULT status = dd_back_buf->Lock(NULL, &ddsd, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);	// lock drawing surface
	if(status != DD_OK) {
//		win_status = GAMEX_LOCK_FAILED;
//		ReportProblem() ;
		DirectDrawError(status);
		return false;
	}

	win_screen = (char *) ddsd.lpSurface;

	drawLocked = true;

	return true;
}

void WindowsDX::DirectDrawError (int status)
{
	if(status == DD_OK || status == dd_last_error) return;
	dd_last_error = status;
	switch (status) {
	case DDERR_ALREADYINITIALIZED: debug.Output ("DirectDraw: This object is already initialized.\n","") ; break;
	case DDERR_BLTFASTCANTCLIP: debug.Output ("DirectDraw: Return if a clipper object is attached to the source surface passed into a BltFast call.\n","") ; break;
	case DDERR_CANNOTATTACHSURFACE: debug.Output ("DirectDraw: This surface can not be attached to the requested surface.\n","") ; break; 
	case DDERR_CANNOTDETACHSURFACE: debug.Output ("DirectDraw: This surface can not be detached from the requested surface.\n","") ;break;
	case DDERR_CANTCREATEDC: debug.Output ("DirectDraw: Windows can not create any more DCs\n","") ; break;
	case DDERR_CANTDUPLICATE: debug.Output ("DirectDraw: Can't duplicate primary & 3D surfaces, or surfaces that are implicitly created.\n","") ; break;
	case DDERR_CLIPPERISUSINGHWND: debug.Output ("DirectDraw: An attempt was made to set a cliplist for a clipper object that is already monitoring an hwnd.\n","") ; break;
	case DDERR_COLORKEYNOTSET: debug.Output ("DirectDraw: No src color key specified for this operation.\n","") ; break;
	case DDERR_CURRENTLYNOTAVAIL: debug.Output ("DirectDraw: Support is currently not available.\n","") ; break;
	case DDERR_DIRECTDRAWALREADYCREATED: debug.Output ("DirectDraw: A DirectDraw object representing this driver has already been created for this process.\n","") ; break;
	case DDERR_EXCEPTION: debug.Output ("DirectDraw: An exception was encountered while performing the requested operation.\n","") ; break;
	case DDERR_EXCLUSIVEMODEALREADYSET: debug.Output ("DirectDraw: An attempt was made to set the cooperative level when it was already set to exclusive.\n","") ; break;
	case DDERR_GENERIC: debug.Output ("DirectDraw: Generic failure. Unfortunately, this means that DirectX has no idea what went wrong.\n","") ; break;
	case DDERR_HEIGHTALIGN: debug.Output ("DirectDraw: Height of rectangle provided is not a multiple of reqd alignment.\n","") ; break;
	case DDERR_HWNDALREADYSET: debug.Output ("DirectDraw: The CooperativeLevel HWND has already been set. It can not be reset while the process has surfaces or palettes created.\n","") ; break;
	case DDERR_HWNDSUBCLASSED: debug.Output ("DirectDraw: HWND used by DirectDraw CooperativeLevel has been subclassed, this prevents DirectDraw from restoring state.\n","") ; break;
	case DDERR_IMPLICITLYCREATED: debug.Output ("DirectDraw: This surface can not be restored because it is an implicitly created surface.\n","") ; break;
	case DDERR_INCOMPATIBLEPRIMARY: debug.Output ("DirectDraw: Unable to match primary surface creation request with existing primary surface.\n","") ; break;
	case DDERR_INVALIDCAPS: debug.Output ("DirectDraw: One or more of the caps bits passed to the callback are incorrect.\n","") ; break;
	case DDERR_INVALIDCLIPLIST: debug.Output ("DirectDraw: DirectDraw does not support the provided cliplist.\n","") ; break;
	case DDERR_INVALIDDIRECTDRAWGUID: debug.Output ("DirectDraw: The GUID passed to DirectDrawCreate is not a valid DirectDraw driver identifier.\n","") ; break;
	case DDERR_INVALIDMODE: debug.Output ("DirectDraw: DirectDraw does not support the requested mode.\n","") ; break;
	case DDERR_INVALIDOBJECT: debug.Output ("DirectDraw: DirectDraw received a pointer that was an invalid DIRECTDRAW object.\n","") ; break;
	case DDERR_INVALIDPARAMS: debug.Output ("DirectDraw: One or more of the parameters passed to the function are incorrect.\n","") ; break;
	case DDERR_INVALIDPIXELFORMAT: debug.Output ("DirectDraw: The pixel format was invalid as specified.\n","") ; break;
	case DDERR_INVALIDPOSITION: debug.Output ("DirectDraw: Returned when the position of the overlay on the destination is no longer legal for that destination->\n","") ; break;
	case DDERR_INVALIDRECT: debug.Output ("DirectDraw: Rectangle provided was invalid.\n","") ; break;
	case DDERR_LOCKEDSURFACES: debug.Output ("DirectDraw: Operation could not be carried out because one or more surfaces are locked.\n","") ; break;
	case DDERR_NO3D: debug.Output ("DirectDraw: There is no 3D present.\n","") ; break;
	case DDERR_NOALPHAHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no alpha accleration hardware present or available.\n","") ; break;
	//case DDERR_NOANTITEARHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no hardware support for synchronizing blits to avoid tearing.	\n","") ; break;
	case DDERR_NOBLTHW: debug.Output ("DirectDraw: No blitter hardware present.\n","") ; break;
	//case DDERR_NOBLTQUEUEHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no hardware support for asynchronous blitting.\n","") ; break;
	case DDERR_NOCLIPLIST: debug.Output ("DirectDraw: No cliplist available.\n","") ; break;
	case DDERR_NOCLIPPERATTACHED: debug.Output ("DirectDraw: No clipper object attached to surface object.\n","") ; break;
	case DDERR_NOCOLORCONVHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no color conversion hardware present or available.\n","") ; break;
	case DDERR_NOCOLORKEY: debug.Output ("DirectDraw: Surface doesn't currently have a color key\n","") ; break;
	case DDERR_NOCOLORKEYHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no hardware support of the destination color key.\n","") ; break;
	case DDERR_NOCOOPERATIVELEVELSET: debug.Output ("DirectDraw: Create function called without DirectDraw object method SetCooperativeLevel being called.\n","") ; break;
	case DDERR_NODC: debug.Output ("DirectDraw: No DC was ever created for this surface.\n","") ; break;
	case DDERR_NODDROPSHW: debug.Output ("DirectDraw: No DirectDraw ROP hardware.\n","") ; break;
	case DDERR_NODIRECTDRAWHW: debug.Output ("DirectDraw: A hardware-only DirectDraw object creation was attempted but the driver did not support any hardware.\n","") ; break;
	case DDERR_NOEMULATION: debug.Output ("DirectDraw: Software emulation not available.\n","") ; break;
	case DDERR_NOEXCLUSIVEMODE: debug.Output ("DirectDraw: Operation requires the application to have exclusive mode but the application does not have exclusive mode.\n","") ; break;
	case DDERR_NOFLIPHW: debug.Output ("DirectDraw: Flipping visible surfaces is not supported.\n","") ; break;
	case DDERR_NOGDI: debug.Output ("DirectDraw: There is no GDI present.\n","") ; break;
	case DDERR_NOHWND: debug.Output ("DirectDraw: Clipper notification requires an HWND or no HWND has previously been set as the CooperativeLevel HWND.\n","") ; break;
	case DDERR_NOMIRRORHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no hardware present or available.\n","") ; break;
	case DDERR_NOOVERLAYDEST: debug.Output ("DirectDraw: Returned when GetOverlayPosition is called on an overlay that UpdateOverlay has never been called on to establish a destination->\n","") ; break;
	case DDERR_NOOVERLAYHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no overlay hardware present or available.\n","") ; break;
	case DDERR_NOPALETTEATTACHED: debug.Output ("DirectDraw: No palette object attached to this surface.\n","") ; break;
	case DDERR_NOPALETTEHW: debug.Output ("DirectDraw: No hardware support for 16 or 256 color palettes.\n","") ; break;
	case DDERR_NORASTEROPHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no appropriate raster op hardware present or available.\n","") ; break;
	case DDERR_NOROTATIONHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no rotation hardware present or available.\n","") ; break;
	case DDERR_NOSTRETCHHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no hardware support for stretching.\n","") ; break;
	case DDERR_NOT4BITCOLOR: debug.Output ("DirectDraw: DirectDrawSurface is not in 4 bit color palette and the requested operation requires 4 bit color palette.\n","") ; break;
	case DDERR_NOT4BITCOLORINDEX: debug.Output ("DirectDraw: DirectDrawSurface is not in 4 bit color index palette and the requested operation requires 4 bit color index palette.\n","") ; break;
	case DDERR_NOT8BITCOLOR: debug.Output ("DirectDraw: DirectDrawSurface is not in 8 bit color mode and the requested operation requires 8 bit color.\n","") ; break;
	case DDERR_NOTAOVERLAYSURFACE: debug.Output ("DirctX: Returned when an overlay member is called for a non-overlay surface.\n","") ; break;
	case DDERR_NOTEXTUREHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no texture mapping hardware present or available.\n","") ; break;
	case DDERR_NOTFLIPPABLE: debug.Output ("DirectDraw: An attempt has been made to flip a surface that is not flippable.\n","") ; break;
	case DDERR_NOTFOUND: debug.Output ("DirectDraw: Requested item was not found.\n","") ; break;
	case DDERR_NOTLOCKED: debug.Output ("DirectDraw: Surface was not locked.  An attempt to unlock a surface that was not locked at all, or by this process, has been attempted.\n","") ; break;
	case DDERR_NOTPALETTIZED: debug.Output ("DirectDraw: The surface being used is not a palette-based surface.\n","") ; break;
	case DDERR_NOVSYNCHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no hardware support for vertical blank synchronized operations.\n","") ; break;
	case DDERR_NOZBUFFERHW: debug.Output ("DirectDraw: Operation could not be carried out because there is no hardware support for zbuffer blitting.\n","") ; break;
	case DDERR_NOZOVERLAYHW: debug.Output ("DirectDraw: Overlay surfaces could not be z layered based on their BltOrder because the hardware does not support z layering of overlays.\n","") ; break;
	case DDERR_OUTOFCAPS: debug.Output ("DirectDraw: The hardware needed for the requested operation has already been allocated.\n","") ; break;
	case DDERR_OUTOFMEMORY: debug.Output ("DirectDraw: DirectDraw does not have enough memory to perform the operation.\n","") ; break;
	case DDERR_OUTOFVIDEOMEMORY: debug.Output ("DirectDraw: DirectDraw does not have enough memory to perform the operation.\n","") ; break;
	case DDERR_OVERLAYCANTCLIP: debug.Output ("DirectDraw: The hardware does not support clipped overlays.\n","") ; break;
	case DDERR_OVERLAYCOLORKEYONLYONEACTIVE: debug.Output ("DirectDraw: Can only have ony color key active at one time for overlays.\n","") ; break;
	case DDERR_OVERLAYNOTVISIBLE: debug.Output ("DirectDraw: Returned when GetOverlayPosition is called on a hidden overlay.\n","") ; break;
	case DDERR_PALETTEBUSY: debug.Output ("DirectDraw: Access to this palette is being refused because the palette is already locked by another thread.\n","") ; break;
	case DDERR_PRIMARYSURFACEALREADYEXISTS: debug.Output ("DirectDraw: This process already has created a primary surface.\n","") ; break;
	case DDERR_REGIONTOOSMALL: debug.Output ("DirectDraw: Region passed to Clipper::GetClipList is too small.\n","") ; break;
	case DDERR_SURFACEALREADYATTACHED: debug.Output ("DirectDraw: This surface is already attached to the surface it is being attached to.\n","") ; break;
	case DDERR_SURFACEALREADYDEPENDENT: debug.Output ("DirectDraw: This surface is already a dependency of the surface it is being made a dependency of.\n","") ; break;
	case DDERR_SURFACEBUSY: debug.Output ("DirectDraw: Access to this surface is being refused because the surface is locked.\nIf the problem is with the back buffer, try calling GameX.DrawClose() first.","") ; break;
	case DDERR_SURFACEISOBSCURED: debug.Output ("DirectDraw: Access to surface refused because the surface is obscured.\n","") ; break;
	case DDERR_SURFACELOST: debug.Output ("DirectDraw: Access to this surface is being refused because the surface memory is gone. The DirectDrawSurface object representing this surface should have Restore called on it.\n","") ; break;
	case DDERR_SURFACENOTATTACHED: debug.Output ("DirectDraw: The requested surface is not attached.\n","") ; break;
	case DDERR_TOOBIGHEIGHT: debug.Output ("DirectDraw: Height requested by DirectDraw is too large.\n","") ; break;
	case DDERR_TOOBIGSIZE: debug.Output ("DirectDraw: Size requested by DirectDraw is too large --	the individual height and width are OK.\n","") ; break;
	case DDERR_TOOBIGWIDTH: debug.Output ("DirectDraw: Width requested by DirectDraw is too large.\n","") ; break;
	case DDERR_UNSUPPORTED: debug.Output ("DirectDraw: Action not supported.\n","") ; break;
	case DDERR_UNSUPPORTEDFORMAT: debug.Output ("DirectDraw: FOURCC format requested is unsupported by DirectDraw.\n","") ; break;
	case DDERR_UNSUPPORTEDMASK: debug.Output ("DirectDraw: Bitmask in the pixel format requested is unsupported by DirectDraw.\n","") ; break;
	case DDERR_VERTICALBLANKINPROGRESS: debug.Output ("DirectDraw: Vertical blank is in progress.\n","") ; break;
	case DDERR_WASSTILLDRAWING: debug.Output ("DirectDraw: Informs DirectDraw that the previous Blt which is transfering information to or from this Surface is incomplete.\n","") ; break;
	case DDERR_WRONGMODE: debug.Output ("DirectDraw: This surface can not be restored because it was created in a different mode.\n","") ; break;
	default: debug.Output ("DirectDraw: Unknown Error.\n","") ; dd_last_error = DD_OK; break;
	}
}

void WindowsDX::Direct3DError (int status)
{
	if(status == D3D_OK || status == dd_last_error) return;
	dd_last_error = status;
	switch (status) {
	case D3DERR_WRONGTEXTUREFORMAT: debug.Output ("Direct3D: D3DERR_WRONGTEXTUREFORMAT.\n","") ; break;
	case D3DERR_UNSUPPORTEDCOLOROPERATION: debug.Output ("Direct3D: D3DERR_UNSUPPORTEDCOLOROPERATION.\n","") ; break;
	case D3DERR_UNSUPPORTEDCOLORARG: debug.Output ("Direct3D: D3DERR_UNSUPPORTEDCOLORARG.\n","") ; break;
	case D3DERR_UNSUPPORTEDALPHAOPERATION: debug.Output ("Direct3D: D3DERR_UNSUPPORTEDALPHAOPERATION.\n","") ; break;
	case D3DERR_UNSUPPORTEDALPHAARG: debug.Output ("Direct3D: D3DERR_UNSUPPORTEDALPHAARG.\n","") ; break;
	case D3DERR_TOOMANYOPERATIONS: debug.Output ("Direct3D: D3DERR_TOOMANYOPERATIONS.\n","") ; break;
	case D3DERR_CONFLICTINGTEXTUREFILTER: debug.Output ("Direct3D: D3DERR_CONFLICTINGTEXTUREFILTER.\n","") ; break;
	case D3DERR_CONFLICTINGRENDERSTATE: debug.Output ("Direct3D: D3DERR_CONFLICTINGRENDERSTATE.\n","") ; break;
	case D3DERR_UNSUPPORTEDTEXTUREFILTER: debug.Output ("Direct3D: D3DERR_UNSUPPORTEDTEXTUREFILTER.\n","") ; break;
	case D3DERR_CONFLICTINGTEXTUREPALETTE: debug.Output ("Direct3D: D3DERR_CONFLICTINGTEXTUREPALETTE.\n","") ; break;
//	case D3DERR_DRIVERINTERNALERROR: debug.Output ("Direct3D: D3DERR_DRIVERINTERNALERROR.\n","") ; break;
//	case D3DERR_NOTFOUND: debug.Output ("Direct3D: D3DERR_NOTFOUND.\n","") ; break;
//	case D3DERR_MOREDATA: debug.Output ("Direct3D: D3DERR_MOREDATA.\n","") ; break;
//	case D3DERR_DEVICELOST: debug.Output ("Direct3D: D3DERR_DEVICELOST.\n","") ; break;
//	case D3DERR_DEVICENOTRESET: debug.Output ("Direct3D: D3DERR_DEVICENOTRESET.\n","") ; break;
//	case D3DERR_NOTAVAILABLE: debug.Output ("Direct3D: D3DERR_NOTAVAILABLE.\n","") ; break;
//	case D3DERR_OUTOFVIDEOMEMORY: debug.Output ("Direct3D: D3DERR_OUTOFVIDEOMEMORY.\n","") ; break;
//	case D3DERR_INVALIDDEVICE: debug.Output ("Direct3D: D3DERR_INVALIDDEVICE.\n","") ; break;
//	case D3DERR_INVALIDCALL: debug.Output ("Direct3D: D3DERR_INVALIDCALL.\n","") ; break;
//	case D3DERR_DRIVERINVALIDCALL: debug.Output ("Direct3D: D3DERR_DRIVERINVALIDCALL.\n","") ; break;
	case E_NOINTERFACE: debug.Output ("Direct3D: E_NOINTERFACE.\n","") ; break;
	default: debug.Output ("Direct3D: Unknown Error.\n","") ; dd_last_error = DD_OK; break;
	}
}

void WindowsDX::DrawColor (int r, int g, int b)
{
	win_r = r;	win_g = g;	win_b = b;
	win_rgb = RGB(win_r, win_g, win_b) ;
}

void WindowsDX::GetDrawColor (int & r, int & g, int & b)
{
	r = win_r;
	g = win_g;
	b = win_b;
}

int WindowsDX::ClipRects (RECT & src_rect, RECT & dst_rect, const RECT & src_bounds, const RECT & dst_bounds)
{
	if (dst_rect.right >= dst_bounds.left && dst_rect.left <= dst_bounds.right) { // In-View Horizontally
		if (dst_rect.bottom >= dst_bounds.top && dst_rect.top <= dst_bounds.bottom) { // In-View Vertically 

			if(dst_rect.right-dst_rect.left == src_rect.right-src_rect.left
			&& dst_rect.bottom-dst_rect.top == src_rect.bottom-src_rect.top) { // If we don't need to scale

				if(dst_rect.left < dst_bounds.left) {
					src_rect.left += dst_bounds.left - dst_rect.left;
					dst_rect.left = dst_bounds.left;
				}
				if(dst_rect.top < dst_bounds.top) {
					src_rect.top += dst_bounds.top - dst_rect.top;
					dst_rect.top = dst_bounds.top;
				}
				if(dst_rect.right > dst_bounds.right) {
					src_rect.right += dst_bounds.right - dst_rect.right;
					dst_rect.right = dst_bounds.right;
				}
				if(dst_rect.bottom > dst_bounds.bottom) {
					src_rect.bottom += dst_bounds.bottom - dst_rect.bottom;
					dst_rect.bottom = dst_bounds.bottom;
				}
				// clip the source too... we could skip this is we were lazy, but then
				// giving an invalid source position would cause the game to crash
				if(src_rect.left < src_bounds.left) {
					dst_rect.left += src_bounds.left - src_rect.left;
					src_rect.left = src_bounds.left;
				}
				if(src_rect.top < src_bounds.top) {
					dst_rect.top += src_bounds.top - src_rect.top;
					src_rect.top = src_bounds.top;
				}
				if(src_rect.right > src_bounds.right) {
					dst_rect.right += src_bounds.right - src_rect.right;
					src_rect.right = src_bounds.right;
				}
				if(src_rect.bottom > src_bounds.bottom) {
					dst_rect.bottom += src_bounds.bottom - src_rect.bottom;
					src_rect.bottom = src_bounds.bottom;
				}
				if(dst_rect.left >= dst_rect.right || dst_rect.top >= dst_rect.bottom) return 0;
				return 1; // Image is in view and should be drawn as-is with the src_rect and dst_rect that we just (possibly) modified
			} else { // need to scale, and if we clip it we need to adjust the clipping amount to match
				
				// the clipping used here isn't perfect since it doesn't allow for pixels to be partially clipped,
				// but it's better than blindly assuming a 1:1 source:destination ratio

				if(dst_rect.left < dst_bounds.left) {
					src_rect.left += (dst_bounds.left - dst_rect.left) * (src_rect.right - src_rect.left) / (dst_rect.right - dst_rect.left);
					dst_rect.left = dst_bounds.left;
				}
				if(dst_rect.top < dst_bounds.top) {
					src_rect.top += (dst_bounds.top - dst_rect.top) * (src_rect.bottom - src_rect.top) / (dst_rect.bottom - dst_rect.top);
					dst_rect.top = dst_bounds.top;
				}
				if(dst_rect.right > dst_bounds.right) {
					src_rect.right += (dst_bounds.right - dst_rect.right) * (src_rect.right - src_rect.left) / (dst_rect.right - dst_rect.left);
					dst_rect.right = dst_bounds.right;
				}
				if(dst_rect.bottom > dst_bounds.bottom) {
					src_rect.bottom += (dst_bounds.bottom - dst_rect.bottom) * (src_rect.bottom - src_rect.top) / (dst_rect.bottom - dst_rect.top);
					dst_rect.bottom = dst_bounds.bottom;
				}
				if(src_rect.left < src_bounds.left) {
					dst_rect.left += (src_bounds.left - src_rect.left) * (dst_rect.right - dst_rect.left) / (src_rect.right - src_rect.left);
					src_rect.left = src_bounds.left;
				}
				if(src_rect.top < src_bounds.top) {
					dst_rect.top += (src_bounds.top - src_rect.top) * (dst_rect.bottom - dst_rect.top) / (src_rect.bottom - src_rect.top);
					src_rect.top = src_bounds.top;
				}
				if(src_rect.right > src_bounds.right) {
					dst_rect.right += (src_bounds.right - src_rect.right) * (dst_rect.right - dst_rect.left) / (src_rect.right - src_rect.left);
					src_rect.right = src_bounds.right;
				}
				if(src_rect.bottom > src_bounds.bottom) {
					dst_rect.bottom += (src_bounds.bottom - src_rect.bottom) * (dst_rect.bottom - dst_rect.top) / (src_rect.bottom - src_rect.top);
					src_rect.bottom = src_bounds.bottom;
				}
				if(dst_rect.left >= dst_rect.right || dst_rect.top >= dst_rect.bottom) return 0;
				return 2; // Image is in view and should be drawn SCALED with the src_rect and dst_rect that we just (possibly) modified
			}
		}
	}
	return 0; // Image is not in view at all and so shouldn't be drawn
}


void WindowsDX::SetMouseMove (int x, int y)
{
	mouse_x = x;
	mouse_y = y;

	if(!win_fullscreen && win_active && CanHideCursor()) {
		int dx = 2*GetMouseDX();
		int dy = 2*GetMouseDY();
		if(x < 0 || y < 0 || x >= win_realx || y >= win_realy) {
			UnhideCursor();
		} else {
			HideCursor();
		}
	}
}

// pauses whatever MIDI song is currently playing (has only been tested internally to GameX)
void WindowsDX::PauseMusic (void)
{
	if(dm_performance == NULL) return; // return if MusicX not initialized yet
/*	int index = dm_musicx_index;
	while(index--)
		if(dm_musicx[index] != NULL)
			if(IsMusicPlaying(dm_musicx[index]) {*/
				float current_tempo = master_music_tempo;
				SetMusicTempo(DMUS_MASTERTEMPO_MIN*2);
				master_music_tempo = current_tempo;

				float current_volume = master_music_volume;
				SetMasterMusicVolume(0);
				master_music_volume = current_volume;
//			}
}

// resumes whatever MIDI song is currently playing (has only been tested internally to GameX)
// due to the way it is paused, a stray note might play for a while after resuming
void WindowsDX::ResumeMusic (void)
{
	if(dm_performance == NULL) return; // return if MusicX not initialized yet
/*	int index = dm_musicx_index;
	while(index--)
		if(dm_musicx[index] != NULL)
			if(IsMusicPlaying(dm_musicx[index]) {*/
				SetMusicTempo(master_music_tempo);
				SetMasterMusicVolume(master_music_volume);
// 				VolumeFadeMusic(dm_musicx[index]->GetVolume(),1); 
//			}
}

void WindowsDX::HideCursor(void)
{
	if(!win_cursor_hidden) {
		ShowCursor(false);
		win_cursor_hidden = true;
	}
}

void WindowsDX::UnhideCursor(void)
{
	if(win_cursor_hidden) {
		ShowCursor(true);
		win_cursor_hidden = false;
	}
}

void WindowsDX::ClipTile (int x, int y, int xres, int yres, int &mx, int &my, int &offx, int &offy)
{
	if (x<0) {
		mx = ((int) (x / xres)) - 1;
		offx = (x % xres) + xres;
	} else {
		mx = ((int) (x / xres)) ;
		offx = x % xres;
	}
	if (y<0) {
		my = ((int) (y / yres)) - 1;		
		offy = (y % yres) + yres;
	} else {
		my = ((int) (y / yres)) ;		
		offy = y % yres;
	}	
}

// clear the view screen, setting all pixels to black
void WindowsDX::DrawClear (void)
{	
	DrawBox(0,0,win_width,win_height,0,0,0); // very fast
}

// Params: left, top, right, bottom, red (out of 255), green (out of 255), blue (out of 255)
// (this is the toned-down but ultra-fast version of DrawBox)
void WindowsDX::DrawBox (int x1, int y1, int x2, int y2, int r, int g, int b)
{
	RECT dst_rect = {x1,y1,x2,y2};

	DrawClose() ; // unlock drawing surface if necessary

	DDBLTFX ddbltfx;
	ddbltfx.dwSize = sizeof(ddbltfx);
	ddbltfx.dwFillColor = (((r >> 3) << win_shiftr) + ((g >> 2) << win_shiftg) + ((b >> 3) << win_shiftb));

	HRESULT status = dd_back_buf->Blt(&dst_rect, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &ddbltfx);
	if(status != DD_OK)
		DirectDrawError(status);
	else return;
	
	// if we failed, try it the old way:

	DrawOpen() ;

	unsigned short *start, *pixel;
	int i, j, clr;

	clr = ddbltfx.dwFillColor;
	start = (unsigned short *) win_screen;
	start += y1*win_pitch + x1;
	for (j=y1; j<=y2; j++) {
		pixel = start;
		for (i=x1; i<=x2; i++)
			*pixel++ = clr;
		start += win_pitch;
	}
}

// the full version of DrawBox (supports 3D-accelerated alpha blending and vertex shading)
void WindowsDX::DrawBox (int flags, int x1, int y1, int x2, int y2, int r, int g, int b, int a, int r2, int g2, int b2, int a2, int r3, int g3, int b3, int a3, int r4, int g4, int b4, int a4, float angle)
{
	if(win_lowquality) { 
		if(flags & DRAWOP_SKIPPABLE || flags & DRAWOP_ALTERNABLE && win_alternator)
			return;
		if(flags & DRAWOP_PLAINABLE) {
			flags = (flags & DRAWOPTIONSMASK) | DRAW_PLAIN;
			if(flags & DRAWOP_DITHER) flags ^= DRAWOP_DITHER;
			if(a||a2||a3||a4)
				a=a2=a3=a4=255;
			r=g=b=r2=g2=b2=r3=g3=b3=r4=g4=b4=255;
		}
	}

	if(r < 0)	r = 0;
	if(r > 255) r = 255;
	if(g < 0)	g = 0;
	if(g > 255) g = 255;
	if(b < 0)	b = 0;
	if(b > 255) b = 255;
	if(a < 0)	a = 0;
	if(a > 255) a = 255;

	if(r2 < 0)	 r2 = 0;
	if(r2 > 255) r2 = 255;
	if(g2 < 0)	 g2 = 0;
	if(g2 > 255) g2 = 255;
	if(b2 < 0)	 b2 = 0;
	if(b2 > 255) b2 = 255;
	if(a2 < 0)	 a2 = 0;
	if(a2 > 255) a2 = 255;

	if(r3 < 0)	 r3 = 0;
	if(r3 > 255) r3 = 255;
	if(g3 < 0)	 g3 = 0;
	if(g3 > 255) g3 = 255;
	if(b3 < 0)	 b3 = 0;
	if(b3 > 255) b3 = 255;
	if(a3 < 0)	 a3 = 0;
	if(a3 > 255) a3 = 255;

	if(r4 < 0)	 r4 = 0;
	if(r4 > 255) r4 = 255;
	if(g4 < 0)	 g4 = 0;
	if(g4 > 255) g4 = 255;
	if(b4 < 0)	 b4 = 0;
	if(b4 > 255) b4 = 255;
	if(a4 < 0)	 a4 = 0;
	if(a4 > 255) a4 = 255;

	if(flags & DRAW_ADD || flags & DRAW_SUBTRACT || flags & DRAW_GHOST || flags & DRAW_INVERT || flags & DRAW_BURN || flags & DRAW_INTENSIFY) {
		// scale the luminosity by the alpha which would otherwise be ignored
		ScaleByAlpha(r,g,b,a,r2,g2,b2,a2,r3,g3,b3,a3,r4,g4,b4,a4);
//		if(win_lowquality && r+r2+r3+r4<128
//		&& g+g2+g3+g4<128 && b+b2+b3+b4<128) return; // time-conservative return for when there's no video acceleration
		if(!(r||g||b||r2||g2||b2||r3||g3||b3||r4||g4||b4))
			return; // return if there isn't any luminosity
	} else {
//		if(win_lowquality && a+a2+a3+a4<128) return; // time-conservative return for when there's no video acceleration
		if(!(a||a2||a3||a4))
			return; // return if there isn't any opacity
	}

	RECT dst_rect = {x1,y1,x2,y2};
	RECT dst_bounds = {0,0,win_width,win_height};
	if(DrawImage3D(d3d_device,NULL,dst_rect,dst_rect,dst_rect,dst_bounds,win_width,win_height,r,g,b,a,r2,g2,b2,a2,r3,g3,b3,a3,r4,g4,b4,a4,angle,flags,0))
		return; // return if succeeded in drawing the box

	// if we failed to draw it in 3D:

// the following code is currently broken
// (maybe has to do with bad pitch?):
/*
	if(!a) { // only support single alpha param when non-D3D for now
		// no alpha, can do much faster version
		DrawBox (x1, y1, x2, y2, r, g, b);
	} else {
		GameX.DrawOpen() ;
		r = (r * a / 255) >> 3;
		g = (g * a / 255) >> 2;
		b = (b * a / 255) >> 3;
		a = 255 - a;

		// dereferencing view_pix is EXTREMELY slow, so we're going to create a storage area for it and
		// memcpy a line of view_pix at a time, then refer to that copy instead, since that's actually much faster
		XBYTE2 *temp = new XBYTE2[win_width];
		for(int y = 0 ; y < win_height ; y++)
		{
			unsigned short * view_start = (unsigned short *) win_screen + win_ytable[y];
			memcpy (temp, view_start, win_width<<1) ; // copy a source-length line of the destination -- the slowest part by far
			for(int x = 0 ; x < win_width ; x++)
			{
				unsigned short * view_pix = view_start + x;
				int v = temp[view_pix-view_start] ; // speedy alternative to: v = *view_pix
				*view_pix = (((((v & win_maskb) >> win_shiftb) * a / 255 + b) << win_shiftb))
						  | (((((v & win_maskg) >> win_shiftg) * a / 255 + g) << win_shiftg))
						  | (((((v & win_maskr) >> win_shiftr) * a / 255 + r) << win_shiftr)) ;
			}
		}
		delete [] temp;
	}
*/

}

// convert lowered (from 255) alpha into lowered r,g,b
// used for drawing modes that draw using red, green, blue, effective opacity determined by brightness,
void WindowsDX::ScaleByAlpha (int & r, int & g, int & b, int & a, int & r2, int & g2, int & b2, int & a2, int & r3, int & g3, int & b3, int & a3, int & r4, int & g4, int & b4, int & a4)
{
	if(a!=255) {
		r  = r	* a  / 255;
		g  = g	* a  / 255;
		b  = b	* a  / 255;
	}
	if(a2!=255) {
		r2 = r2 * a2 / 255;
		g2 = g2 * a2 / 255;
		b2 = b2 * a2 / 255;
	}
	if(a3!=255) {
		r3 = r3 * a3 / 255;
		g3 = g3 * a3 / 255;
		b3 = b3 * a3 / 255;
	}
	if(a4!=255) {
		r4 = r4 * a4 / 255;
		g4 = g4 * a4 / 255;
		b4 = b4 * a4 / 255;
	}
	a = a2 = a3 = a4 = 255;
}

void WindowsDX::GetAverageViewportColors(int & r,int & g,int & b,int & r2,int & g2,int & b2,int & r3,int & g3,int & b3,int & r4,int & g4,int & b4)
{
	DrawClose();

	RECT src_rect = {0, 0, win_width, win_height} ;
	RECT dst_rect = {0, 0, 6, 6} ;

	HRESULT status = temp_surface->Blt(&dst_rect, dd_back_buf,
									   &src_rect,
									   DDBLT_WAIT,
									   NULL);

	DDSURFACEDESC2 ddsd;
	ddsd.dwSize = sizeof(ddsd) ;
	status = temp_surface->Lock(NULL, &ddsd, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);	// lock drawing surface
	if(status != DD_OK) {
		if(status == DDERR_SURFACELOST) {
			temp_surface->Restore();
			status = temp_surface->Lock(NULL, &ddsd, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);	// lock drawing surface
			if(status != DD_OK) {
				DirectDrawError(status);
				return;
			}
		} else {
	//		win_status = GAMEX_LOCK_FAILED;
	//		ReportProblem() ;
			DirectDrawError(status);
			return;
		}
	}

	XBYTE2 * dest = (XBYTE2 *) ddsd.lpSurface;
	int v;
	int adj_pitch = ddsd.lPitch >> 1; // >> 1 for 16 bit

	v = dest[1+1*adj_pitch];
	b = ((v & win_maskb) >> win_shiftb) << 3;
	g = ((v & win_maskg) >> win_shiftg) << 2;
	r = ((v & win_maskr) >> win_shiftr) << 3;
	v = dest[4+1*adj_pitch];
	b2 = ((v & win_maskb) >> win_shiftb) << 3;
	g2 = ((v & win_maskg) >> win_shiftg) << 2;
	r2 = ((v & win_maskr) >> win_shiftr) << 3;
	v = dest[1+4*adj_pitch];
	b3 = ((v & win_maskb) >> win_shiftb) << 3;
	g3 = ((v & win_maskg) >> win_shiftg) << 2;
	r3 = ((v & win_maskr) >> win_shiftr) << 3;
	v = dest[4+4*adj_pitch];
	b4 = ((v & win_maskb) >> win_shiftb) << 3;
	g4 = ((v & win_maskg) >> win_shiftg) << 2;
	r4 = ((v & win_maskr) >> win_shiftr) << 3;

	temp_surface->Unlock(NULL);
}

// draws over the screen in 4 colors chosen from the colors already onscreen,
// a is the opacity with which to draw over the screen
void WindowsDX::SpecialViewFade(int a)
{
	int r,g,b, r2,g2,b2, r3,g3,b3, r4,g4,b4, min, max;
	r=g=b=r2=g2=b2=r3=g3=b3=r4=g4=b4=0; // keep compiler happy
	GetAverageViewportColors(r,g,b,r2,g2,b2,r3,g3,b3,r4,g4,b4);
	g++;
	b+=2;
	g2++;
	b2+=2;
	g3++;
	b3+=2;
	g4++;
	b4+=2;
	max = max(r,max(g,min(255,b)));
	min = min(r,min(g,b));
	r = 64 + (r - min) * 64 / (max-min);
	g = 64 + (g - min) * 64 / (max-min);
	b = 64 + (b - min) * 64 / (max-min);
	max = max(r2,max(g2,min(255,b2)));
	min = min(r2,min(g2,b2));
	r2 = 64 + (r2 - min) * 64 / (max-min);
	g2 = 64 + (g2 - min) * 64 / (max-min);
	b2 = 64 + (b2 - min) * 64 / (max-min);
	max = max(r3,max(g3,min(255,b3)));
	min = min(r3,min(g3,b3));
	r3 = 64 + (r3 - min) * 64 / (max-min);
	g3 = 64 + (g3 - min) * 64 / (max-min);
	b3 = 64 + (b3 - min) * 64 / (max-min);
	max = max(r4,max(g4,min(255,b4)));
	min = min(r4,min(g4,b4));
	r4 = 64 + (r4 - min) * 64 / (max-min);
	g4 = 64 + (g4 - min) * 64 / (max-min);
	b4 = 64 + (b4 - min) * 64 / (max-min);
	DrawBox(DRAW_PLAIN|DRAWOP_DITHER,0,0,GameX.win_width,GameX.win_height,r,g,b,a,r2,g2,b2,a,r3,g3,b3,a,r4,g4,b4,a);
}

// blurs the target image or the viewport by scaling down then up
// amount of 0.0f means not blurred at all, 0.5f means somewhat blurred, 1.0f means blurred all the way into a single color
void WindowsDX::Blur(LPDIRECTDRAWSURFACE7 surface, RECT target_rect, float blur)
{
	DrawClose();

	int width = (target_rect.right-target_rect.left);
	int height = (target_rect.bottom-target_rect.top);
	RECT mini_rect = {target_rect.left+(width*(1-blur))/2, target_rect.top+(height*(1-blur))/2, target_rect.left+(width*(1+blur))/2, target_rect.top+(height*(1+blur))/2} ;
	if(mini_rect.right == mini_rect.left) mini_rect.left--;
	if(mini_rect.bottom == mini_rect.top) mini_rect.top--;

	DrawImage2D(surface, surface, target_rect, mini_rect, 1);
	DrawImage2D(surface, surface, mini_rect, target_rect, 1);
}


void WindowsDX::Blur(ImageX * img, int x1, int y1, int x2, int y2, float amount)
{
	if(amount <= 0.01f) return;
//	if(amount >= 0.99f) amount = 0.99f;
	amount = (1.0f - amount) * (1.0f - amount); // invert about 0.5 and adjust amount toward 0
	LPDIRECTDRAWSURFACE7 surface;
	if(img == VIEWPORT) surface = dd_back_buf;
	else					surface = img->m_surface;
	if(x2==DEFAULT) {
		if(img == VIEWPORT) {
			x2 = win_width;
			y2 = win_height;
		} else {
			x2 = img->m_xres;
			y2 = img->m_yres;
		}
	}
	RECT target_rect = {x1,y1,x2,y2};
	Blur(surface,target_rect,amount); // calls the above function
}


int WindowsDX::DrawText (int x, int y, char *msg)
{
	GameX.DrawOpen() ;

	HDC hdc = NULL;
	if (FAILED(dd_back_buf->GetDC(&hdc))) { // Get the device context handle.
		#ifdef WINDX_DEBUG
			debug.Output ("WinDX::DrawText: Unable to get device context handle.\n", "") ;
		#endif	
		return FALSE; 
	}
										// Write the message.
	SetBkMode(hdc, TRANSPARENT) ;
	SetTextColor(hdc, win_rgb) ;
	TextOut(hdc, x, y, msg, strlen(msg)) ; // just use the Windows function for now, although it's slow

	dd_back_buf->ReleaseDC(hdc) ;		// Release the device context.
	return TRUE;
}

int WindowsDX::DrawOutlinedText (int x, int y, int br, int bg, int bb, char *msg)
{
	GameX.DrawOpen() ;

	HDC hdc = NULL;
	if (FAILED(dd_back_buf->GetDC(&hdc))) { // Get the device context handle.
		#ifdef WINDX_DEBUG
			debug.Output ("WinDX::DrawOutlinedText: Unable to get device context handle.\n", "") ;
		#endif	
		return FALSE; 
	}
										// Write the message.
	SetBkMode(hdc, TRANSPARENT) ;

	// Draw an outline around the text
	const int dir[8] = {1,1,1,0,-1,-1,-1,0}; // direction table for outline
	SetTextColor(hdc, RGB(br,bg,bb)) ;
	for (int i=0 ; i<8 ; i++)
		TextOut(hdc, x+dir[i], y+dir[(i+6)%8], msg, strlen(msg));

	// Now draw the text normally
	SetTextColor(hdc, win_rgb) ;
	TextOut(hdc, x, y, msg, strlen(msg)) ;

	dd_back_buf->ReleaseDC(hdc) ;		// Release the device context.
	return TRUE;
}

// draws a polygon the slow and easy way, using the windows API
// not yet tested, should not really be used in final implementation anyway
// since that's what Direct3D should be used for.
void WindowsDX::DrawPolygon (CONST POINT * points, int num_points)
{
	GameX.DrawOpen() ;

	HDC hdc = NULL;
	if (FAILED(dd_back_buf->GetDC(&hdc))) { // Get the device context handle.
		#ifdef WINDX_DEBUG
			debug.Output ("WinDX::DrawPolygon: Unable to get device context handle.\n", "") ;
		#endif	
		return; 
	}

	Polygon(hdc, points, num_points) ; // just use the Windows function for now, although it's slow

	dd_back_buf->ReleaseDC(hdc) ;		// Release the device context.
}


// draws an image tiled over the entire screen,
// good for things such as drawing backgrounds
int WindowsDX::DrawBackground (int flags, ImageX * img, int translation_x, int translation_y, int rotation_center_x, int rotation_center_y, int r, int g, int b, int a, float angle, float scale)
{
	if(win_lowquality) {
		if(flags & DRAWOP_SKIPPABLE || flags & DRAWOP_ALTERNABLE && win_alternator)
			return 1;

		angle = 0.0f; // not rotating the background can speed it up a lot too
	}

	int xres, yres;
	int map_x1, map_y1;
	int map_x2, map_y2; 
	int map_numx, map_numy; 
	int off_x, off_y;
	int img_x, img_y;	
	int mx, my;

	xres = img->m_xres*scale; yres = img->m_yres*scale;
	if(xres < 1 || yres < 1) return 0;

	ClipTile (translation_x+win_width, translation_y+win_height, xres, yres, map_x2, map_y2, off_x, off_y) ;
	ClipTile (translation_x, translation_y, xres, yres, map_x1, map_y1, off_x, off_y) ;

	map_numx = map_x2 - map_x1 + 1;
	map_numy = map_y2 - map_y1 + 1;

	if(angle == 0.0f) {

		for (my = 0; my < map_numy; my++) {
			img_y = my * yres - off_y;
			img_x = -off_x;
			for (mx = 0; mx < map_numx; mx++) {
				DrawImage (flags, img, img_x, img_y, VIEWPORT, r,g,b,a, 0.0f, scale) ;
				img_x += xres;
			}
		}

	} else {

#ifdef _DEBUG
	if(img->m_usage & LOAD_NON3D) {
		MessageBox (win_hwnd, "You should not load a background image (or any image) as 2D (with the LOAD_NON3D flag) unless\nyou do not want it rotated (or scaled or shaded).", "GameX Notice", MB_OK|MB_ICONWARNING) ;
	}
#endif

		float sinAngle, cosAngle, xFixFactor, yFixFactor;

		sinAngle = sinf((360-angle)*DEGtoRAD);
		cosAngle = cosf((360-angle)*DEGtoRAD);

		xFixFactor = max(0.5f,(float)(xres-2)/xres); // having trouble getting the tiles to join together seamlessly,
		yFixFactor = max(0.5f,(float)(yres-2)/yres); // temporary fix for now: always have an extra overlap up to 2 pixels

		map_numx *= 3;
		map_numy *= 3;

		float xres_extent = xres*sqrt(2);
		float yres_extent = yres*sqrt(2);

		for (my = -map_numy; my < map_numy; my++) {
			img_y = my * yres;
			img_x = -map_numx * xres;
			for (mx = -map_numx; mx < map_numx; mx++) {
				int real_x, real_y, x_offset, y_offset;
				x_offset = rotation_center_x - xres/2;
				y_offset = rotation_center_y - yres/2;
				img_x -= x_offset;
				img_y -= y_offset;
				real_x = img_x * cosAngle * xFixFactor - img_y * sinAngle * yFixFactor - translation_x;
				real_y = img_x * sinAngle * xFixFactor + img_y * cosAngle * yFixFactor - translation_y;
				real_x += x_offset;
				real_y += y_offset;
				img_x += x_offset;
				img_y += y_offset;
				if(real_x > -xres_extent && real_x < win_width+xres_extent
				&& real_y > -yres_extent && real_y < win_height+yres_extent) {
					DrawImage (flags, img, real_x, real_y, VIEWPORT, r,g,b,a, angle, scale) ;
				}
				img_x += xres;
			}
		}
	}
	return 1;
}

// this function is very slow if compiler optimizations are off (i.e. in debug mode)
// because it is made to be inlined.
static void DrawPixel16 (int x, int y, int r, int g, int b) // r, g, and b are out of 255
{
	// hmm... to support 15 bit mode, may need to change 2 to 3, i.e. store the 2s and 3s in variables determined by screen depth
	((unsigned short *)GameX.win_screen)[x+GameX.win_ytable[y]] = (((r >> 3) << win_shiftr) + ((g >> 2) << win_shiftg) + ((b >> 3) << win_shiftb));
}

// this function is pretty slow, if you need to use it a lot in real-time consider using DrawAdded() instead,
// or writing a custom drawing function that is more optimized than this general one.
static void DrawPixelAdded16 (int x, int y, int ri, int gi, int bi) // ri, gi, and bi are -255 for full subtract, 255 for full add
{
	// too big to be inline so might as well do some bounds checking...
	if(!(ri||gi||bi) || y<0 || x<0 || y>=GameX.win_height || x>=GameX.win_width) return;
	unsigned short * view_pix = ((unsigned short *)GameX.win_screen + x + GameX.win_ytable[y]);
	int v = *view_pix;
	int blue =	((v & win_maskb) + ((31 * bi / 255) << win_shiftb)) ;
	int green = ((v & win_maskg) + ((63 * gi / 255) << win_shiftg)) ;
	int red =	((v & win_maskr) + ((31 * ri / 255) << win_shiftr)) ;
	*view_pix = ((bi < 0) ? max(0,blue)  : min(win_maskb, blue))
			  | ((gi < 0) ? max(0,green) : min(win_maskg, green))
			  | ((ri < 0) ? max(0,red)	 : min(win_maskr, red)) ;
}



// the Draw function that is finally called to draw things -- assumes all values have been converted from DEFAULT by now
// can be called directly but probably shouldn't, that's what the ton of overloaded inline Draw functions are for
void WindowsDX::DrawImageMaster(int flags, int x1s,int y1s,int x2s,int y2s, ImageX * source, int x1d,int y1d,int x2d,int y2d, ImageX * destination, int r,int g,int b,int a, int r2,int g2,int b2,int a2, int r3,int g3,int b3,int a3, int r4,int g4,int b4,int a4, float angle, float scalex, float scaley)
{

#ifdef _DEBUG
	bool error = false;
	char str[500];

	if(!(flags & DRAWMODESMASK)) {
		sprintf(str,"%s","You must specify one of the following flags when drawing an image:\nDRAW_PLAIN  DRAW_ADD  DRAW_SUBTRACT  DRAW_GHOST  DRAW_INVERT  DRAW_BURN  DRAW_INTENSIFY  DRAW_MULTIPLY");
		error = true;
	} else if(!!(flags & DRAW_PLAIN)  + !!(flags & DRAW_ADD)  + !!(flags & DRAW_SUBTRACT)	 + !!(flags & DRAW_GHOST)
			+ !!(flags & DRAW_INVERT) + !!(flags & DRAW_BURN) + !!(flags & DRAW_INTENSIFY) + !!(flags & DRAW_MULTIPLY) > 1) {
		sprintf(str,"Attempted to combine flags that cannot be combined when drawing: %s%s%s%s%s%s%s%s",
			(flags			& DRAW_PLAIN)		? " DRAW_PLAIN" 	: "",
			(flags			& DRAW_ADD) 		? " DRAW_ADD"		: "",
			(flags			& DRAW_SUBTRACT)	? " DRAW_SUBTRACT"	: "",
			(flags			& DRAW_GHOST)		? " DRAW_GHOST" 	: "",
			(flags			& DRAW_INVERT)		? " DRAW_INVERT"	: "",
			(flags			& DRAW_BURN)		? " DRAW_BURN"		: "",
			(flags			& DRAW_INTENSIFY) ? " DRAW_INTENSIFY": "",
			(flags			& DRAW_MULTIPLY)	? " DRAW_MULTIPLY"	: "");
		error = true;
	} else if(destination != VIEWPORT && !(destination->m_usage & LOAD_TARGETABLE)) {
		sprintf(str,"Destination image was loaded with: %s%s%s%s%s%s\nBut you have to specify the LOAD_TARGETABLE flag when loading an image that you want to draw into.",
			(destination->m_usage	& LOAD_ALPHA)		? " LOAD_ALPHA" 	: "",
			(destination->m_usage	& LOAD_MASKED)		? " LOAD_MASKED"	: "",
			(destination->m_usage	& LOAD_NON3D)		? " LOAD_NON3D" 	: "",
//			(destination->m_usage	& LOAD_DITHER)		? " LOAD_DITHER"	: "",
//			(destination->m_usage	& LOAD_OPTIMIZED)	? " LOAD_OPTIMIZED" : "",
			(destination->m_usage	& LOADMODESMASK)	? ""				: "NO LOAD FLAGS");
		error = true;
	} else if(flags & DRAWOP_CENTERED && x2s != DEFAULT2 && x2d != DEFAULT2) {
		sprintf(str,"%s"," The DRAWOP_CENTERED flag should only be specified when drawing an image with\n either a source or destination position that is determined by a point (2 parameters, x and y)\n as opposed to a rectangle (4 parameters) or the entire source or destination (0 parameters) for both.");
		error = true;
	}
	if(error) {
		char error_message[1000];
		char * source_name = (source == VIEWPORT) ? "VIEWPORT" : ((source->m_filename != NULL) ? source->m_filename : "an ImageX");
		char * dest_name = (destination == VIEWPORT) ? "VIEWPORT" : ((destination->m_filename != NULL) ? destination->m_filename : "an ImageX");
		sprintf(error_message,"Error when drawing from %s into %s:\n\n%s",source_name,dest_name,str);
		MessageBox (win_hwnd, error_message, "GameX Error", MB_OK|MB_ICONSTOP) ;
		return;
	}
#endif

	if(win_lowquality) { 
		if(flags & DRAWOP_SKIPPABLE || flags & DRAWOP_ALTERNABLE && win_alternator)
			return;
		if(flags & DRAWOP_PLAINABLE) {
//			if(flags & DRAWOP_PLAIN && source != VIEWPORT) {
//				if(source->m_usage & LOAD_ALPHA) {
//					flags = (flags & DRAWOPTIONSMASK) | DRAW_PLAIN | DRAWOP_ALPHATOMASK;
//				}
//			} else {
				flags = (flags & DRAWOPTIONSMASK) | DRAW_PLAIN | DRAWOP_NOFILTER;
//			}
			if(flags & DRAWOP_DITHER) flags ^= DRAWOP_DITHER;
			if(a||a2||a3||a4)
				a=a2=a3=a4=255;
			r=g=b=r2=g2=b2=r3=g3=b3=r4=g4=b4=255;
		}
	}

	if(x2s == DEFAULT) { // default value for entire source
		if(source == VIEWPORT)
			x2s = win_width;
		else
			x2s = source->m_xres;
		if(source == VIEWPORT)
			y2s = win_height;
		else
			y2s = source->m_yres;
	}
	if(x2d == DEFAULT) { // default value for entire destination

		if(flags & DRAWOP_CENTERED) {	// if DRAWOP_CENTERED, assume they meant to center the image in the destination
			x2d = DEFAULT2; 	// instead of stretching it to fill the entire destination
			if(destination == VIEWPORT) {
				x1d = win_width/2;
				y1d = win_height/2;
			} else {
				x1d = destination->m_xres/2;
				y1d = destination->m_yres/2;
			}
		}
		else {
			if(destination == VIEWPORT)
				x2d = win_width;
			else
				x2d = destination->m_xres;
			if(destination == VIEWPORT)
				y2d = win_height;
			else
				y2d = destination->m_yres;
		}
	}
	if(scalex == DEFAULT)
		scalex = scaley = 1.0f;
	else {
		if(scaley == DEFAULT)
			scaley = scalex;
		if(flags & DRAWOP_SCALECANFLIP) {
			if(scalex == 0.0f || scaley == 0.0f) return;
			if(scalex < 0.0f) {
				scalex *= -1;
				flags ^= DRAWOP_HMIRROR;
			}
			if(scaley < 0.0f) {
				scaley *= -1;
				flags ^= DRAWOP_VMIRROR;
			}
		} else
			if(scalex <= 0.0f || scaley <= 0.0f)
				return;
	}
	if(x2s == DEFAULT2) { // default value that means we have to convert a point into a rectangle
		if(scalex == 1.0f) {
			x2s = x1s + (x2d - x1d);
		} else {
			x2s = x1s + (int)(scalex*(x2d - x1d));
		}
		if(scaley == 1.0f) {
			y2s = y1s + (y2d - y1d);
		} else {
			y2s = y1s + (int)(scaley*(y2d - y1d));
		}
		if(flags & DRAWOP_CENTERED) { // if the point is the center, not the top-left corner, of the source
			int halfWidth = (x2s - x1s) >> 1;
			int halfHeight = (y2s - y1s) >> 1;
			x1s -= halfWidth;
			x2s -= halfWidth;
			y1s -= halfHeight;
			y2s -= halfHeight;
			flags ^= DRAWOP_CENTERED;
		}
	}
	if(x2d == DEFAULT2) { // default value that means we have to convert a point into a rectangle
		if(scalex == 1.0f) {
			x2d = x1d + (x2s - x1s);
		} else {
			x2d = x1d + (int)(scalex*(x2s - x1s));
		}
		if(scaley == 1.0f) {
			y2d = y1d + (y2s - y1s);
		} else {
			y2d = y1d + (int)(scaley*(y2s - y1s));
		}
		if(flags & DRAWOP_CENTERED) { // if the point is the center, not the top-left corner, of the destination
			int halfWidth = (x2d - x1d) >> 1;
			int halfHeight = (y2d - y1d) >> 1;
			x1d -= halfWidth;
			x2d -= halfWidth;
			y1d -= halfHeight;
			y2d -= halfHeight;
			flags ^= DRAWOP_CENTERED;
		}
	}

	if(r == DEFAULT) { // default value that means we have to set a color that makes sense for the drawing mode
		r = g = b = a = 255;
	} else {
		// cap the r,g,b values to avoid unintentionally strange image color behaviour
		// this is only for the parameters, not for each pixel (which is capped either way)
			if(r < 0)	r = 0;
	   else if(r > 255) r = 255;
			if(g < 0)	g = 0;
	   else if(g > 255) g = 255;
			if(b < 0)	b = 0;
	   else if(b > 255) b = 255;
			if(a < 0)	a = 0;
	   else if(a > 255) a = 255;
	}
	if(r2 == DEFAULT) {
		r4 = r3 = r2 = r;
		g4 = g3 = g2 = g;
		b4 = b3 = b2 = b;
		a4 = a3 = a2 = a;
	} else {
			if(r2 < 0)	 r2 = 0;
	   else if(r2 > 255) r2 = 255;
			if(g2 < 0)	 g2 = 0;
	   else if(g2 > 255) g2 = 255;
			if(b2 < 0)	 b2 = 0;
	   else if(b2 > 255) b2 = 255;
			if(a2 < 0)	 a2 = 0;
	   else if(a2 > 255) a2 = 255;
			if(r3 < 0)	 r3 = 0;
	   else if(r3 > 255) r3 = 255;
			if(g3 < 0)	 g3 = 0;
	   else if(g3 > 255) g3 = 255;
			if(b3 < 0)	 b3 = 0;
	   else if(b3 > 255) b3 = 255;
			if(a3 < 0)	 a3 = 0;
	   else if(a3 > 255) a3 = 255;
			if(r4 < 0)	 r4 = 0;
	   else if(r4 > 255) r4 = 255;
			if(g4 < 0)	 g4 = 0;
	   else if(g4 > 255) g4 = 255;
			if(b4 < 0)	 b4 = 0;
	   else if(b4 > 255) b4 = 255;
			if(a4 < 0)	 a4 = 0;
	   else if(a4 > 255) a4 = 255;
	}
	if(flags & DRAW_ADD || flags & DRAW_SUBTRACT || flags & DRAW_GHOST || flags & DRAW_INVERT || flags & DRAW_BURN || flags & DRAW_INTENSIFY) {
		// scale the luminosity by the alpha which would otherwise be ignored
		ScaleByAlpha(r,g,b,a,r2,g2,b2,a2,r3,g3,b3,a3,r4,g4,b4,a4);
//		if(win_lowquality && r+r2+r3+r4<256
//		&& g+g2+g3+g4<256 && b+b2+b3+b4<256) return; // time-conservative return for when there's no video acceleration
//		if(win_lowquality && r<64&&r2<64&&r3<64&&r4<64
//		&& g<64&&g2<64&&g3<64&&g4<64 && b<64&&b2<64&&b3<64&&b4<64) return; // time-conservative return for when there's no video acceleration
		if(!(r||g||b||r2||g2||b2||r3||g3||b3||r4||g4||b4))
			return; // return if there isn't any luminosity
	} else {
//		if(win_lowquality && a+a2+a3+a4<96*4) return; // time-conservative return for when there's no video acceleration
//		if(win_lowquality && a<96&&a2<96&&a3<96&&a4<96) return; // time-conservative return for when there's no video acceleration
		if(!(a||a2||a3||a4))
			return; // return if there isn't any opacity
	}

	
	int sourceHasSurface, destHasSurface;
	bool sourceHasData, sourceHasAlpha, destHasData, destHasAlpha, destHasDevice;
	LPDIRECTDRAWSURFACE7 p_src_surface, p_dst_surface;
	LPDIRECT3DDEVICE7 p_dst_device;
	XBYTE * p_src_data;
	XBYTE * p_src_alpha;
	XBYTE * p_dst_data;
	XBYTE * p_dst_alpha;
	RECT src_bounds = {0,0,0,0};
	RECT dst_bounds = {0,0,0,0};
	int src_pitch, dst_pitch;

	if(source == VIEWPORT) {
		sourceHasSurface = true;
		p_src_surface = dd_back_buf;
		sourceHasAlpha = false;
//		src_pitch = win_adjustedpitch;
		src_pitch = win_pitch * 8 / win_screenbpp; // the adjustment here has to be inverted
		src_bounds.right = win_width;
		src_bounds.bottom = win_height;
	} else {

		if(source->m_hassurface) {
			sourceHasSurface = source->m_hassurface;
			p_src_surface = source->m_surface;
			if(GameX.WasLost()) source->Restore();
		}
		else sourceHasSurface = 0;

		if(source->m_data != NULL) {
			sourceHasData = true;
			p_src_data = source->m_data;
		}
		else
			sourceHasData = false;

		if(source->m_alpha != NULL) {
			sourceHasAlpha = true;
			p_src_alpha = source->m_alpha;
		}
		else
			sourceHasAlpha = false;

		src_pitch = source->m_xres; // may have to change this line to do depths other than 16 bit
		src_bounds.right = source->m_xres;
		src_bounds.bottom = source->m_yres;
	}

	if(destination == VIEWPORT) {
		destHasSurface = 3;
		p_dst_surface = dd_back_buf;

		destHasDevice = true;
		p_dst_device = d3d_device;

		destHasAlpha = false;

//		dst_pitch = win_adjustedpitch;
		dst_pitch = win_pitch * 8 / win_screenbpp; // the adjustment here has to be inverted
		dst_bounds.right = win_width;
		dst_bounds.bottom = win_height;
	} else {
		if(destination->m_hassurface) {
			destHasSurface = destination->m_hassurface;
			p_dst_surface = destination->m_surface;
			if(GameX.WasLost()) destination->Restore();
//			dst_pitch = destination->m_surf_pitch; // untested
		}
		else {
			dst_pitch = destination->m_xres; // may have to change this line to do depths other than 16 bit...maybe
			destHasSurface = 0;
		}

		if(destination->m_data != NULL) {
			destHasData = true;
			p_dst_data = destination->m_data;
		}
		else
			destHasData = false;

		if(destination->m_alpha != NULL) {
			destHasAlpha = true;
			p_dst_alpha = destination->m_alpha;
		}
		else
			destHasAlpha = false;

		if(destination->m_device != NULL) {
			destHasDevice = true;
			p_dst_device = destination->m_device;
		}
		else
			destHasDevice = false;

		dst_bounds.right = destination->m_xres;
		dst_bounds.bottom = destination->m_yres;
	}

/// if(flags & DRAW_PLAIN)
/// 	if(source != VIEWPORT && destination != VIEWPORT)
/// 		if(source->m_options & IMG_ALPHA && destination->m_options & IMG_ALPHA)
/// 			flags |= ALPHACOPY;

	int alphatype = 0;
	if(source != VIEWPORT) {
		if((source->m_usage & LOAD_MASKED))
			alphatype = 1;
		else if((source->m_usage & LOAD_ALPHA)) {
			if(destination == VIEWPORT || !(destination->m_usage & LOAD_ALPHA))
			alphatype = 2; // alpha channel in image
		}
	}

	// find out whether we can use 3D acceleration and features to draw this image,
	// which is true if the source is a texture and the destination has a 3D device
	bool can_3D_accelerate = (sourceHasSurface==2 && destHasDevice) ;
	bool can_2D_accelerate = (flags & DRAW_PLAIN && a==255 && alphatype != 2 && sourceHasSurface && destHasSurface);
	bool priority3D = can_3D_accelerate;

	// now we check to see if 3D acceleration would just slow us down
	if(can_3D_accelerate && (win_lowquality || ((destHasSurface == 2 || destHasSurface == 1)
	&& x2s - x1s == x2d - x1d
	&& y2s - y1s == y2d - y1d))
	&& angle == 0.0f // currently no support for non-3D rotation
	&& (flags & DRAW_PLAIN && alphatype != 2)
	&& r ==255 && g ==255 && b ==255 && a ==255
	&& r2==255 && g2==255 && b2==255 && a2==255
	&& r3==255 && g3==255 && b3==255 && a3==255
	&& r4==255 && g4==255 && b4==255 && a4==255
/*	&& r+g+b+a+r2+g2+b2+a2+r3+g3+b3+a3+r4+g4+b4+a4 == 255*16*/) // currently no support for non-3D coloration/transparency
		priority3D = false;

	// NOTE: support for all advanced features is not complete for if we can't draw using 3D,
	//		 but that *should* only happen when drawing FROM the back buffer to somewhere,

	RECT sourceRect = {x1s,y1s,x2s,y2s}; // temporary -- will stop using rectangles soon
	RECT destRect	= {x1d,y1d,x2d,y2d}; // since DrawImage3D can easily be modified to map images to any four-sided polygonal shape

	int ok = 0;
	if(can_3D_accelerate && priority3D) {
		// try to draw it using Direct3D (can be extremely slow if there is insufficient video hardware, in which case D3D emulates it in software)
		ok = DrawImage3D(p_dst_device, p_src_surface, sourceRect, src_bounds,destRect, dst_bounds, source->m_surf_width, source->m_surf_height, r,g,b,a, r2,g2,b2,a2, r3,g3,b3,a3, r4,g4,b4,a4, angle, flags, alphatype);
	} 
	if(!ok){ // if we couldn't draw it with Direct3D
		if(can_2D_accelerate) {
			// if we can draw it using DirectDraw's 2D acceleration
			// (this is the fastest possible way of drawing anything, but also the most limited.
			//  DirectDraw supports masks (color keys) and (somewhat slowly but often filtered) supports scaling,
			//  but cannot handle rotation or any kind of alpha/added/anything blending or coloration effects)

			if(source == VIEWPORT || destination == VIEWPORT)
				DrawClose();
			if(dd_was_lost) {
				if(source != VIEWPORT) {source->Restore(); p_src_surface = source->m_surface;}
				if(destination != VIEWPORT) {destination->Restore(); p_dst_surface = destination->m_surface;}
			}

			ok = DrawImage2D(p_src_surface, p_dst_surface, sourceRect, destRect, !alphatype);
		}
		if(!ok && !priority3D && can_3D_accelerate) {
			// if we decided to try drawing 2D before drawing 3D, and if drawing 2D failed,
			ok = DrawImage3D(p_dst_device, p_src_surface, sourceRect, src_bounds,destRect, dst_bounds, source->m_surf_width, source->m_surf_height, r,g,b,a, r2,g2,b2,a2, r3,g3,b3,a3, r4,g4,b4,a4, angle, flags, alphatype);
		} 
		if(!ok){ // if we couldn't draw it with Direct3D or DirectDraw

			if(destination == VIEWPORT) {
				if(DrawOpen()) {
					destHasData = true;
					p_dst_data = (XBYTE *) win_screen;
				} else destHasData = false;
			}

			if(source == VIEWPORT) {
				if(DrawOpen()) {
					sourceHasData = true;
					p_src_data = (XBYTE *) win_screen;
				} else sourceHasData = false;
			}

			if(sourceHasData && destHasData) {

#ifdef _DEBUG
	if(!win_no_acc_err) { // so we only warn once instead of every frame which would effectively freeze the game
		win_no_acc_err = true;
		if(source == VIEWPORT) {
			MessageBox (win_hwnd, "Warning: The VIEWPORT is being drawn without the help of DirectDraw or Direct3D.\nThis can cause the game to perform extremely poorly.\nTo correct this issue, make sure to only draw the VIEWPORT into an image that is loaded with the LOAD_NON3D flag,\nand only draw the VIEWPORT with DRAW_PLAIN. If you need to draw it with a special mode such as DRAW_ADD,\ncopy it from the NON3D image into a normal image and draw from that.", "GameX Warning", MB_OK|MB_ICONWARNING) ;
		} else {
			MessageBox (win_hwnd, "Warning: An image is being drawn without the help of DirectDraw or Direct3D.\nThis can cause the game to perform extremely poorly.\nPerhaps you are using the LOAD_NON3D flag when it should not be used.", "GameX Warning", MB_OK|MB_ICONWARNING) ;
		}
		if(win_screenbpp != 16)
			MessageBox (win_hwnd, "Warning: Drawing images without 2D or 3D acceleration\nis currently only supported in 16-bit (HIGHCOLOR) display mode.", "GameX Warning", MB_OK|MB_ICONWARNING) ;
	}
#endif

	// the following code is the slow, hopefully rarely-used backup code
	
				if(flags & DRAW_PLAIN) {
					if(alphatype != 2) { // non-alpha or mask
						if(a!=255)
							DrawBlended(p_src_data, p_dst_data, sourceRect, destRect, src_bounds, dst_bounds, src_pitch,dst_pitch, r,g,b,a, !alphatype);
						else
							DrawPlain(p_src_data, p_dst_data, sourceRect, destRect, src_bounds, dst_bounds, src_pitch,dst_pitch, !alphatype);
					} else { // draw using image's alpha channel
						if(sourceHasAlpha)
							DrawAlphaUnmerged(p_src_data, p_src_alpha, p_dst_data, sourceRect, destRect, src_bounds, dst_bounds, src_pitch,dst_pitch, a);
						else
							DrawAlphaMerged(p_src_data, p_dst_data, sourceRect, destRect, src_bounds, dst_bounds, src_pitch,dst_pitch, a);
					}
				}
				else if(flags & DRAW_ADD) {
					if(alphatype != 2) { // non-alpha
						DrawAdded(p_src_data, p_dst_data, sourceRect, destRect, src_bounds, dst_bounds, src_pitch,dst_pitch, r,g,b);
					} else { // draw added AND take alpha channel into account
						DrawAddedMerged(p_src_data, p_dst_data, sourceRect, destRect, src_bounds, dst_bounds, src_pitch,dst_pitch, r,g,b);
					}
				}
				else if(flags & DRAW_SUBTRACT) {
					DrawAdded(p_src_data, p_dst_data, sourceRect, destRect, src_bounds, dst_bounds, src_pitch,dst_pitch, -r,-g,-b);
				}
				else if(flags & DRAW_INVERT) {
					DrawInvert(p_src_data, p_dst_data, sourceRect, destRect, src_bounds, dst_bounds, src_pitch,dst_pitch, r,g,b);
				}
				else if(flags & DRAW_GHOST) {
					// not yet fully supported without D3D, this is just a bad approximation:
					DrawAdded(p_src_data, p_dst_data, sourceRect, destRect, src_bounds, dst_bounds, src_pitch,dst_pitch, -r,-g,-b);
					DrawAdded(p_src_data, p_dst_data, sourceRect, destRect, src_bounds, dst_bounds, src_pitch,dst_pitch, r,g,b);
				}
				else if(flags & DRAW_INTENSIFY) {
					// not yet fully supported without D3D, this is just a bad approximation:
					DrawAdded(p_src_data, p_dst_data, sourceRect, destRect, src_bounds, dst_bounds, src_pitch,dst_pitch, r/2,g/2,b/2);
				}
				else if(flags & DRAW_BURN) {
					// not yet fully supported without D3D, this is just a bad approximation:
					DrawAdded(p_src_data, p_dst_data, sourceRect, destRect, src_bounds, dst_bounds, src_pitch,dst_pitch, (r-255)/2,(g-255)/2,(b-255)/2);
					DrawAdded(p_src_data, p_dst_data, sourceRect, destRect, src_bounds, dst_bounds, src_pitch,dst_pitch, r/2,g/2,b/2);
				}

			}
		}
	}

	if(destination != VIEWPORT && destHasSurface == 2) {
		// if the destination is a texture, we have to tell D3D to refresh it by
		// locking and unlocking it, or else the image won't change after the first time it's drawn.
		// note that this makes drawing to somewhere else than the screen even slower... oh well
		DDSURFACEDESC2 ddsd;
		ddsd.dwSize = sizeof(ddsd);
		p_dst_surface->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
		p_dst_surface->Unlock(NULL);
	}

	DrawClose();
}


// Draws a 2D image using 3D acceleration and features
int WindowsDX::DrawImage3D(LPDIRECT3DDEVICE7 device, LPDIRECTDRAWSURFACE7 image, RECT & src_rect, RECT & src_bounds, RECT & dst_rect, RECT & dst_bounds, int surf_width, int surf_height, int r, int g, int b, int a, int r2, int g2, int b2, int a2, int r3, int g3, int b3, int a3, int r4, int g4, int b4, int a4, float angle, int flags, int alphatype)
{
	if(device == NULL) return 0;

	DrawClose();

	// create the vertices for a rectangle with points ordered like so:
	//
	//	0----1			  
	//	.	/.		 
	//	.  / .		 
	//	. /  .		   
	//	./	 .				
	//	2----3				  
	//
	// to allow for D3D to use it for a triangle strip
	
	D3DTLVERTEX quad1[4];
	
	float x[4] = {dst_rect.left, dst_rect.right, dst_rect.left, dst_rect.right};
	float y[4] = {dst_rect.top, dst_rect.top, dst_rect.bottom, dst_rect.bottom};
	
	while(angle >= 360.0f) angle -= 360.0f;
	while(angle < 0.0f)   angle += 360.0f;
	int rotate;
	if(angle==	0.0f || angle== 45.0f || angle== 90.0f || angle==135.0f
	|| angle==180.0f || angle==225.0f || angle==270.0f || angle==315.0f)
		rotate = 0;
	else
		rotate = 1;

	float x_adjust, y_adjust;

	if(src_rect.right - src_rect.left == src_bounds.right - src_bounds.left) {
		x_adjust = 0.0f;
	} else {
		float scalex = (float)(dst_rect.right-dst_rect.left)/(src_rect.right-src_rect.left);
		x_adjust = ((scalex <= 1.0f) ? (0.0f) : (.525f-0.5f/scalex));
		if(rotate) x_adjust = min(0.5f, x_adjust+0.25f);
	}
	if(src_rect.bottom - src_rect.top == src_bounds.bottom - src_bounds.top) {
		y_adjust = 0.0f;
	} else {
		float scaley = (float)(dst_rect.bottom-dst_rect.top)/(src_rect.bottom-src_rect.top);
		y_adjust = ((scaley <= 1.0f) ? (0.0f) : (.525f-0.5f/scaley));
		if(rotate) y_adjust = min(0.5f, y_adjust+0.25f);
	}
	if(angle == 0.0f) // no rotation at all
		rotate = 0;
	else if(angle == 180.0f) { // exactly 180 degrees
		flags ^= DRAWOP_HMIRROR;	   // same as mirroring across x and y
		flags ^= DRAWOP_VMIRROR;
		rotate = 1;
	}
	else {
		float x_offset = (dst_rect.right + dst_rect.left) / 2;
		float y_offset = (dst_rect.bottom + dst_rect.top) / 2;
		angle = 360.0f - angle; // convert to clockwise
		float cosAngle = cosf(angle*DEGtoRAD);
		float sinAngle = sinf(angle*DEGtoRAD);

		// rotate the image by simply rotating the four points
		// of the rectangular polygon that the image is being mapped onto
		// like so:
		//				   >			  0
		//				  / 			/\				
		//	0----1		 0----1 	   /  \ 			
		//	|	 |		 |	  |\    2 /    \	   
		//	|	 |		 |	  | v	 /		\	   
		//	|	 |	   ^ |	  | 	 \		/		 
		//	|	 |		\|	  | 	  \    / 1 		   
		//	2----3	->	 2----3   ->   \  / 				
		//					 /			\/				  
		//					<		   3	 

		for(int i = 0 ; i < 4 ; i++) {
			x[i] -= x_offset;
			y[i] -= y_offset;
			float xr = x[i]*cosAngle - y[i]*sinAngle;
			float yr = x[i]*sinAngle + y[i]*cosAngle;
			x[i] = xr + x_offset;
			y[i] = yr + y_offset;

/// 		if(abs(x[i]-dst_rect.left)<=1) x[i] = dst_rect.left;		// tiny error correction
/// 		else if(abs(x[i]-dst_rect.right)<=1) x[i] = dst_rect.right; // so images never have an extra pixel sticking out
/// 		if(abs(y[i]-dst_rect.top)<=1) y[i] = dst_rect.top;
/// 		else if(abs(y[i]-dst_rect.bottom)<=1) y[i] = dst_rect.bottom;
		}
		rotate = 2;
	}
	
	// D3D's automatic clipping doesn't usually seem to work right,
	// it clips but it takes just as long as if it draws everything even if completely off the screen
	// so, at least catch the things that are completely out of bounds and skip drawing them:
	bool inside = false;
	for(int i = 0 ; i < 4 ; i++) {
		if(x[i] >= dst_bounds.left && x[i] <= dst_bounds.right && y[i] >= dst_bounds.top && y[i] <= dst_bounds.bottom) {
			inside = true;
			break;
		}
	}
	if(!inside) {
		// none of the points are in bounds, but that doesn't necessarily mean it's safe to skip drawing it...
		// we'll just do a simple check to see if all of the ones that were out of bounds
		// were on the same side, in which case they couldn't possibly intersect the boundary:
		
		int i;
		bool allout;

		allout = true;
		for(i = 0 ; i < 4 ; i++) {
			if(x[i] >= dst_bounds.left) {
				allout = false;
				break;
			}
		}
		if(allout) return 1; // all points are too far left; don't draw

		allout = true;
		for(i = 0 ; i < 4 ; i++) {
			if(y[i] >= dst_bounds.top) {
				allout = false;
				break;
			}
		}
		if(allout) return 1; // all points are too far up; don't draw

		allout = true;
		for(i = 0 ; i < 4 ; i++) {
			if(x[i] <= dst_bounds.right) {
				allout = false;
				break;
			}
		}
		if(allout) return 1; // all points are too far right; don't draw

		allout = true;
		for(i = 0 ; i < 4 ; i++) {
			if(y[i] <= dst_bounds.bottom) {
				allout = false;
				break;
			}
		}
		if(allout) return 1; // all points are too far down; don't draw
		
		// if we got here, it *probably* needs to be drawn...
		// here is the only no-draw case we may have missed:
		//
		//			polygon (with image on it)
		//			  /\
		//			 /	\
		//			/	 \
		//		   <	  \
		//		   .\	   >
		//		   . \	  /
		// +--------+ \  /
		// |		|..\/
		// |		|
		// |		|
		// |		|
		// +--------+
		// screen (or other destination boundary)
		// 
		// we could do boundary intersection checks at this point but it wouldn't be worth it.
		// since all completely out of bound images would have to go through the check
		// every single time they're drawn, and the time used for intersection checks could add up fast
	}

	// respond to mirroring request flags (note that the mirroring happens BEFORE any rotation)
	if(flags & DRAWOP_VMIRROR) { // flip the top and bottom mapping values if we want to mirror the image vertically
		int temp = src_rect.top;
		src_rect.top = src_rect.bottom;
		src_rect.bottom = temp;
		y_adjust = -y_adjust;
	}
	if(flags & DRAWOP_HMIRROR) { // flip the left and right mapping values if we want to mirror the image horizontally
		int temp = src_rect.left;
		src_rect.left = src_rect.right;
		src_rect.right = temp;
		x_adjust = -x_adjust;
	}

	float leftMap	= (src_rect.left   + x_adjust) / surf_width;  // determine source mapping values
	float topMap	= (src_rect.top    + y_adjust) / surf_height; // (the 0.5 pixel adjustments fix an
	float rightMap	= (src_rect.right  - x_adjust) / surf_width;  //  annoying "sometimes off by one" bug)
	float bottomMap = (src_rect.bottom - y_adjust) / surf_height;

	// vertex structure: {x, y, zbuff, zcorrection, color{r,g,b,a}, specular, texture_u, texture_v}
	// to extend this to 3D, make zcorrection decrease approaching 0 as the vertex's distance approaches infinity -- it can also be used for weird warping effects
	D3DTLVERTEX vertex0 = {x[0]-0.5f,y[0]-0.5f,0.0,1.0,RGBA_MAKE(r ,g ,b ,a ),0,leftMap,topMap};
	D3DTLVERTEX vertex1 = {x[1]-0.5f,y[1]-0.5f,0.0,1.0,RGBA_MAKE(r2,g2,b2,a2),0,rightMap,topMap};
	D3DTLVERTEX vertex2 = {x[2]-0.5f,y[2]-0.5f,0.0,1.0,RGBA_MAKE(r3,g3,b3,a3),0,leftMap,bottomMap};
	D3DTLVERTEX vertex3 = {x[3]-0.5f,y[3]-0.5f,0.0,1.0,RGBA_MAKE(r4,g4,b4,a4),0,rightMap,bottomMap};

	quad1[0]=vertex0;
	quad1[1]=vertex1;
	quad1[2]=vertex2;
	quad1[3]=vertex3;

	int color_shift = 0; // amount to left-shift brightness

	if(flags & DRAWOP_BRIGHT)
		color_shift++; // multiply by 2 if BRIGHT

	device->SetTextureStageState(0,D3DTSS_ALPHAOP, D3DTOP_MODULATE);

	if(flags & DRAW_ADD) { // additive -- r, g, b at 255 is full additive, reduce r for cyan-tinted added, etc.
		device->SetRenderState(D3DRENDERSTATE_SRCBLEND,D3DBLEND_ONE); // Cs = 1
		device->SetRenderState(D3DRENDERSTATE_DESTBLEND,D3DBLEND_ONE); // Cd = 1
		// so C = source * Cs + dest * Cd
		//      = source * (1) + dest * (1)
		//      = dest + source
		if(alphatype == 2) { // use the alpha channel too if there is one
			device->SetRenderState(D3DRENDERSTATE_SRCBLEND,D3DBLEND_SRCALPHA);
		//    C = dest + source*alpha
		}
	} else if(flags & DRAW_SUBTRACT) { // subtractive -- r, g, b at 255 is full subtractive, reduce r to subtract less red, etc.
		// note: not totally accurate subtraction because this does not use DirectX 8.0
		//       so the closest thing in DX7 is this, the inverse of DRAW_INTENSIFY
		device->SetRenderState(D3DRENDERSTATE_SRCBLEND,D3DBLEND_ZERO); // Cs = 0
		device->SetRenderState(D3DRENDERSTATE_DESTBLEND,D3DBLEND_INVSRCCOLOR); // Cd = 1-source
		// so C = source * Cs + dest * Cd
		//      = source * (0) + dest * (1-source)
		//      = source - source*dest
		color_shift++; // increase effect 2x because it's usually too pale otherwise (so we don't have to render it twice)
		//    C = 2*source - 2*source*dest
		if(alphatype == 2) { // not perfect, but attempt to use alpha channel somehow
			device->SetRenderState(D3DRENDERSTATE_SRCBLEND,D3DBLEND_INVSRCALPHA);
		//    C = 2*source - 2*source*alpha - 2*source*dest
		}
	} else if(flags & DRAW_GHOST) { // color varying alpha blend -- r, g, b determine amount, produces neat auto-alpha effect
		device->SetRenderState(D3DRENDERSTATE_SRCBLEND,D3DBLEND_SRCCOLOR); // Cs = source
		device->SetRenderState(D3DRENDERSTATE_DESTBLEND,D3DBLEND_INVSRCCOLOR); // Cd = 1-source
		// so C = source * Cs + dest * Cd
		//      = source * (source) + dest * (1-source)
		//      = dest + source*source - source*dest
	} else if(flags & DRAW_INVERT) { // invert by the amount of color in the source -- r,g,b at 255 is full inversion, otherwise hard to describe
		device->SetRenderState(D3DRENDERSTATE_SRCBLEND,D3DBLEND_INVDESTCOLOR); // Cs = 1-dest
		device->SetRenderState(D3DRENDERSTATE_DESTBLEND,D3DBLEND_INVSRCCOLOR); // Cd = 1-source
		// so C = source * Cs + dest * Cd
		//      = source * (1-dest) + dest * (1-source)
		//      = dest + 2*source - 2*source*dest
	} else if(flags & DRAW_MULTIPLY) { // multiply source by destination
		device->SetRenderState(D3DRENDERSTATE_SRCBLEND,D3DBLEND_ZERO); // Cs = 0
		device->SetRenderState(D3DRENDERSTATE_DESTBLEND,D3DBLEND_SRCCOLOR); // Cd = source
		// so C = source * Cs + dest * Cd
		//      = source * (0) + dest * (source)
		//      = dest * source

        // alternate way:
///		device->SetRenderState(D3DRENDERSTATE_SRCBLEND,D3DBLEND_DESTCOLOR); // Cs = dest
///		device->SetRenderState(D3DRENDERSTATE_DESTBLEND,D3DBLEND_ZERO); // Cd = 0

		if(alphatype == 2) { // not perfect, but attempt to use alpha channel somehow
			device->SetRenderState(D3DRENDERSTATE_SRCBLEND,D3DBLEND_INVSRCALPHA);
		//    C = dest*source + source - source*alpha
		}
	} else if(flags & DRAW_BURN) { // similar to invert but peaks at midway
		device->SetRenderState(D3DRENDERSTATE_SRCBLEND,D3DBLEND_INVSRCCOLOR); // Cs = 1-source
		device->SetRenderState(D3DRENDERSTATE_DESTBLEND,D3DBLEND_INVSRCCOLOR); // Cd = 1-source
		// so C = source * Cs + dest * Cd
		//      = source * (1-source) + dest * (1-source)
		//      = source - source*source + dest - source*dest
	} else if(flags & DRAW_INTENSIFY) { // add, but darker as destination becomes darker
		device->SetRenderState(D3DRENDERSTATE_SRCBLEND,D3DBLEND_DESTCOLOR); // Cs = dest
		device->SetRenderState(D3DRENDERSTATE_DESTBLEND,D3DBLEND_ONE); // Cd = 1
		// so C = source * Cs + dest * Cd
		//      = source * (dest) + dest * (1)
		//      = dest + source*dest
		color_shift++; // increase effect 2x because it's usually too pale otherwise (so we don't have to render it twice)
		//    C = dest + 2*source*dest
		if(alphatype == 2) { // not perfect, but attempt to use alpha channel somehow
			device->SetRenderState(D3DRENDERSTATE_DESTBLEND,D3DBLEND_SRCALPHA);
		//    C = dest*alpha + 2*source*dest
		}
	} else { // DRAW_PLAIN
		device->SetRenderState(D3DRENDERSTATE_SRCBLEND,D3DBLEND_SRCALPHA); // Cs = alpha
		device->SetRenderState(D3DRENDERSTATE_DESTBLEND,D3DBLEND_INVSRCALPHA); // Cd = 1-alpha
		// so C = source * Cs + dest * Cd
		//      = source * (alpha) + dest * (1-alpha)
		//      = dest + source*alpha - dest*alpha
		if(alphatype != 2) {
			if(a == 255 && a2 == 255 && a3 == 255 && a4 == 255) {
				device->SetRenderState(D3DRENDERSTATE_SRCBLEND,D3DBLEND_ONE); // Cs = 1
				device->SetRenderState(D3DRENDERSTATE_DESTBLEND,D3DBLEND_ZERO); // Cd = 0
				// so C = source * Cs + dest * Cd
				//      = source * (1) + dest * (0)
				//      = source
			}
		}
		
	}
	if(alphatype == 1) { // if masked
		device->SetRenderState(D3DRENDERSTATE_COLORKEYENABLE,TRUE); // turn on color keying
		device->SetRenderState(D3DRENDERSTATE_COLORKEYBLENDENABLE,TRUE); // and color key blending (may do nothing)
	}

	if(flags & DRAWOP_INVERTED) { // if inverting, set to use the complement of the source texture
		device->SetTextureStageState(0,D3DTSS_COLORARG1, D3DTA_TEXTURE|D3DTA_COMPLEMENT);
	}

	switch(color_shift) {
	case 0:
		device->SetTextureStageState(0,D3DTSS_COLOROP, D3DTOP_MODULATE);
		break;
	case 1: // DRAWOP_BRIGHT
		device->SetTextureStageState(0,D3DTSS_COLOROP, D3DTOP_MODULATE2X);
		break;
	case 2:
		device->SetTextureStageState(0,D3DTSS_COLOROP, D3DTOP_MODULATE4X);
		break;
	}

/// if(flags & ALPHACOPY) {
/// 	device->SetTextureStageState(0,D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
/// }
/// device->SetTextureStageState(0, D3DTSS_MIPFILTER, 3);

	if(flags & DRAWOP_NOFILTER) {
		device->SetTextureStageState(0,D3DTSS_MAGFILTER,D3DTFG_POINT);
		device->SetTextureStageState(0,D3DTSS_MINFILTER,D3DTFG_POINT);
	} else {
		device->SetTextureStageState(0,D3DTSS_MAGFILTER,D3DTFG_LINEAR);
		device->SetTextureStageState(0,D3DTSS_MINFILTER,D3DTFG_LINEAR);
	}

	if(flags & DRAWOP_DITHER)
		device->SetRenderState(D3DRENDERSTATE_DITHERENABLE, TRUE);


	HRESULT status = D3D_OK;

	if(image != NULL) {
		// if the image isn't NULL, set it as the texture to draw.
		status = device->SetTexture(0,image);
	} else {
		// otherwise, leave it as just a colored polygon (so DrawBox() can use this function)
		status = device->SetTexture(0,NULL); 
	}
	

	if(status == D3D_OK)
		status = device->BeginScene();

	if(status == D3D_OK)
		status = device->DrawPrimitive(D3DPT_TRIANGLESTRIP, D3DFVF_TLVERTEX,quad1,4,0);

	if(status == D3D_OK)
		status = device->EndScene(); // if this fails, then GameX probably forgot to restore the 3D device from a lost state (which should not happen, that bug is now fixed in ViewUpdate())
	else
		device->EndScene(); // (always try to end the scene anyway, but don't overwrite knowledge of previous failure)

	// reset flags for next time:

	if(flags & DRAWOP_DITHER)
		device->SetRenderState(D3DRENDERSTATE_DITHERENABLE, FALSE);

	if(alphatype == 1) { // if masked
		device->SetRenderState(D3DRENDERSTATE_COLORKEYENABLE,FALSE);
		device->SetRenderState(D3DRENDERSTATE_COLORKEYBLENDENABLE,FALSE);
	}

	if(flags & DRAWOP_INVERTED) {
		device->SetTextureStageState(0,D3DTSS_COLORARG1, D3DTA_TEXTURE);
	}

	// report our success or failure to do the actual drawing:

	if(status != D3D_OK) {
		Direct3DError(status);
		DirectDrawError(status);
		return 0;
	} else {
		return 1;
	}
}

// the simplest and fastest image drawing function -- using offscreen surfaces, usually very fast
// includes an option for masked drawing too, just input nomask=0
int WindowsDX::DrawImage2D (LPDIRECTDRAWSURFACE7 p_sourceSurface, LPDIRECTDRAWSURFACE7 p_destSurface, RECT & src_rect, RECT & dst_rect, int nomask)
{
	// this function assumes:
	// both source and destination have already been UNlocked (if not, will refuse to draw anything)
	// the destination has a valid clipper attached to it (if not, won't clip and might crash)
	// (if the source and destination rectangles differ, Blt() will automatically scale the image,
	//	using hardware acceleration which will probably instantly scale it with bicubic filtering or something)

	int options = DDBLT_WAIT;
	if(!nomask) options |= DDBLT_KEYSRC;

//	DDBLTFX ddbltfx;						  // incredibly slow, just use DrawImage3D with the HMIRROR or VMIRROR flags for mirroring instead
//	ddbltfx.dwSize = sizeof(ddbltfx);		  // (for mirroring, would need to change a little more code here and hope for
//	ddbltfx.dwDDFX = DDBLTFX_MIRRORLEFTRIGHT; //  non-existent hardware acceleration features for it to work right in this function)

	HRESULT status = p_destSurface->Blt(&dst_rect, p_sourceSurface,
										&src_rect,
										options,
										NULL);
	if(status != DD_OK) {
		DirectDrawError(status);
		return 0;
	}
	return 1;
}


// includes an option for masked drawing too, just input nomask=0
int WindowsDX::DrawPlain (XBYTE * source_pix, XBYTE * dest_pix, RECT & src_rect, RECT & dst_rect, RECT & src_bounds, RECT & dst_bounds, int src_pitch, int dst_pitch, int nomask)
{
	int drawType = ClipRects(src_rect, dst_rect, src_bounds, dst_bounds);

	if(drawType == 0) return 1; // image is entirely out of bounds, so just return

//	DrawOpen() ; // buffer must be locked before 

	XBYTE2 *img_start;
	XBYTE2 *view_start;
	int img_xs = dst_rect.right-dst_rect.left;
	int img_ys = dst_rect.bottom-dst_rect.top;

	int img_offset = src_rect.top * src_pitch + src_rect.left;
	int view_offset = dst_rect.top * dst_pitch + dst_rect.left;


	if(drawType == 1) {
		// if source and destination rectangles are exactly the same size,
		// use this optimized code that assumes no scaling:

		if(nomask) {
			
			// even slightly more optimized code that assumes no mask

			img_start = (XBYTE2 *) source_pix + img_offset;
			view_start = (XBYTE2 *) dest_pix + view_offset;
			for (int y=img_ys; y--;) {			
				memcpy (view_start, img_start, img_xs<<1) ; // the	<< 1  is 16-bit specific
				view_start += dst_pitch;
				img_start += src_pitch;
			} 
		} else {

			XBYTE2 *img_now;
			XBYTE2 *view_pix;

			int src_height = src_rect.bottom-src_rect.top;
			int src_width = src_rect.right-src_rect.left;

			img_start = (XBYTE2 *) source_pix + img_offset;
			view_start = (XBYTE2 *) dest_pix + view_offset;
			for (int y=0 ; y < img_ys; y++) {

				view_pix = view_start;
				img_now = img_start;
				for (int x=0 ; x < img_xs; x++) {
					if (* img_now) * view_pix = * img_now;
					view_pix++; img_now++;
				}
				view_start += dst_pitch;
				img_start += src_pitch;
			}
		}
	} else {
		// if drawType == 2, that means we have to stretch or shrink the image accordingly:
		// (using nearest-neighbor technique, since it's the easiest to implement)

		XBYTE2 *img_rowstart;
		XBYTE2 *img_now;
		XBYTE2 *view_pix;

		int src_height = src_rect.bottom-src_rect.top;
		int src_width = src_rect.right-src_rect.left;

		img_start = (XBYTE2 *) source_pix + img_offset;
		view_start = (XBYTE2 *) dest_pix + view_offset;
		for (int y=0 ; y < img_ys; y++) {

			img_rowstart = img_start + y*src_height/img_ys*src_pitch;
			view_pix = view_start;
			for (int x=0 ; x < img_xs; x++) {
				img_now = img_rowstart + x*src_width/img_xs;
				if(nomask || *img_now) * view_pix = * img_now;
				view_pix++;
			}
			view_start += dst_pitch;
		}
	}
	return 1;
}

int WindowsDX::DrawAlphaUnmerged (XBYTE * source_pix, XBYTE * src_alpha_pix, XBYTE * dest_pix, RECT & src_rect, RECT & dst_rect, RECT & src_bounds, RECT & dst_bounds, int src_pitch, int dst_pitch, int alphafactor)
{
	// Unfortunately, DirectDraw doesn't support alpha blending yet.
	// Fortunately, Direct3D does support alpha blending, but this function is for in case we can't use D3D

	int drawType = ClipRects(src_rect, dst_rect, src_bounds, dst_bounds);

	if(drawType == 0) return 1; // image is entirely out of bounds, so just return

	if(drawType == 1) {
		// if source and destination rectangles are exactly the same size:

		XBYTE2 *img_start;
		XBYTE2 *img_now;

		XBYTE *alpha_start;
		XBYTE *alpha_now;

		XBYTE2 *view_start;
		XBYTE2 *view_pix;

		int img_offset = src_rect.top * src_pitch + src_rect.left;
		int view_offset = dst_rect.top * dst_pitch + dst_rect.left;
		int img_xs = dst_rect.right-dst_rect.left;
		int img_ys = dst_rect.bottom-dst_rect.top;
		int src_height = src_rect.bottom-src_rect.top;
		int src_width = src_rect.right-src_rect.left;

		img_start = (XBYTE2 *) source_pix + img_offset;
		alpha_start = (XBYTE *) src_alpha_pix + img_offset;
		view_start = (XBYTE2 *) dest_pix + view_offset;

		XBYTE2 *temp = new XBYTE2[img_xs];

		for (int y=0 ; y < img_ys; y++) {
			memcpy (temp, view_start, img_xs<<1) ; // copy a source-length line of the destination -- the slowest part by far, but there doesn't seem to be a way to optimize it any more...
			view_pix = view_start;
			img_now = img_start;
			alpha_now = alpha_start;
			for (int x=0 ; x < img_xs; x++) {
				int a = (255 - *alpha_now) * alphafactor / 255;
				if (a) { // the >> 8 introduces a slight inaccuracy because it's / 256 but should be / 255, this check makes it unnoticeable
					int i = *img_now;
					int v = temp[view_pix-view_start] ; // speedy alternative to: v = *view_pix
					*view_pix = (((((v & win_maskb) * (255-a)) >> 8) & win_maskb) + ((((i & win_maskb) * a) >> 8) & win_maskb))
							  | (((((v & win_maskg) * (255-a)) >> 8) & win_maskg) + ((((i & win_maskg) * a) >> 8) & win_maskg))
							  | (((((v & win_maskr) * (255-a)) >> 8) & win_maskr) + ((((i & win_maskr) * a) >> 8) & win_maskr));
//					*view_pix = ((((v & win_maskb) * (255-a)) >> 8) & win_maskb) // old alpha code,
//							  | ((((v & win_maskg) * (255-a)) >> 8) & win_maskg) // results in alpha source
//							  | ((((v & win_maskr) * (255-a)) >> 8) & win_maskr) // and added destination
//							  + *img_now;
				}
				view_pix++; img_now++; alpha_now++;
			}
			view_start += dst_pitch;
			img_start += src_pitch;
			alpha_start += src_pitch;
		}
		delete [] temp;

	} else {
		// if ClipRects() returned 2, that means we have to stretch or shrink the image accordingly:
		// (using nearest-neighbor technique, since it's the easiest to implement)

		XBYTE2 *img_start;
		XBYTE2 *img_rowstart;
		XBYTE2 *img_now;

		XBYTE *alpha_start;
		XBYTE *alpha_rowstart;
		XBYTE *alpha_now;

		XBYTE2 *view_start;
		XBYTE2 *view_pix;

		int img_offset = src_rect.top * src_pitch + src_rect.left;
		int view_offset = dst_rect.top * dst_pitch + dst_rect.left;
		int img_xs = dst_rect.right-dst_rect.left;
		int img_ys = dst_rect.bottom-dst_rect.top;
		int src_height = src_rect.bottom-src_rect.top;
		int src_width = src_rect.right-src_rect.left;

		img_start = (XBYTE2 *) source_pix + img_offset;
		alpha_start = (XBYTE *) src_alpha_pix + img_offset;
		view_start = (XBYTE2 *) dest_pix + view_offset;

		XBYTE2 *temp = new XBYTE2[img_xs];

		for (int y=0 ; y < img_ys; y++) {
			memcpy (temp, view_start, img_xs<<1) ; // copy a source-length line of the destination -- the slowest part by far, but there doesn't seem to be a way to optimize it any more...

			int yOffset = y*src_height/img_ys*src_pitch;
			img_rowstart = img_start + yOffset;
			alpha_rowstart = alpha_start + yOffset;
			view_pix = view_start;
			for (int x=0 ; x < img_xs; x++) {
				int xOffset = x*src_width/img_xs;
				img_now = img_rowstart + xOffset;
				alpha_now = alpha_rowstart + xOffset;

				int a = (255 - *alpha_now) * alphafactor / 255;
				if (a) { // the >> 8 introduces a slight inaccuracy because it's / 256 but should be / 255, this check makes it unnoticeable
					int i = *img_now;
					int v = temp[view_pix-view_start] ; // speedy alternative to: v = *view_pix
					*view_pix = (((((v & win_maskb) * (255-a)) >> 8) & win_maskb) + ((((i & win_maskb) * a) >> 8) & win_maskb))
							  | (((((v & win_maskg) * (255-a)) >> 8) & win_maskg) + ((((i & win_maskg) * a) >> 8) & win_maskg))
							  | (((((v & win_maskr) * (255-a)) >> 8) & win_maskr) + ((((i & win_maskr) * a) >> 8) & win_maskr));
//					*view_pix = ((((v & win_maskb) * (255-a)) >> 8) & win_maskb) // old alpha code,
//							  | ((((v & win_maskg) * (255-a)) >> 8) & win_maskg) // results in alpha source
//							  | ((((v & win_maskr) * (255-a)) >> 8) & win_maskr) // and added destination
//							  + *img_now;
				}
				view_pix++;
			}
			view_start += dst_pitch;
		}
		delete [] temp;
	}
	return 1;
}

int WindowsDX::DrawAlphaMerged (XBYTE * source_pix, XBYTE * dest_pix, RECT & src_rect, RECT & dst_rect, RECT & src_bounds, RECT & dst_bounds, int src_pitch, int dst_pitch, int alphafactor)
{
	// Unfortunately, DirectDraw doesn't support alpha blending yet.
	// Fortunately, Direct3D does support alpha blending, but this function is for in case we can't use D3D

	int drawType = ClipRects(src_rect, dst_rect, src_bounds, dst_bounds);

	if(drawType == 0) return 1; // image is entirely out of bounds, so just return

	if(drawType == 1) {
		// if source and destination rectangles are exactly the same size:

		XBYTE4 *img_start;
		XBYTE4 *img_now;

		XBYTE2 *view_start;
		XBYTE2 *view_pix;

		int img_offset = src_rect.top * src_pitch + src_rect.left;
		int view_offset = dst_rect.top * dst_pitch + dst_rect.left;
		int img_xs = dst_rect.right-dst_rect.left;
		int img_ys = dst_rect.bottom-dst_rect.top;
		int src_height = src_rect.bottom-src_rect.top;
		int src_width = src_rect.right-src_rect.left;

		img_start = (XBYTE4 *) source_pix + img_offset;
		view_start = (XBYTE2 *) dest_pix + view_offset;

		XBYTE2 *temp = new XBYTE2[img_xs];

		for (int y=0 ; y < img_ys; y++) {
			memcpy (temp, view_start, img_xs<<1) ; // copy a source-length line of the destination -- the slowest part by far, but there doesn't seem to be a way to optimize it any more...
			view_pix = view_start;
			img_now = img_start;
			for (int x=0 ; x < img_xs; x++) {
				int i = *img_now;
				int a = ((i & 0xFF000000) >> 24);
//				int a = (255 - *alpha_now) * alphafactor / 255;
				if (a) { // the >> 8 introduces a slight inaccuracy because it's / 256 but should be / 255, this check makes it unnoticeable
					int r = ((i & 0x00FF0000) >> 16);
					int g = ((i & 0x0000FF00) >> 8);
					int b = ((i & 0x000000FF) >> 0);
					i = ((r >> 3) << win_shiftr) | ((g >> 2) << win_shiftg) | ((b >> 3) << win_shiftb);
					int v = temp[view_pix-view_start] ; // speedy alternative to: v = *view_pix
					*view_pix = (((((v & win_maskb) * (255-a)) >> 8) & win_maskb) + ((((i & win_maskb) * a) >> 8) & win_maskb))
							  | (((((v & win_maskg) * (255-a)) >> 8) & win_maskg) + ((((i & win_maskg) * a) >> 8) & win_maskg))
							  | (((((v & win_maskr) * (255-a)) >> 8) & win_maskr) + ((((i & win_maskr) * a) >> 8) & win_maskr));
//					*view_pix = ((((v & win_maskb) * (255-a)) >> 8) & win_maskb) // old alpha code,
//							  | ((((v & win_maskg) * (255-a)) >> 8) & win_maskg) // results in alpha source
//							  | ((((v & win_maskr) * (255-a)) >> 8) & win_maskr) // and added destination
//							  + *img_now;
				}
				view_pix++; img_now++;
			}
			view_start += dst_pitch;
			img_start += src_pitch;
		}
		delete [] temp;

	} else {
		// if ClipRects() returned 2, that means we have to stretch or shrink the image accordingly:
		// (using nearest-neighbor technique, since it's the easiest to implement)

		XBYTE4 *img_start;
		XBYTE4 *img_rowstart;
		XBYTE4 *img_now;

		XBYTE2 *view_start;
		XBYTE2 *view_pix;

		int img_offset = src_rect.top * src_pitch + src_rect.left;
		int view_offset = dst_rect.top * dst_pitch + dst_rect.left;
		int img_xs = dst_rect.right-dst_rect.left;
		int img_ys = dst_rect.bottom-dst_rect.top;
		int src_height = src_rect.bottom-src_rect.top;
		int src_width = src_rect.right-src_rect.left;

		img_start = (XBYTE4 *) source_pix + img_offset;
		view_start = (XBYTE2 *) dest_pix + view_offset;

		XBYTE2 *temp = new XBYTE2[img_xs];

		for (int y=0 ; y < img_ys; y++) {
			memcpy (temp, view_start, img_xs<<1) ; // copy a source-length line of the destination -- the slowest part by far, but there doesn't seem to be a way to optimize it any more...

			int yOffset = y*src_height/img_ys*src_pitch;
			img_rowstart = img_start + yOffset;
			view_pix = view_start;
			for (int x=0 ; x < img_xs; x++) {
				int xOffset = x*src_width/img_xs;
				img_now = img_rowstart + xOffset;

				int i = *img_now;
				int a = ((i & 0xFF000000) >> 24);
//				int a = (255 - *alpha_now) * alphafactor / 255;
				if (a) { // the >> 8 introduces a slight inaccuracy because it's / 256 but should be / 255, this check makes it unnoticeable
					int r = ((i & 0x00FF0000) >> 16);
					int g = ((i & 0x0000FF00) >> 8);
					int b = ((i & 0x000000FF) >> 0);
					i = ((r >> 3) << win_shiftr) | ((g >> 2) << win_shiftg) | ((b >> 3) << win_shiftb);
					int v = temp[view_pix-view_start] ; // speedy alternative to: v = *view_pix
					*view_pix = (((((v & win_maskb) * (255-a)) >> 8) & win_maskb) + ((((i & win_maskb) * a) >> 8) & win_maskb))
							  | (((((v & win_maskg) * (255-a)) >> 8) & win_maskg) + ((((i & win_maskg) * a) >> 8) & win_maskg))
							  | (((((v & win_maskr) * (255-a)) >> 8) & win_maskr) + ((((i & win_maskr) * a) >> 8) & win_maskr));
//					*view_pix = ((((v & win_maskb) * (255-a)) >> 8) & win_maskb) // old alpha code,
//							  | ((((v & win_maskg) * (255-a)) >> 8) & win_maskg) // results in alpha source
//							  | ((((v & win_maskr) * (255-a)) >> 8) & win_maskr) // and added destination
//							  + *img_now;
				}
				view_pix++;
			}
			view_start += dst_pitch;
		}
		delete [] temp;
	}
	return 1;
}


// ri, gi, bi are RGB intensity values, they default to 255 if not passed in
// 255 means full additive, -255 means full subtractive,
// 128 means half additive, -128 means half subtractive,
// 255,0,0 means full-add red only, 0,0,-255 means full-subtract blue only, etc.
// they can go outside the -255 to 255 range too, such as 512 for doubly-additive
int WindowsDX::DrawAdded (XBYTE * source_pix, XBYTE * dest_pix, RECT & src_rect, RECT & dst_rect, RECT & src_bounds, RECT & dst_bounds, int src_pitch, int dst_pitch, int ri, int gi, int bi)
{
	if(!(ri||gi||bi)) return 1; // if no intensity at all, return to save processor time

	int drawType = ClipRects(src_rect, dst_rect, src_bounds, dst_bounds);

	if(drawType == 0) return 1; // image is entirely out of bounds, so just return

	if(drawType == 1) {
		// if source and destination rectangles are exactly the same size:

		XBYTE2 *img_start;
		XBYTE2 *img_now;

		XBYTE2 *view_start;
		XBYTE2 *view_pix;

		int img_offset = src_rect.top * src_pitch + src_rect.left;
		int view_offset = dst_rect.top * dst_pitch + dst_rect.left;
		int img_xs = dst_rect.right-dst_rect.left;
		int img_ys = dst_rect.bottom-dst_rect.top;
		int src_height = src_rect.bottom-src_rect.top;
		int src_width = src_rect.right-src_rect.left;

		img_start = (XBYTE2 *) source_pix + img_offset;
		view_start = (XBYTE2 *) dest_pix + view_offset;

		XBYTE2 *temp = new XBYTE2[img_xs];

		if(ri==255 && gi==255 && bi==255) { // slightly faster code in fully additive case (which r,g,b default to)
			for (int y=0 ; y < img_ys; y++) {
				memcpy (temp, view_start, img_xs<<1) ; // copy a source-length line of the destination -- the slowest part by far, but there doesn't seem to be a way to optimize it any more...
				view_pix = view_start;
				img_now = img_start;
				for (int x=0 ; x < img_xs; x++) {
					int i, v;
					if ((i = *img_now) && ((v = temp[view_pix-view_start]) != 65535)) { // if destination pixel isn't white and source pixel isn't black
						*view_pix = min(((v & win_maskb) + (i & win_maskb)), win_maskb)
								  | min(((v & win_maskg) + (i & win_maskg)), win_maskg)
								  | min(((v & win_maskr) + (i & win_maskr)), win_maskr) ;
					}
					view_pix++; img_now++;
				}
				view_start += dst_pitch;
				img_start += src_pitch;
			}
		}
		else {
			for (int y=0 ; y < img_ys; y++) {
				memcpy (temp, view_start, img_xs<<1) ; // copy a source-length line of the destination -- the slowest part by far, but there doesn't seem to be a way to optimize it any more...
				view_pix = view_start;
				img_now = img_start;
				for (int x=0 ; x < img_xs; x++) {
					int i, v;
					if ((i = *img_now)) { // if source pixel isn't black
						v = temp[view_pix-view_start] ; // speedy alternative to: v = *view_pix
						int blue  = ((v & win_maskb) + (((((i & win_maskb) >> win_shiftb) * bi) >> 8) << win_shiftb)) ;
						int green = ((v & win_maskg) + (((((i & win_maskg) >> win_shiftg) * gi) >> 8) << win_shiftg)) ;
						int red   = ((v & win_maskr) + (((((i & win_maskr) >> win_shiftr) * ri) >> 8) << win_shiftr)) ;
						*view_pix = ((bi < 0) ? max(0,blue)  : min(win_maskb, blue))
								  | ((gi < 0) ? max(0,green) : min(win_maskg, green))
								  | ((ri < 0) ? max(0,red)	 : min(win_maskr, red)) ;
					}
					view_pix++; img_now++;
				}
				view_start += dst_pitch;
				img_start += src_pitch;
			}
		}
		delete [] temp;

	} else {
		// if ClipRects() returned 2, that means we have to stretch or shrink the image accordingly:
		// (using nearest-neighbor technique, since it's the easiest to implement)

		XBYTE2 *img_start;
		XBYTE2 *img_rowstart;
		XBYTE2 *img_now;

		XBYTE2 *view_start;
		XBYTE2 *view_pix;

		int img_offset = src_rect.top * src_pitch + src_rect.left;
		int view_offset = dst_rect.top * dst_pitch + dst_rect.left;
		int img_xs = dst_rect.right-dst_rect.left;
		int img_ys = dst_rect.bottom-dst_rect.top;
		int src_height = src_rect.bottom-src_rect.top;
		int src_width = src_rect.right-src_rect.left;

		img_start = (XBYTE2 *) source_pix + img_offset;
		view_start = (XBYTE2 *) dest_pix + view_offset;

		XBYTE2 *temp = new XBYTE2[img_xs];

		for (int y=0 ; y < img_ys; y++) {
			memcpy (temp, view_start, img_xs<<1) ; // copy a source-length line of the destination -- the slowest part by far, but there doesn't seem to be a way to optimize it any more...
			img_rowstart = img_start + y*src_height/img_ys*src_pitch;
			view_pix = view_start;
			for (int x=0 ; x < img_xs; x++) {
				img_now = img_rowstart + x*src_width/img_xs;

				// assignment, not equality... and v = temp[view_pix] is the speedy alternative to: v = *view_pix
				int i, v;
				if ((i = *img_now)) { // if source pixel isn't black
					v = temp[view_pix-view_start] ; // speedy alternative to: v = *view_pix
					int blue  = ((v & win_maskb) + (((((i & win_maskb) >> win_shiftb) * bi) >> 8) << win_shiftb)) ;
					int green = ((v & win_maskg) + (((((i & win_maskg) >> win_shiftg) * gi) >> 8) << win_shiftg)) ;
					int red   = ((v & win_maskr) + (((((i & win_maskr) >> win_shiftr) * ri) >> 8) << win_shiftr)) ;
					*view_pix = ((bi < 0) ? max(0,blue)  : min(win_maskb, blue))
							  | ((gi < 0) ? max(0,green) : min(win_maskg, green))
							  | ((ri < 0) ? max(0,red)	 : min(win_maskr, red)) ;
				}

				view_pix++;
			}
			view_start += dst_pitch;
		}
		delete [] temp;
	}
	return 1;
}

int WindowsDX::DrawAddedMerged (XBYTE * source_pix, XBYTE * dest_pix, RECT & src_rect, RECT & dst_rect, RECT & src_bounds, RECT & dst_bounds, int src_pitch, int dst_pitch, int ri, int gi, int bi)
{
	if(!(ri||gi||bi)) return 1; // if no intensity at all, return to save processor time

	int drawType = ClipRects(src_rect, dst_rect, src_bounds, dst_bounds);

	if(drawType == 0) return 1; // image is entirely out of bounds, so just return

	if(drawType == 1) {
		// if source and destination rectangles are exactly the same size:

		XBYTE4 *img_start;
		XBYTE4 *img_now;

		XBYTE2 *view_start;
		XBYTE2 *view_pix;

		int img_offset = src_rect.top * src_pitch + src_rect.left;
		int view_offset = dst_rect.top * dst_pitch + dst_rect.left;
		int img_xs = dst_rect.right-dst_rect.left;
		int img_ys = dst_rect.bottom-dst_rect.top;
		int src_height = src_rect.bottom-src_rect.top;
		int src_width = src_rect.right-src_rect.left;

		img_start = (XBYTE4 *) source_pix + img_offset;
		view_start = (XBYTE2 *) dest_pix + view_offset;

		XBYTE2 *temp = new XBYTE2[img_xs];

		if(ri==255 && gi==255 && bi==255) { // slightly faster code in fully additive case (which r,g,b default to)
			for (int y=0 ; y < img_ys; y++) {
				memcpy (temp, view_start, img_xs<<1) ; // copy a source-length line of the destination -- the slowest part by far, but there doesn't seem to be a way to optimize it any more...
				view_pix = view_start;
				img_now = img_start;
				for (int x=0 ; x < img_xs; x++) {
					int i = *img_now;
					int a = ((i & 0xFF000000) >> 24);
					int v;
					if(a && (i & 0x00FFFFFF) && ((v = temp[view_pix-view_start]) != 65535)) { // if destination pixel isn't white and source pixel isn't black
						int r = (((i & 0x00FF0000) >> 16) * a) >> 8; // >> 8 is approximation for / 255
						int g = (((i & 0x0000FF00) >> 8) * a) >> 8; // >> 8 is approximation for / 255
						int b = (((i & 0x000000FF) >> 0) * a) >> 8; // >> 8 is approximation for / 255
						i = ((r >> 3) << win_shiftr) | ((g >> 2) << win_shiftg) | ((b >> 3) << win_shiftb);
						*view_pix = min(((v & win_maskb) + (i & win_maskb)), win_maskb)
								  | min(((v & win_maskg) + (i & win_maskg)), win_maskg)
								  | min(((v & win_maskr) + (i & win_maskr)), win_maskr) ;
					}
					view_pix++; img_now++;
				}
				view_start += dst_pitch;
				img_start += src_pitch;
			}
		}
		else {
			for (int y=0 ; y < img_ys; y++) {
				memcpy (temp, view_start, img_xs<<1) ; // copy a source-length line of the destination -- the slowest part by far, but there doesn't seem to be a way to optimize it any more...
				view_pix = view_start;
				img_now = img_start;
				for (int x=0 ; x < img_xs; x++) {
					int i = *img_now;
					int a = ((i & 0xFF000000) >> 24);
					int v = temp[view_pix-view_start] ; // speedy alternative to: v = *view_pix
					if(a && (i & 0x00FFFFFF)) { // if source pixel isn't black
						int r = ((((i & 0x00FF0000) >> 16) * a*ri)) >> 16; // >> 16 is approximation for / 255 / 255
						int g = ((((i & 0x0000FF00) >> 8) * a*gi)) >> 16; // >> 16 is approximation for / 255 / 255
						int b = ((((i & 0x000000FF) >> 0) * a*bi)) >> 16; // >> 16 is approximation for / 255 / 255
						i = ((r >> 3) << win_shiftr) | ((g >> 2) << win_shiftg) | ((b >> 3) << win_shiftb);
						r = min(((v & win_maskr) + (i & win_maskr)), win_maskr);
						g = min(((v & win_maskg) + (i & win_maskg)), win_maskg);
						b = min(((v & win_maskb) + (i & win_maskb)), win_maskb);
						*view_pix = ((bi < 0) ? max(0,b) : min(win_maskb, b))
								  | ((gi < 0) ? max(0,g) : min(win_maskg, g))
								  | ((ri < 0) ? max(0,r) : min(win_maskr, r)) ;
					}
					view_pix++; img_now++;
				}
				view_start += dst_pitch;
				img_start += src_pitch;
			}
		}
		delete [] temp;

	} else {
		// if ClipRects() returned 2, that means we have to stretch or shrink the image accordingly:
		// (using nearest-neighbor technique, since it's the easiest to implement)

		XBYTE4 *img_start;
		XBYTE4 *img_rowstart;
		XBYTE4 *img_now;

		XBYTE2 *view_start;
		XBYTE2 *view_pix;

		int img_offset = src_rect.top * src_pitch + src_rect.left;
		int view_offset = dst_rect.top * dst_pitch + dst_rect.left;
		int img_xs = dst_rect.right-dst_rect.left;
		int img_ys = dst_rect.bottom-dst_rect.top;
		int src_height = src_rect.bottom-src_rect.top;
		int src_width = src_rect.right-src_rect.left;

		img_start = (XBYTE4 *) source_pix + img_offset;
		view_start = (XBYTE2 *) dest_pix + view_offset;

		XBYTE2 *temp = new XBYTE2[img_xs];

		for (int y=0 ; y < img_ys; y++) {
			memcpy (temp, view_start, img_xs<<1) ; // copy a source-length line of the destination -- the slowest part by far, but there doesn't seem to be a way to optimize it any more...
			img_rowstart = img_start + y*src_height/img_ys*src_pitch;
			view_pix = view_start;
			for (int x=0 ; x < img_xs; x++) {
				img_now = img_rowstart + x*src_width/img_xs;
				int i = *img_now;
				int a = ((i & 0xFF000000) >> 24);
				int v = temp[view_pix-view_start] ; // speedy alternative to: v = *view_pix
				if(a && (i & 0x00FFFFFF)) { // if source pixel isn't black
					int r = ((((i & 0x00FF0000) >> 16) * a*ri)) >> 16; // >> 16 is approximation for / 255 / 255
					int g = ((((i & 0x0000FF00) >> 8) * a*gi)) >> 16; // >> 16 is approximation for / 255 / 255
					int b = ((((i & 0x000000FF) >> 0) * a*bi)) >> 16; // >> 16 is approximation for / 255 / 255
					i = ((r >> 3) << win_shiftr) | ((g >> 2) << win_shiftg) | ((b >> 3) << win_shiftb);
					r = min(((v & win_maskr) + (i & win_maskr)), win_maskr);
					g = min(((v & win_maskg) + (i & win_maskg)), win_maskg);
					b = min(((v & win_maskb) + (i & win_maskb)), win_maskb);
					*view_pix = ((bi < 0) ? max(0,b) : min(win_maskb, b))
							  | ((gi < 0) ? max(0,g) : min(win_maskg, g))
							  | ((ri < 0) ? max(0,r) : min(win_maskr, r)) ;
				}

				view_pix++;
			}
			view_start += dst_pitch;
		}
		delete [] temp;
	}
	return 1;
}


// ro, go, bo are RGB opacity values, they default to 128 if not passed in,
// 255,255,255 means fully opaque, 0,0,0 means fully transparent, 128,128,128 means half transparent,
// 255,128,0 means fully opaque red, half transparent green and fully transparent blue, etc.
int WindowsDX::DrawBlended (XBYTE * source_pix, XBYTE * dest_pix, RECT & src_rect, RECT & dst_rect, RECT & src_bounds, RECT & dst_bounds, int src_pitch, int dst_pitch, int r, int g, int b, int a, int nomask)
{
	if(!a) return 1; // if no opacity at all, return to save processor time

	int drawType = ClipRects(src_rect, dst_rect, src_bounds, dst_bounds);

	if(drawType == 0) return 1; // image is entirely out of bounds, so just return

	XBYTE2 *img_start;
	XBYTE2 *img_now;
	XBYTE2 *view_start;
	XBYTE2 *view_pix;
	int img_offset = src_rect.top * src_pitch + src_rect.left;
	int view_offset = dst_rect.top * dst_pitch + dst_rect.left;
	int img_xs = dst_rect.right-dst_rect.left;
	int img_ys = dst_rect.bottom-dst_rect.top;

	if(drawType == 1) {
		// if source and destination rectangles are exactly the same size:


		img_start = (XBYTE2 *) source_pix + img_offset;
		view_start = (XBYTE2 *) dest_pix + view_offset;

		XBYTE2 *temp = new XBYTE2[img_xs];

		for (int y=0 ; y < img_ys; y++) {
			memcpy (temp, view_start, img_xs<<1) ; // copy a source-length line of the destination -- the slowest part by far, but there doesn't seem to be a way to optimize it any more...
			view_pix = view_start;
			img_now = img_start;
			for (int x=0 ; x < img_xs; x++) {
				int i, v;
				if ((i = *img_now) || nomask) { // if source pixel isn't black
					v = temp[view_pix-view_start] ; // speedy alternative to: v = *view_pix
					*view_pix =
						  ((((((v & win_maskb) >> win_shiftb) *   (255-a)) >> 8) << win_shiftb)
						 + (((((i & win_maskb) >> win_shiftb) * (b*a/255)) >> 8) << win_shiftb))
						| ((((((v & win_maskg) >> win_shiftg) *   (255-a)) >> 8) << win_shiftg)
						 + (((((i & win_maskg) >> win_shiftg) * (g*a/255)) >> 8) << win_shiftg))
						| ((((((v & win_maskr) >> win_shiftr) *   (255-a)) >> 8) << win_shiftr)
						 + (((((i & win_maskr) >> win_shiftr) * (r*a/255)) >> 8) << win_shiftr)) ;
				}
				view_pix++; img_now++;
			}
			view_start += dst_pitch;
			img_start += src_pitch;
		}
		delete [] temp;

	} else {
		// if ClipRects() returned 2, that means we have to stretch or shrink the image accordingly:
		// (using nearest-neighbor technique, since it's the easiest to implement)

		XBYTE2 *img_rowstart;

		int src_height = src_rect.bottom-src_rect.top;
		int src_width = src_rect.right-src_rect.left;

		img_start = (XBYTE2 *) source_pix + img_offset;
		view_start = (XBYTE2 *) dest_pix + view_offset;

		XBYTE2 *temp = new XBYTE2[img_xs];

		for (int y=0 ; y < img_ys; y++) {
			memcpy (temp, view_start, img_xs<<1) ; // copy a source-length line of the destination -- the slowest part by far, but there doesn't seem to be a way to optimize it any more...
			img_rowstart = img_start + y*src_height/img_ys*src_pitch;
			view_pix = view_start;
			for (int x=0 ; x < img_xs; x++) {
				img_now = img_rowstart + x*src_width/img_xs;

				// assignment, not equality... and v = temp[view_pix] is the speedy alternative to: v = *view_pix
				int i, v;
				if ((i = *img_now) || nomask) { // if source pixel isn't black
					v = temp[view_pix-view_start] ; // speedy alternative to: v = *view_pix
					*view_pix =
						  ((((((v & win_maskb) >> win_shiftb) *   (255-a)) >> 8) << win_shiftb)
						 + (((((i & win_maskb) >> win_shiftb) * (b*a/255)) >> 8) << win_shiftb))
						| ((((((v & win_maskg) >> win_shiftg) *   (255-a)) >> 8) << win_shiftg)
						 + (((((i & win_maskg) >> win_shiftg) * (g*a/255)) >> 8) << win_shiftg))
						| ((((((v & win_maskr) >> win_shiftr) *   (255-a)) >> 8) << win_shiftr)
						 + (((((i & win_maskr) >> win_shiftr) * (r*a/255)) >> 8) << win_shiftr)) ;
				}

				view_pix++;
			}
			view_start += dst_pitch;
		}
		delete [] temp;
	}
	return 1;
}

int WindowsDX::DrawInvert (XBYTE * source_pix, XBYTE * dest_pix, RECT & src_rect, RECT & dst_rect, RECT & src_bounds, RECT & dst_bounds, int src_pitch, int dst_pitch, int ri, int gi, int bi)
{
	if(!(ri||gi||bi)) return 1; // if no intensity at all, return to save processor time

	int drawType = ClipRects(src_rect, dst_rect, src_bounds, dst_bounds);

	if(drawType == 0) return 1; // image is entirely out of bounds, so just return

	if(drawType == 1) {
		// if source and destination rectangles are exactly the same size:

		XBYTE2 *img_start;
		XBYTE2 *img_now;

		XBYTE2 *view_start;
		XBYTE2 *view_pix;

		int img_offset = src_rect.top * src_pitch + src_rect.left;
		int view_offset = dst_rect.top * dst_pitch + dst_rect.left;
		int img_xs = dst_rect.right-dst_rect.left;
		int img_ys = dst_rect.bottom-dst_rect.top;
		int src_height = src_rect.bottom-src_rect.top;
		int src_width = src_rect.right-src_rect.left;

		img_start = (XBYTE2 *) source_pix + img_offset;
		view_start = (XBYTE2 *) dest_pix + view_offset;

		XBYTE2 *temp = new XBYTE2[img_xs];

//		device->SetRenderState(D3DRENDERSTATE_SRCBLEND,D3DBLEND_INVDESTCOLOR);
//		device->SetRenderState(D3DRENDERSTATE_DESTBLEND,D3DBLEND_INVSRCCOLOR);
//		which means
//		Cs = (1 - Rd, 1 - Gd, 1 - Bd, 1 - Ad)
//		Cd = (1 - Rs, 1 - Gs, 1 - Bs, 1 - As)
//		and
//		C  = i * Cs + v * Cd 
//		so
//		C  = i * Cs 					  + v * Cd 
//		   = i * (1 - Rd, 1 - Gd, 1 - Bd) + v * (1 - Rs, 1 - Gs, 1 - Bs)
//		   = i * (1 - v[r], 1 - v[g], 1 - v[b]) + v * (1 - i[r], 1 - i[g], 1 - i[b])
//		C  = i[r] * (1 - v[r]) + i[g] * (1 - v[g]) + i[b] * (1 - v[b])
//		   + v[r] * (1 - i[r]) + v[g] * (1 - i[g]) + v[b] * (1 - i[b])
//
//		C  = i[r] + i[g] + i[b]
//		   + v[r] + v[g] + v[b]
//		   - 2*i[r]*v[r] - 2*i[g]*v[g] - 2*i[b]*v[b]
//
//		C  = i + v - 2*i[r]*v[r] - 2*i[g]*v[g] - 2*i[b]*v[b]


//		if(ri==255 && gi==255 && bi==255) { // slightly faster code in fully additive case (which r,g,b default to)
			for (int y=0 ; y < img_ys; y++) {
				memcpy (temp, view_start, img_xs<<1) ; // copy a source-length line of the destination -- the slowest part by far, but there doesn't seem to be a way to optimize it any more...
				view_pix = view_start;
				img_now = img_start;
				for (int x=0 ; x < img_xs; x++) {
					int i = *img_now;
					int v = temp[view_pix-view_start];

					int iBlue = ((i & win_maskb) >> win_shiftb) * bi / 255;
					int iGreen = ((i & win_maskg) >> win_shiftg) * gi / 255;
					int iRed = ((i & win_maskr) >> win_shiftr) * ri / 255;
					int vBlue = (v & win_maskb) >> win_shiftb;
					int vGreen = (v & win_maskg) >> win_shiftg;
					int vRed = (v & win_maskr) >> win_shiftr;
					int bSub = ((iBlue) * (vBlue)) / 4;
					
					// almost right...
					// C  = i + v - 2*i[r]*v[r] - 2*i[g]*v[g] - 2*i[b]*v[b]
					// should work though...

					int blue  = ((v & win_maskb) + (i & win_maskb) - ((( (iBlue)  * (vBlue))  / 16) << win_shiftb)) ;
					int green = ((v & win_maskg) + (i & win_maskg) - ((( (iGreen) * (vGreen)) / 32) << win_shiftg)) ;
					int red   = ((v & win_maskr) + (i & win_maskr) - ((( (iRed)   * (vRed))   / 16) << win_shiftr)) ;
					*view_pix = ((bi < 0) ? max(0,blue)  : min(win_maskb, blue))
							  | ((gi < 0) ? max(0,green) : min(win_maskg, green))
							  | ((ri < 0) ? max(0,red)	 : min(win_maskr, red)) ;

					view_pix++; img_now++;
				}
				view_start += dst_pitch;
				img_start += src_pitch;
			}
		delete [] temp;

	} else {
		// if ClipRects() returned 2, that means we have to stretch or shrink the image accordingly:
		// (using nearest-neighbor technique, since it's the easiest to implement)

		XBYTE2 *img_start;
		XBYTE2 *img_rowstart;
		XBYTE2 *img_now;

		XBYTE2 *view_start;
		XBYTE2 *view_pix;

		int img_offset = src_rect.top * src_pitch + src_rect.left;
		int view_offset = dst_rect.top * dst_pitch + dst_rect.left;
		int img_xs = dst_rect.right-dst_rect.left;
		int img_ys = dst_rect.bottom-dst_rect.top;
		int src_height = src_rect.bottom-src_rect.top;
		int src_width = src_rect.right-src_rect.left;

		img_start = (XBYTE2 *) source_pix + img_offset;
		view_start = (XBYTE2 *) dest_pix + view_offset;

		XBYTE2 *temp = new XBYTE2[img_xs];

		for (int y=0 ; y < img_ys; y++) {
			memcpy (temp, view_start, img_xs<<1) ; // copy a source-length line of the destination -- the slowest part by far, but there doesn't seem to be a way to optimize it any more...
			img_rowstart = img_start + y*src_height/img_ys*src_pitch;
			view_pix = view_start;
			for (int x=0 ; x < img_xs; x++) {
				img_now = img_rowstart + x*src_width/img_xs;

				int i = *img_now;
				int v = temp[view_pix-view_start];

				int iBlue = ((i & win_maskb) >> win_shiftb) * bi / 255;
				int iGreen = ((i & win_maskg) >> win_shiftg) * gi / 255;
				int iRed = ((i & win_maskr) >> win_shiftr) * ri / 255;
				int vBlue = (v & win_maskb) >> win_shiftb;
				int vGreen = (v & win_maskg) >> win_shiftg;
				int vRed = (v & win_maskr) >> win_shiftr;
				int bSub = ((iBlue) * (vBlue)) / 4;
				
				// almost right...
				// C  = i + v - 2*i[r]*v[r] - 2*i[g]*v[g] - 2*i[b]*v[b]
				// should work though...

				int blue  = ((v & win_maskb) + (i & win_maskb) - ((( (iBlue)  * (vBlue))  / 16) << win_shiftb)) ;
				int green = ((v & win_maskg) + (i & win_maskg) - ((( (iGreen) * (vGreen)) / 32) << win_shiftg)) ;
				int red   = ((v & win_maskr) + (i & win_maskr) - ((( (iRed)   * (vRed))   / 16) << win_shiftr)) ;
				*view_pix = ((bi < 0) ? max(0,blue)  : min(win_maskb, blue))
						  | ((gi < 0) ? max(0,green) : min(win_maskg, green))
						  | ((ri < 0) ? max(0,red)	 : min(win_maskr, red)) ;

				view_pix++;
			}
			view_start += dst_pitch;
		}
		delete [] temp;
	}
	return 1;
}

/*
static int DrawImage24 (ImageX *img, int x, int y)
{
	XBYTE *img_start;
	XBYTE *view_start;
	int img_offset, view_offset;
	int img_xs, img_ys;
	int img_clipwid, view_wid, img_xwid;

	GameX.DrawOpen() ;

	img_start = (XBYTE*) img->m_data;
	if (img_start != NULL && GameX.ClipRects(x,y,img, view_offset, img_offset, img_xs, img_ys)) {
		view_start = (XBYTE *) GameX.win_screen + view_offset*3;
		img_start += img_offset*3;
		img_clipwid = img_xs * 3;
		img_xwid = img->m_xres * 3;
		view_wid = GameX.win_width * 3;
		for (y=img_ys; y--;) {			
			memcpy (view_start, img_start, img_clipwid) ;
			view_start += view_wid;
			img_start += img_xwid;
		}		
	}
	return 1;
}

static int DrawMasked24 (ImageX *img, int x, int y)
{
	XBYTE *img_start;
	XBYTE *view_start;	
	int img_offset, view_offset;
	int img_xs, img_ys;
	XBYTE *img_pix;
	XBYTE *view_pix;
	int img_clipwid;
	int img_xwid, view_wid;
	
	GameX.DrawOpen() ;

	img_start = (XBYTE *) img->m_data;
	if (img_start != NULL && GameX.ClipRects(x,y,img, view_offset, img_offset, img_xs, img_ys)) {
		view_start = (XBYTE *) GameX.win_screen + view_offset*3;
		img_start += img_offset;
		img_clipwid = img_xs * 3;
		img_xwid = img->m_xres * 3;
		view_wid = GameX.win_width * 3;
		for (y=img_ys; y--;) {			
			img_pix = img_start;
			for (view_pix=view_start; view_pix<view_start+img_clipwid;) {
				if (*img_pix>0) {
					*view_pix++ = *img_pix++;
					*view_pix++ = *img_pix++;
					*view_pix++ = *img_pix++;
				} else {
					view_pix+=3; img_pix+=3;
				}				
			}
			view_start += view_wid;
			img_start += img_xwid;
		}		
	}	
	return 1;
}

static int DrawAlpha24 (ImageX *img, int x, int y)
{
	XBYTE *img_start;
	XBYTE *alpha_start;
	XBYTE *view_start;
	int img_offset, view_offset;
	int img_xs, img_ys;
	XBYTE *img_pix;
	XBYTE *alpha_pix;
	XBYTE *view_pix;
	int i;
	int img_clipwid, img_xwid, view_wid;
	
	GameX.DrawOpen() ;

	img_start = (XBYTE*) img->m_data;
	alpha_start = (XBYTE*) img->m_alpha;
	if (img_start!=NULL && alpha_start!=NULL && GameX.ClipRects(x,y,img, view_offset, img_offset, img_xs, img_ys)) {
		view_start = (XBYTE*) GameX.win_screen + view_offset*3;
		img_start += img_offset*3;
		alpha_start += img_offset*3;
		img_clipwid = img_xs * 3;
		img_xwid = img->m_xres * 3;
		view_wid = GameX.win_width * 3;
		for (y=img_ys; y--;) {			
			img_pix = img_start;
			alpha_pix = alpha_start;
			for (view_pix=view_start; view_pix<view_start+img_clipwid;) {
				i = *alpha_pix;
				if (i!=255) {
					*view_pix++ = ((*view_pix * i) >> 8) + *img_pix;
					*view_pix++ = ((*view_pix * i) >> 8) + *img_pix;
					*view_pix++ = ((*view_pix * i) >> 8) + *img_pix;					
				} else {
					view_pix+=3; img_pix+=3;
				}
				alpha_pix++;
			}
			view_start += view_wid;
			img_start += img_xwid;
			alpha_start += img->m_xres;
		}		
	}
	return 1;
}
*/
void WindowsDX::ReportProblem (int problem)
{
	win_status = problem;
	ReportProblem() ;
}

void WindowsDX::ReportProblem (void)
{	
	if(win_status == GAMEX_READY)
		return; // apparently there's no problem

	char error[128] ;
	char recommend[640] ;	
	char disp[1280] ;
	
	switch (win_status) {
	case GAMEX_NO_INIT:
		sprintf (error, "Error: NO_INIT (User did not initialize GameX)") ;
		sprintf (recommend, "You must start GameX by calling:\n  GameX.Initialize (flags, xres, yres, speed)\n\nfrom the initialize() function.\nFor example:\n  GameX.Initialize(VIDEO_16BIT, 640, 480, 100);\n\nIf you are sure that you are making this call before anything else in initialize(), make sure that you don't have any\nstatically constructed objects whose constructors use GameX in some way, such as by loading an image.\n\nSee documentation for details on starting GameX.") ;
		break;
	case GAMEX_INITCLASS_FAILED:
		sprintf (error, "Error: INITCLASS_FAILED (Cannot Initialize Window Class)") ;
		sprintf (recommend, "Are you using Windows 98/2000/NT/XP? GameX requires one of these. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_INITWINTEMP_FAILED:
		sprintf (error, "Error: INITWINTEMP_FAILED (Cannot Initialize Temp Window)") ;
		sprintf (recommend, "Are you using Windows 98/2000/NT/XP? GameX requires one of these. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_INITWINREAL_FAILED:
		sprintf (error, "Error: INITWINREAL_FAILED (Cannot Initialize Main Window)") ;
		sprintf (recommend, "Are you using Windows 98/2000/NT/XP? GameX requires one of these. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_SHOWWIN_FAILED:
		sprintf (error, "Error: SHOWWIN_FAILED (Unable to show Main Window)") ;
		sprintf (recommend, "Are you using Windows 98/2000/NT/XP? GameX requires one of these. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_DDSTART_FAILED:
		sprintf (error, "Error: DDSTART_FAILED (Unable to start DirectDraw or DirectX)") ;
		sprintf (recommend, "Have you installed DirectX 7.0 or later? GameX requires it. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_D3DSTART_FAILED:
		sprintf (error, "Error: D3DSTART_FAILED (Unable to start Direct3D)") ;
		sprintf (recommend, "Have you installed DirectX 7.0 or later? GameX requires it. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_DDSETLEVEL_FAILED:
		sprintf (error, "Error: DDSETLEVEL_FAILED (Unable to set DirectX Cooperative Level)") ;
		sprintf (recommend, "Have you installed DirectX 7.0 or later? GameX requires it. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_SETMODE_FAILED:
		sprintf (error, "Error: SETMODE_FAILED (Unable to set screen mode)") ;
		sprintf (recommend, "Have you installed DirectX 7.0 or later? GameX requires it. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_FRONTSURFACE_FAILED:
		sprintf (error, "Error: FRONTSURFACE_FAILED (Unable to create primary surface)") ;
		sprintf (recommend, "Have you installed DirectX 7.0 or later? GameX requires it. Still having trouble? Please report error and system configuration.") ; 	
		break;
	case GAMEX_BACKSURFACE_FAILED:
		sprintf (error, "Error: BACKSURFACE_FAILED (Unable to create secondary surface)") ;
		sprintf (recommend, "Have you installed DirectX 7.0 or later? GameX requires it. Still having trouble? Please report error and system configuration.") ;
		break;
	case GAMEX_REQSIZE_FAILED:
		sprintf (error, "Error: REQSIZE_FAILED (Requested size invalid)") ;
		sprintf (recommend, "The requested game size was invalid.\nThe absolute maximum size allowed is 2048x2047, and the absolute minimum allowed is 64x64.\nTry to request a reasonable size though, such as something between 320x240 and 1600x1200.\nIf you are unsure of screen size, try initializing GameX as follows:\nGameX.Initialize (VIDEO_16BIT, 800, 600);") ;
		break;
	case GAMEX_SOUNDINIT_FAILED:
		sprintf (error, "Error: SOUNDINIT_FAILED (Unable to initialize Direct Sound)") ;
		sprintf (recommend, "GameX was unable to initialize sound. Is sound card installed and working?") ;
		break;
	case GAMEX_SOUNDCOOP_FAILED:
		sprintf (error, "Error: SOUNDCOOP_FAILED (Unable to set Direct Sound Cooperative Level") ;
		sprintf (recommend, "GameX was unable to set Direct Sound cooperative level. Is sound card installed and working?") ;
		break;	
	case GAMEX_SOUNDCREATE_FAILED:
		sprintf (error, "Error: SOUNDCREATE_FAILED (Unable to create Direct Sound buffer.") ;
		sprintf (recommend, "GameX was unable to create a sound buffer for your sound. Try shortening the sound file (total length), reducing the sound resolution (bits per sample), reducing the sound quality (samples per second), or eliminating a channel (mono vs. stereo).") ;
		break;
	case GAMEX_SOUNDMATCH_FAILED:
		sprintf (error, "Error: SOUNDMATCH_FAILED (Loaded sounds do not match.") ;
		sprintf (recommend, "You are attempting to load several sounds with different quality, resolution, and/or channel info. All sounds to be played simultaneously must have the same resolution (bits per sample), same quality (samples per second), and same number of channels (mono vs. stereo). Edit WAV files accordinging.") ;
		break;
	case GAMEX_SOUNDCAPS_FAILED:
		sprintf (error, "Error: SOUNDCAPS_FAILED (Unable to get Sound Buffer Capabilities.") ;
		sprintf (recommend, "GameX was unable to retrieve sound buffer capabilities. You may be running out of memory. Close any open applications. Make sure you have DirectX 7.0 or later.") ;
		break;
	case GAMEX_SOUNDBUFLEN_FAILED:
		sprintf (error, "Error: SOUNDBUFLEN_FAILED (Direct Sound buffer and SoundX buffers do not match.") ;
		sprintf (recommend, "Direct Sound and SoundX buffer lengths do not match. Internal error. Please report this problem and the code that generated it.") ;	
		break;
	case GAMEX_SOUNDLOST_FAIL:
		sprintf (error, "Error: SOUNDLOST_FAIL (Direct Sound buffer was lost.") ;
		sprintf (recommend, "Direct Sound buffer was lost. Make sure you are using DirectX 7.0 or later. Close any open applications. Try restarting system. If you are using a large or high quality sound file, try reducing it.") ;
		break;
	case GAMEX_SOUNDINVALID_FAIL:
		sprintf (error, "Error: SOUNDINVALID_FAIL (Direct Sound Invalid Call.") ;
		sprintf (recommend, "DirectX, via GameX, made an invalid sound call. Make sure you are using DirectX 7.0 or later. Close any open applications. Try restarting system. If you are using a large or high quality sound file, try reducing it.") ;
		break;
	case GAMEX_SOUNDPARAM_FAIL:
		sprintf (error, "Error: SOUNDPARAM_FAIL (Direct Sound Invalid Parameter.") ;
		sprintf (recommend, "DirectX, via GameX, used an invalid parameter. Make sure you are using DirectX 7.0 or later. Close any open applications. Try restarting system. If you are using a large or high quality sound file, try reducing it.") ;
		break;
	case GAMEX_SOUNDPRIOLEVEL_FAIL:
		sprintf (error, "Error: SOUNDPARAM_FAIL (Direct Sound needs Priority Level") ;
		sprintf (recommend, "DirectX, via GameX, is using an invalid priority level. Make sure you are using DirectX 7.0 or later. Close any open applications. Try restarting system. If you are using a large or high quality sound file, try reducing it.") ;		
		break;	
	case GAMEX_IMAGEUSE_FAILED:
		sprintf (error, "Error: IMAGEUSE_FAILED (Image used incorrectly)") ;
		sprintf (recommend, "Purpose of image on load does not match its use. Images loaded with ALPHA must be drawn Draw(ALPHA,...).\nImages loaded with DRAW_ADD must be drawn with Draw(DRAW_ADD,...). Etc.\nIf you want to be able to draw an image in more than one way, combine the usage types on load like this: Load(...,ALPHA | DRAW_ADD)\nThe reason that this usage type setting is required is that the image needs to be prepared differently when loaded depending on what it is going to be used for.\nNote that performance or memory usage may not be optimal for images that are prepared to be able to be drawn in multiple ways.") ;
		break;
	case GAMEX_CLIPPER_FAILED:
		sprintf (error, "Error: CLIPPER_FAILED (Failed to create or set a clipper for the back buffer)") ;
		sprintf (recommend, "Are you using DirectX 7.0 or later? GameX requires one of these. Still having trouble? Please report error and system configuration.") ;
		break;
	default:
		sprintf (error, "Error: UNKNOWN") ;
		sprintf (recommend, "Unregistered error with GameX. Please report. (Unless you caused this error by editing GameX)") ;
		break;
	}
	sprintf (disp, "GameX %s\n", error) ;
	debug.Output (disp, "") ; 
	int choice;
	if(win_status == GAMEX_DDSTART_FAILED) {
		sprintf (disp, "GameX has encountered a fatal error.\n\n%s\n\n%s\n\nIf restarting does not fix this problem, please (re)install DirectX and try again.", error, recommend) ;
		MessageBox (win_hwnd, disp, "GameX", MB_ICONSTOP) ;
		choice = IDNO;
	} else {
		sprintf (disp, "GameX has encountered an error and may not be able to continue.\n\n%s\n\n%s\n\nAttempt to continue execution anyway?", error, recommend) ;
		if(dd_main != NULL)
			win_status = GAMEX_READY;
		choice = MessageBox (win_hwnd, disp, "GameX", MB_YESNO|MB_DEFBUTTON2|MB_ICONSTOP) ;
	}
	if(choice == IDNO)
		ExitProcess(1);
} // insert breakpoint here :-)

// vol: 0.0 means silent, 1.0 means loud, etc.
// pan: -1.0 means left only, 0.0 means center (both), 1.0 means right only, 1.5 means right only half volume, -1.5 means left only half volume, <=-2.0 and >=2.0 mean no volume, etc.
// freq: 1.0 means normal frequency, 2.0 means twice as high-pitched, 0.5 means half pitch, etc.
void WindowsDX::PlaySound (SoundXPtr snd, int how, float vol, float pan, float freq)
{		
	if(sound_support) {
		if(snd->NeedUpdate()==true) UpdateSound(snd) ;	
		int num = snd->GetIndex() ;
		if(ds_sound_list[num] == NULL) return;
		if(win_request_reverse_speakers) pan = -pan;
		if(win_request_no_stereo) pan = 0.0f;
		if(pan > 1.0f) {
			vol *= max(0.0f, 1.0f - (pan-1.0f));
			pan = 1.0f;
		} else if(pan < -1.0f) {
			vol *= max(0.0f, 1.0f - (-pan-1.0f));
			pan = -1.0f;
		}
		if(vol > 1.0f) vol = 1.0f; // DirectSound does not support sound amplification 
		if(vol < 0.0f) vol = 0.0f;
		if(how & PLAY_NOTIFLESSVOL) {
			DWORD status;
			ds_sound_list[num]->GetStatus(&status);
			if(status & DSBSTATUS_PLAYING) {
				long old_vol;
				ds_sound_list[num]->GetVolume(&old_vol);
				if((long)(-5000.0*(1-vol)*(1-vol)) < old_vol)
					return;
			}
		}
		if(how & PLAY_REWIND) ds_sound_list[num]->SetCurrentPosition(0) ;
 		ds_sound_list[num]->SetVolume((long) (-5000.0*(1-vol)*(1-vol))) ;
		ds_sound_list[num]->SetPan((long) (pan*2048.0)) ;
		ds_sound_list[num]->SetFrequency(ds_samples_per_sec*freq) ;
		ds_sound_list[num]->Play(0,0,(how & PLAY_LOOP ? DSBPLAY_LOOPING : 0)) ;
	}
}




// Plays a MusicX object, and immediately stops playing any other music that may be playing
// times of 0 means infinite loop, 1 means play once, 2 means play twice, etc.
// speed_factor of 1.0 means normal speed, 2.0 means double speed, etc.
void WindowsDX::PlayMusic (MusicX * music, int times, float fade_in_seconds, float volume, float tempo, float pitch)
{
	if(music->GetSegment() == NULL) {
#ifdef _DEBUG
		MessageBox(GameX.GetWindow(), "Error: Attempting to play a music file that did not load properly.\nMake sure your music files are in the proper location and in .mid format.\nAlso make sure any music files needed by the game are not open in another program.", "GameX Error",MB_OK|MB_ICONSTOP) ;
#endif
		// if we're not in debug mode, don't bother the player with the music failure message
		return;
	}

	VolumeFadeMusic(music, 0.0f, 0.0f);

	if(music->GetPlayState()) { // if already playing, have to wait for change to take effect
		int clocker = clock();
		while(clock() - clocker < 80) {}
	}

	if(times) times--;
	else times = DMUS_SEG_REPEAT_INFINITE;
	music->GetSegment()->SetRepeats(times);
	SetMusicTempo(tempo);
	IDirectMusicSegmentState * idmss = music->GetSegmentState();
	dm_performance->PlaySegment(music->GetSegment(), 0, 0, &idmss);

	PitchBendMusic(music, pitch,0.0f);
	VolumeFadeMusic(music, volume, fade_in_seconds);
	music->SetPlayState(1);
}

// Fades the volume to 0 over given seconds if seconds>0, then stops the music from playing
void WindowsDX::StopMusic (MusicX * music, float seconds)
{
	if(music->GetSegment() == NULL) return;

	VolumeFadeMusic(music, 0.0f, seconds); // fade for given seconds

	// calculate time until after fade
	DMUS_TEMPO_PARAM tempo;
	music->GetSegment()->GetParam(GUID_TempoParam, 0xFFFF, 0, 0, NULL, &tempo);
	seconds *= tempo.dblTempo * 10;
	MUSIC_TIME musictime;
	dm_performance->GetTime(NULL, &musictime);
	musictime += seconds;

	dm_performance->Stop(music->GetSegment(), NULL, musictime, 0); // stop after fade

	music->SetPlayState(2);
}

// Fades the music up or down to the given volume factor, over the given time
// seconds is how many seconds to alter volume over
// target_factor of 0.0f is fade to silence, 0.5f is fade to half master volume, 1.0f is no change, 1.27f is fade to max amplified
void WindowsDX::VolumeFadeMusic(MusicX * music, float volume_factor, float seconds)
{
	if(seconds < 0.05f) seconds = 0.05f;
// 	if(volume_factor == 0) volume_factor = -0.2f;
	music->Curve(1000*seconds, max(-20,min(127, 100*volume_factor)), DMUS_CURVET_CCCURVE, DMUS_CURVES_LINEAR, max(-20,min(127, 100*music->GetVolume())));
	music->RememberVol(volume_factor);
}

// Bends the pitch of the mucis up or down to the given pitch factor, over the given time
// seconds is how many seconds to alter pitch over. This method does not change speed of play
// pitch_factor of 0.5f is half pitch (low frequency), 2.0f is double pitch (high frequency), 1.0f is no change, etc.
void WindowsDX::PitchBendMusic(MusicX * music, float pitch_factor, float seconds)
{
	if(seconds < 0.05f) seconds = 0.05f;
	music->Curve(1000*seconds, max(1000, min(32767, 8000*pitch_factor)), DMUS_CURVET_PBCURVE, DMUS_CURVES_LINEAR);
}

// Sets the tempo, or speed of play, of all music -- does not change the pitch
// 0.5f is half speed, 1.0f is normal speed, 2.0f is double speed, etc.
// Can be called before or after you start playing the song
// note: there may be a small time delay after calling this function before the change takes effect
void WindowsDX::SetMusicTempo(float tempo_factor)
{
	tempo_factor = min(DMUS_MASTERTEMPO_MAX, max(DMUS_MASTERTEMPO_MIN, tempo_factor));
	dm_performance->SetGlobalParam(GUID_PerfMasterTempo, &tempo_factor, sizeof(float));
	master_music_tempo = tempo_factor;
}

// Sets the volume of all music
// Can be called before or after you start playing a song
// If an individual music song has its volume faded to 50% and the master volume is 50%
// that will result in the music being played at 25% volume
void WindowsDX::SetMasterMusicVolume(float volume_factor)
{
	long volume = (long)(-3000*(1 - volume_factor));
	dm_performance->SetGlobalParam(GUID_PerfMasterVolume, &volume, sizeof(long));
	master_music_volume = volume_factor;
}

// returns true if the given song is still playing, looping, or paused, false otherwise
bool WindowsDX::IsMusicPlaying(MusicX * music)
{
	return(dm_performance->IsPlaying(music->GetSegment(),NULL)==S_OK);
}



bool WindowsDX::IsSoundPlaying(SoundXPtr snd)
{
	if(snd->NeedUpdate()==true) UpdateSound(snd) ;	
	int num = snd->GetIndex() ;
	if(ds_sound_list[num] == NULL) return false;

	DWORD status;
	ds_sound_list[num]->GetStatus(&status);
	if(status & DSBSTATUS_PLAYING)
		return true;
	else
		return false;
}

void WindowsDX::RewindSound (SoundXPtr snd)
{
	if (sound_support) {
		if (snd->NeedUpdate()==true) UpdateSound (snd) ;	
		int num = snd->GetIndex() ; 
		ds_sound_list[num]->SetCurrentPosition(0) ;
	}
}


void WindowsDX::StopSound (SoundXPtr snd)
{
	if (sound_support) {
		int num = snd->GetIndex() ;
		if (num>=0) ds_sound_list[num]->Stop() ;
	}
}

void WindowsDX::UpdateSound (SoundXPtr snd)
{
	signed short *buf_a, *buf_b;	
	unsigned long bytes_a, bytes_b; 

	int num_samples, num_channels;
	int result, num;
	int n, c;
	
	// Determine Direct Sound buffer to use based on SoundX object.
	// (If Direct Sound buffer does not exist for SoundX object, create one)
	num = snd->GetIndex() ;
	if (num==-1) {
		num = AddSound (snd) ;
		snd->SetIndex (num) ;
	}

	if(ds_sound_list[num]==NULL) return;

	// Get requested Direct Sound buffer capabilities
	DSBCAPS caps;
	ZeroMemory(&caps, sizeof(caps)) ;
	caps.dwSize = sizeof(caps) ;
	if (DS_OK != ds_sound_list[num]->GetCaps (&caps)) ReportProblem(GAMEX_SOUNDCAPS_FAILED) ;
		
	// Get and check number of samples (SoundX sample count should match DS buffer count)
	num_channels = snd->GetNumChannels() ;
	num_samples = caps.dwBufferBytes / num_channels;

	// causes legitimate sounds (such as ones that DirectSound changed the length of) to fail to play.
//	if (num_samples!=snd->GetNumSamples()) ReportProblem(GAMEX_SOUNDBUFLEN_FAILED) ;

	if (num_samples!=snd->GetNumSamples()) num_samples = snd->GetNumSamples();
		
	
	// Lock Direct Sound buffer
	result = ds_sound_list[num]->Lock (0, caps.dwBufferBytes, (void**) &buf_a, &bytes_a, (void**) &buf_b, &bytes_b, 0) ;
	n = 0;
	while (result==DSERR_BUFFERLOST && n<10) {
		ds_sound_list[num]->Restore() ;
		result = ds_sound_list[num]->Lock (0, caps.dwBufferBytes, (void**) &buf_a, &bytes_a, (void**) &buf_b, &bytes_b, 0) ;
		n++;
	}
	// Check for possible buffer errors
	if (DS_OK!=result) {		
		if (result==DSERR_BUFFERLOST)			win_status = GAMEX_SOUNDLOST_FAIL;	
		else if (result==DSERR_INVALIDCALL) 	win_status = GAMEX_SOUNDINVALID_FAIL;
		else if (result==DSERR_INVALIDPARAM)	win_status = GAMEX_SOUNDPARAM_FAIL;
		else if (result==DSERR_PRIOLEVELNEEDED) win_status = GAMEX_SOUNDPRIOLEVEL_FAIL;
		ReportProblem() ;
	}
	// Copy SoundX data to Direct Sound buffer
	if (snd->GetBPS()<=8) {
		unsigned char *data = (unsigned char*) snd->GetData() ;
		unsigned char *curr_a = (unsigned char*) buf_a;
		for (n = 0; n < num_samples; n++)
			for (c=0; c < num_channels; c++)
				*curr_a++ = *data++;
	} else {
		signed short *data = (signed short*) snd->GetData() ;
		signed short *curr_a = buf_a;
		curr_a = buf_a;
		for (n = 0; n < num_samples; n++)
			for (c=0; c < num_channels; c++) 
				*curr_a++ = *data++;			
	}
	
	// Unlock Direct Sound buffer
	ds_sound_list[num]->Unlock (&buf_a, bytes_a, &buf_b, bytes_b) ;

	snd->SetRefresh (false) ;
}

int WindowsDX::AddSound (SoundXPtr snd)
{
	SoundXPtr*				new_list;
	LPDIRECTSOUNDBUFFER*	new_ds_list;
	
	if (num_sounds==max_sounds) {
		max_sounds += 10;
		new_list = new SoundXPtr[max_sounds] ;
		new_ds_list = new LPDIRECTSOUNDBUFFER[max_sounds] ;
		memcpy (new_list, sound_list, sizeof(SoundXPtr)*num_sounds) ;
		memcpy (new_ds_list, ds_sound_list, sizeof(LPDIRECTSOUNDBUFFER)*num_sounds) ;
		if (sound_list!=NULL) {delete [] sound_list;}
		if (ds_sound_list!=NULL) {delete [] ds_sound_list;}
		sound_list = new_list;
		ds_sound_list = new_ds_list;
	}	
	sound_list[num_sounds] = snd;
	
	if (num_sounds==0) {
		ds_num_channels = snd->GetNumChannels() ;
		ds_bits_per_sample = snd->GetBPS() ;
		ds_samples_per_sec = snd->GetSamplesPerSec() ;
	} else {
		if (snd->GetNumChannels()!=ds_num_channels) ReportProblem (GAMEX_SOUNDMATCH_FAILED) ;
		if (snd->GetBPS()!=ds_bits_per_sample) ReportProblem (GAMEX_SOUNDMATCH_FAILED) ;
		if (snd->GetSamplesPerSec()!=ds_samples_per_sec) ReportProblem (GAMEX_SOUNDMATCH_FAILED) ;
	}

	ds_sound_list[num_sounds] = NULL;
	
	WAVEFORMATEX format; 
	ZeroMemory(&format,sizeof(format)) ;		
	format.wFormatTag = WAVE_FORMAT_PCM;
	format.nChannels = ds_num_channels;
	format.nSamplesPerSec = ds_samples_per_sec;
	format.wBitsPerSample = ds_bits_per_sample;
	format.nBlockAlign = (format.wBitsPerSample / 8) * format.nChannels; 
	format.nAvgBytesPerSec = format.nSamplesPerSec * format.nBlockAlign;	
	
	DSBUFFERDESC bufferDesc; 
	ZeroMemory(&bufferDesc,sizeof(bufferDesc)) ; 
	bufferDesc.dwSize = sizeof(bufferDesc) ; 
	bufferDesc.dwFlags = DSBCAPS_CTRLFREQUENCY | DSBCAPS_CTRLPAN | DSBCAPS_CTRLVOLUME; 
	bufferDesc.dwBufferBytes = snd->GetNumSamples() * format.nBlockAlign;
	bufferDesc.lpwfxFormat = (LPWAVEFORMATEX) &format;

	if (FAILED (ds_main->CreateSoundBuffer(&bufferDesc, &ds_sound_list[num_sounds], NULL))) {		
		ReportProblem (GAMEX_SOUNDCREATE_FAILED) ;
	}
	
	num_sounds++;
	return num_sounds-1;
}

